/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

/**
 * Direction restricted tile (startable).
 * Note: At least 1 and at most 3 directions can be restricted.
 * @author V.Termanis
 */
public class ArwTile extends StartableTile {

  /**
   * Indicates upward/north direction.
   */
  public static final int UP = 0;
  /**
   * Indicates downward/south direction.
   */
  public static final int DOWN = 1;
  /**
   * Indicates left/west direction.
   */
  public static final int LEFT = 2;
  /**
   * Indicates right/east direction.
   */
  public static final int RIGHT = 3;
  
  /**
   * Details which directions are restricted.
   */
  private boolean[] directions;


  /**
   * Constructs a new <code>ArwTile</code> instance with the given attributes.
   * @param height of the pile
   * @param levelStart indicates whether this is a start
   * @param directions details which directions are restricted
   * @throws NullPointerException if <code>directions</code> is
   * <code>null</code>.
   * @throws TileException if either <code>height</code> or the combination of
   * of restricted directions is invalid.
   */
  public ArwTile(int height, boolean levelStart, boolean[] directions)
    throws NullPointerException, TileException {
    this.height = height;
    tileCheck();
    setStart(levelStart);
    if (directions == null || directions.length != 4) {
      throw new TileException("Invalid directions array length");
    }
    this.directions = new boolean[4];
    System.arraycopy(directions, 0, this.directions, 0, 4);
    arwCheck();
    type = ARROW;
  }


  /**
   * Used to check for direction restriction validity upon creation.
   * @throws NullPointerException if <code>directions</code> is
   * <code>null</code>.
   * @throws TileException if combination of restricted directions is invalid.
   */
  private void arwCheck() throws NullPointerException, TileException {
    if (directions == null) {
      throw new NullPointerException("directions parameter is null");
    }
    if ((directions[0] && directions[1] && directions[2] && directions[3]) ||
	!(directions[0] || directions[1] || directions[2] || directions[3])) {
      throw new TileException("ArwTile: " + 
			      "Invalid tile directions (either all or none)");
    }
  }
  

  /**
   * Check whether a particular direction is not restricted.
   * @param dir one of <code>UP</code>, <code>DOWN</code>, <code>LEFT</code> or
   * <code>RIGHT</code>
   * @return <code>true</code> if the direction is allowed.
   */
  public boolean allowsDirection(int dir) {
    return directions[dir];
  }


  /**
   * Used to compare two tiles, ignoring their respective heights.
   * @param t the <code>Tile</code> instance to compare the current tile with.
   * @return <code>true</code> if the two <code>Tile</code> instances
   * are equal disregarding their heights.
   */
  public boolean equalsApartFromHeight(Tile t) {
    if (t.getType() == ARROW) {
      ArwTile aT = (ArwTile) t;
      return (levelStart == aT.isStart() &&
	      directions[UP] == aT.allowsDirection(UP) &&
	      directions[DOWN] == aT.allowsDirection(DOWN) &&
	      directions[LEFT] == aT.allowsDirection(LEFT) &&
	      directions[RIGHT] == aT.allowsDirection(RIGHT));
    } else {
      return false;
    }
  }


  /**
   * Creates an exact (non-shallow) copy of the current tile
   * @return a copy of the <code>Tile</code> instance.
   */
  public Tile copy() {
    try {
      return new ArwTile(height, levelStart, directions);
      // this should never happen
    }	catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

}