/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.util.Random;

/**
 * Generates levels for one and two-player challenge modes.
 * @author V.Termanis
 */
public class Challenge {

  // limitins gaussian distribution range
  private static final double DEVIATIONS = 2.5;


  /**
   * Generates a new set of levels with the given attributes.
   * @param numberOfLevels the number of levels to generate
   * @param minTileCount number of tiles required in first level
   * @param minTileCountIncrement increase of tiles per level. For example
   * level 10 would have <code>minTileCount +
   * (minTileCountIncrement * 9)</code> tiles.
   * @param grid width and height of all levels
   * @param offset from outside of grid which is not to be used. For example
   * if the grid is 15x15 and the offset has been specified as 2x2 then only
   * the 11x11 area in the middle of the 15x15 grid will be used
   * @param maxTileHeight maximum height of any pile
   * @param forceGenRetries number of times to regenerate level if it does
   * not have enough tiles (this can happen if the current position is
   * surrounded by piles of height <code>maxTileHeight</code>.
   * @param cooperative if set, a two-player levelset will be generated
   * @return the generated levelset
   * @throws IllegalArgumentException if <code>offset</code> is too large for
   * the specified <code>grid</code>.
   */
  public static LevelSet genSet(int numberOfLevels, int minTileCount,
				int minTileCountIncrement, Dimension grid,
				Point offset, int maxTileHeight,
				int forceGenRetries, boolean cooperative)
    throws IllegalArgumentException {

    if (grid.width <= offset.x * 2 || grid.height <= offset.y) {
      throw new IllegalArgumentException("Offset too large");
    }

    Random r = new Random();
    LevelSet set = new LevelSet(cooperative ? "2P Challenge" : "Challenge",
				grid.width, grid.height, maxTileHeight,
				cooperative);
    int levelSize = minTileCount;

    for (int i = 0; i < numberOfLevels; i++) {
      int retries = 0;
      set.addLevel(genLevel(grid, offset, levelSize, maxTileHeight,
			    cooperative, r));
      // if not exact tile number in generated level, try again
      while (set.getLevel(i).size() != levelSize &&
      	     retries < forceGenRetries) {
	set.removeLevel(i);
	set.addLevel(genLevel(grid, offset, levelSize, maxTileHeight,
			      cooperative, r));
	retries++;
      }
      levelSize += minTileCountIncrement;
    }
    return set;
  }


  /**
   * Generates a new set of levels with the given attributes and a square grid.
   * @param numberOfLevels the number of levels to generate
   * @param minTileCount number of tiles required in first level
   * @param minTileCountIncrement increase of tiles per level. For example
   * level 10 would have <code>minTileCount +
   * (minTileCountIncrement * 9)</code> tiles.
   * @param gridSize width and height of all levels
   * @param offset from outside of grid which is not to be used. For example
   * if the grid is 15x15 and the offset has been specified as 2 then only
   * the 11x11 area in the middle of the 15x15 grid will be used
   * @param maxTileHeight maximum height of any pile
   * @param forceGenRetries number of times to regenerate level if it does
   * not have enough tiles (this can happen if the current position is
   * surrounded by piles of height <code>maxTileHeight</code>.
   * @param cooperative if set, a two-player levelset will be generated
   * @return the generated levelset
   * @throws IllegalArgumentException if <code>offset</code> is too large for
   * the specified <code>grid</code>.
   */
  public static LevelSet genSet(int numberOfLevels, int minTileCount,
				int minTileCountIncrement, int gridSize,
				int offset, int maxTileHeight,
				int forceGenRetries, boolean cooperative) {
    return genSet(numberOfLevels, minTileCount, minTileCountIncrement,
		  new Dimension(gridSize, gridSize), new Point(offset, offset),
		  maxTileHeight, forceGenRetries, cooperative);
  }


  /**
   * Generates a new level with the given attributes.
   * @param grid width and height of the level
   * @param offset from outside of grid which is not to be used. For example
   * if the grid is 15x15 and the offset has been specified as 2 then only
   * the 11x11 area in the middle of the 15x15 grid will be used
   * @param tileCount number of tiles the level should have. Note: The level
   * can end up with less than the specified number of tiles. This is not
   * very likely though, unless <code>tileCount</code> is very large and
   * and <code>maxTileHeight</code> fairly low.
   * @param maxTileHeight maximum height of any pile
   * @param cooperative if set, a two-player level will be generated
   * @param r instance of <code>Random</code> required to use
   * <code>Random.nextGaussian()</code> method
   * @return the generated level
   */
  public static Level genLevel(Dimension grid, Point offset, int tileCount,
			       int maxTileHeight, boolean cooperative,
			       Random r) {
    Level lvl = new Level(grid.width, grid.height, maxTileHeight,
			  (cooperative ? 30 : 60), cooperative);
    Point nextP, currentP = randomGridPosition(grid, offset, r);
    int tileCounter = 0, direction = 1;
    

    // add start
    lvl.pushTile(currentP, true);
    while (tileCounter <= tileCount &&
	   canStillPush(lvl, currentP, grid, offset, maxTileHeight)) {
      nextP = randomGridPosition(grid, offset, r);
      // vertical
      direction = (currentP.y < nextP.y) ? 1 : -1;
      while (currentP.y != nextP.y) {
	currentP.translate(0, direction);
	// stop advance tile with maximum height reached
	if (!lvl.pushTile(currentP, false)) {
	  currentP.translate(0, -direction);
	  break;
	}
	tileCounter++;
	// make sure at most <tileCount> tiles are laid
	if (tileCounter > tileCount) {
	  finaliseLevel(lvl, currentP, cooperative);
	  return lvl;
	}
      }
      // horizontal
      direction = (currentP.x < nextP.x) ? 1 : -1;
      while (currentP.x != nextP.x) {
	currentP.translate(direction, 0);
	// stop advance tile with maximum height reached
	if (!lvl.pushTile(currentP, false)) {
	  currentP.translate(-direction, 0);
	  break;
	}
	tileCounter++;
	// make sure at most <tileCount> tiles are laid
	if (tileCounter > tileCount) {
	  finaliseLevel(lvl, currentP, cooperative);
	  return lvl;
	}
      }
    }
    finaliseLevel(lvl, currentP, cooperative);
    return lvl;
  }
  

  /**
   * Adds start for 2nd player to a level (used by <code>genLevel</code>)
   */
  private static void finaliseLevel(Level l, Point p, boolean cooperative) {
    // end of generated level = start for 2nd player
    if (cooperative) {
      l.pushTile(p, true);
      // or remove one tile if single player (otherwise will have too many)
    } else {
      l.popTile(p);
    }
  }
  

  /**
   * Returns <code>true</code> if it still possible to add more tile around
   * the given position.
   */
  private static boolean canStillPush(Level l, Point p, Dimension grid,
				      Point offset, int maxHeight) {
    if ((p.x == offset.x || tileHeight(l, p.x - 1, p.y) == maxHeight) &&
	(p.y == offset.y || tileHeight(l, p.x, p.y - 1) == maxHeight) &&
	(p.x == grid.width - 1 || tileHeight(l, p.x + 1, p.y) == maxHeight) &&
	(p.y == grid.height - 1 || tileHeight(l, p.x, p.y + 1) == maxHeight)) {
      return false;
    } else {
      return true;
    }
  }


  /**
   * Simplies tile height check by automatically returning 0 if it grid
   * position is <code>null</code>.
   */
  private static int tileHeight(Level l, int x, int y) {
    if (l.tile(x, y) == null) {
      return 0;
    } else {
      return l.tile(x, y).getHeight();
    }
  }

  
  /**
   * Returns random position on grid with the given offset.
   */
  private static Point randomGridPosition(Dimension grid, Point offset,
					  Random r) {
    return new Point((int) (gaussianDouble(r, DEVIATIONS) *
			    (grid.width - (offset.x * 2))) +
		     offset.x,
		     (int) (gaussianDouble(r, DEVIATIONS) *
			    (grid.height - (offset.y * 2))) +
		     offset.y);
  }


  /**
   * Maps Random.nextGaussian() to (0,1) (non-inclusive)
   * devationsFromMean specifies how bigger range of values to map from
   * (nextGaussian() uses mean 0 and standard deviation 1)
   */
  private static double gaussianDouble(Random r, double deviationsFromMean) {
    double d;
    // discard values outside specified range
    while (Math.abs(d = r.nextGaussian()) >= deviationsFromMean);
    return (((d / deviationsFromMean) + 1) / 2);
  }

}