/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.swing.*;
import java.io.*;


/**
 * Defines function and rendering of the editor component.
 * @author V.Termanis
 */
public class Editor {

  // possible mouseover states
  public static final int NONE = -1;
  public static final int NORMAL = 0;
  public static final int ICE = 1;
  public static final int TELEPORT = 2;
  public static final int GRAVITY = 3;
  public static final int ARROW_UP = 4;
  public static final int ARROW_DOWN = 5;
  public static final int ARROW_LEFT = 6;
  public static final int ARROW_RIGHT = 7;
  public static final int START = 8;
  public static final int START2 = 9;
  public static final int LEVEL_SWAP = 10;
  public static final int LEVEL_FIRST = 11;
  public static final int LEVEL_LAST = 12;
  public static final int LEVEL_MESSAGES = 13;
  public static final int LEVEL_BACK = 14;
  public static final int LEVEL_FORWARD = 15;
  public static final int LEVEL_TIME = 16;
  public static final int LEVEL_CLEAR = 17;
  public static final int LEVEL_PLAY = 18;
  public static final int SAVE = 19;
  public static final int CLOSE = 20;
  public static final int SAVECLOSE = 21;
  public static final int GRID = 22;

  // button images
  private Images imgData;
  private TilePainter tPainter;
  // locations, dimensions
  private Rectangle[] buttonLocations;
  private ScreenFrames frames;
  private Rectangle levelSetName, levelNumberBg, levelNumber1s, levelNumber10s;
  private Point levelSetNameOffset, levelNumber1sOffset, levelNumber10sOffset,
    levelNumberMiddleOffset;
  // tile piling height, row and column count
  private int tileHeightLimit, gridSize;
  private Color gridColor, gridHighlight, txtColor, buttonBorderColor;
  private Font txtFont;
  // drawing clip dimensions for next update
  private Rectangle clip, allFieldClipArea, raisedGridClip;
  // button & grid states
  private int selectedButton, activeButton, pressedButton, oldButton;
  private Point activeTile, pressedTile, oldTile, teleportTile;
  private int buttonSpacing;

  private boolean
  // indicates either activeTile or oldTile has teleport destination displayed
    displayTeleportLine,
  /* indicates whether level set will be saved or not (also stops save buttons
   * from being displayed
   */
    readOnly,
  /* used to determine whether to ask for user confirmation when exiting
   * with possible pending changes to one of the levels
  */
    changedSinceLastSave,
  // using mouse first time after key presses requires full redraw
    forceRedraw;
  // up, down, left, right - for arrow tile
  private boolean[] selectedArrows; 
  // level data
  private LevelSet levelSet;
  // possibly shortened string if name too wide to fit on screen
  private String shortLevelSetName;
  // file details required for saving to disk
  private String levelSetFilename, fileExtension;
  // active level and location of it in level set
  private Level currentLevel;
  private int
    currentLevelIndex,
  /* left & right mouse buttons cannot be used at the same time, so disregard
   * other mouse press until mouse button has been released
   */
    mouseDown,
    buttonBorderThickness;
  // pop-up messages
  private JOptionPane optionPane;
  /* to select "No" as the active button in a dialog a custom list of options
   * has to be used (limitation of JOptionPane)
   */
  private String[] yesNo = {"Yes", "No"};
  private BasicStroke buttonBorderStroke;
  // so parent query for test level and mode
  private LevelSet testSet;


  /**
   * Creates a new <code>Editor</code> instance with the specified parameters.
   * @param imgData access to global image resources
   * @param frames responsible for background and frame painting
   * @param tPainter responsible for drawing grid and tiles
   * @param gridSize width and height of grid (e.g. 15x15)
   * @param tileHeightLimit maximum height a pile can have
   * @param buttonSpacing horizontal and vertical spacing inbetween buttons
   * @param buttonBorderThickness thickness of highlighted button border
   * @param buttonBorderColor colour of highlighted button border
   * @param txtFont style of all text (except tile numbers)
   * @param gridColor colour of grid drawn underneath tiles
   * @param gridHighlight colour of highlighted grid position
   * @param txtColor colour of all text (except tile numbers)
   * @param g graphics context for rendering
   * @param fileExtension so upon new levelset creation the correct extension
   * is added if required
   * @param levelSetFilename a levelset to load initially, can be
   * <code>null</code>
   * @throws IOException if the specified levelset cannot be loaded/created
   */
  public Editor(Images imgData, ScreenFrames frames, TilePainter tPainter, 
		int gridSize, int tileHeightLimit, int buttonSpacing,
		float buttonBorderThickness, Color buttonBorderColor,
		Font txtFont, Color gridColor, Color gridHighlight,
		Color txtColor, Graphics2D g, String fileExtension,
		String levelSetFilename) throws IOException { 
    activeButton = pressedButton = oldButton = NONE;
    this.imgData = imgData;
    this.frames = frames;
    this.tPainter = tPainter;
    this.gridSize = gridSize;
    this.tileHeightLimit = tileHeightLimit;
    this.buttonSpacing = buttonSpacing;
    this.txtFont = txtFont;
    this.buttonBorderColor = buttonBorderColor;
    this.gridColor = gridColor;
    this.gridHighlight = gridHighlight;
    this.txtColor = txtColor;
    this.fileExtension = fileExtension;
    this.buttonBorderThickness = Math.round(buttonBorderThickness);
    buttonBorderStroke = new BasicStroke(buttonBorderThickness);
    optionPane = new JOptionPane();
    optionPane.setOptionType(JOptionPane.OK_CANCEL_OPTION);
    optionPane.setWantsInput(true);
    selectedArrows = new boolean[4];
    if (levelSetFilename != null) {
      loadLevelSet(levelSetFilename, null, g);
    }
    mouseDown = NONE;
    testSet = null;
  }
  

  /**
   * Loads a levelset from disk or creates a new one if it does not exist.
   * @param filename absolute path for new/existing levelset
   * @param c parent component for drawing dialogs if required
   * @param g graphics context for rendering
   * @throws IOException if either the levelset could not be loaded or a new
   * one could not be created.
   */
  public void loadLevelSet(String filename, Component c, Graphics2D g)
    throws IOException {
    File f = new File(filename);
    levelSetFilename = filename;
    readOnly = false;
    // LOAD EXISTING
    if (f.exists()) {
      readOnly = !f.canWrite();
      try {
	levelSet = LevelSet.load(filename);
      } catch (ClassNotFoundException e) {
	throw new IOException("Invalid levelset file");
      }
      currentLevel = levelSet.getLevel(0);
      // check whether the level is a development level
      readOnly = levelSet.readOnly();
      // CREATE NEW ONE
    } else {
      // test whether writing will be possible (throws IOException if not)
      f.createNewFile();
      f.delete();
      // Remove file extension for level set name (only from end)
      String setName = f.getName();
      int i =
	(new String(setName)).toLowerCase().lastIndexOf("." + fileExtension);
      if (i > 0 && i == setName.length() - (fileExtension.length() + 1)) {
	setName = setName.substring(0, i);
      }
      // ask whether 2 or 1 player set wanted (if not possible default to 1P)
      boolean twoPlayer = false;
      if (c != null) {
	twoPlayer =
	  (JOptionPane.showOptionDialog(c, "Make new level set two-player?",
					"Tile Pile - Editor",
					JOptionPane.YES_NO_OPTION,
					JOptionPane.QUESTION_MESSAGE,
					imgData.getIcon(Images.O_QUESTION),
					yesNo, yesNo[1])
	   == JOptionPane.YES_OPTION);
      }
      levelSet = new LevelSet(setName, gridSize, gridSize, tileHeightLimit,
			      twoPlayer);
      currentLevel = levelSet.addLevel();
    }
    // restet controls for new level set
    selectedButton = NORMAL;
    resetArrows();
    currentLevelIndex = 0;
    calculateGeometry(g);
    changedSinceLastSave = false;
  }


  /**
   * Handles mouse (up) events based on last mouse location from last
   * <code>updateUI</code> call.
   * @param button the mouse button which was released
   * @param screen parent component so can draw dialogs if required
   * @return the performed function or <code>NONE</code>
   */
  public int clickResult(int button, Component screen) {
    // disregard other mouse button presses (if several at once)
    if (mouseDown != button) {
      return NONE;
    } else {
      mouseDown = NONE;
    }

    if (pressedButton != NONE && pressedButton == activeButton) {
      pressedButton = NONE;

      // LEVEL EDITING
      if (activeButton == GRID) {
	// SINGLE SELECT
	if (pressedTile.equals(activeTile)) {
	  pressedTile = null;
	  if (button == MouseEvent.BUTTON1) {
	    return addTile();
	  } else if (button == MouseEvent.BUTTON3) {
	    if (teleportTile == null) {
	      currentLevel.popTile(activeTile);
	      changedSinceLastSave = true;
	    } else {
	      teleportTile = null;
	    }
	  }
	  // MULTI SELECT
	} else {
	  if (button == MouseEvent.BUTTON1 &&
	      !(selectedButton == TELEPORT || selectedButton == START ||
		selectedButton == START2)) {
	      currentLevel.manipulateMultiple(pressedTile, activeTile,
					generateTile());
	      changedSinceLastSave = true;	      
	  } else if (button == MouseEvent.BUTTON3) {
	    if (teleportTile == null) {
	      currentLevel.manipulateMultiple(pressedTile, activeTile, null);
	      changedSinceLastSave = true;
	    } else {
	      teleportTile = null;
	    }
	  }
	  pressedTile = null;
	}
      } else if (button == MouseEvent.BUTTON1) {
	// SWAP LEVEL (with next in set, if any)
	if (activeButton == LEVEL_SWAP) {
	  if (levelSet.swapLevel(currentLevelIndex)) {
	    currentLevel = levelSet.getLevel(currentLevelIndex);
	    changedSinceLastSave = true;
	  }
	} else if (activeButton == LEVEL_FIRST) {
	  if (currentLevelIndex > 0) {
	    currentLevelIndex = 0;
	    currentLevel = levelSet.getLevel(currentLevelIndex);
	  } else {
	    return NONE;
	  }
	} else if (activeButton == LEVEL_LAST) {
	  if (currentLevelIndex < levelSet.getSize() - 1) {
	    currentLevelIndex = levelSet.getSize() - 1;
	    currentLevel = levelSet.getLevel(currentLevelIndex);
	  } else {
	    return NONE;
	  }
	  // PREVIOUS LEVEL
	} else if (activeButton == LEVEL_BACK) {
	  if (currentLevelIndex > 0) {
	    currentLevelIndex--;
	    currentLevel = levelSet.getLevel(currentLevelIndex);
	  } else {
	    return NONE;
	  }
	  // NEXT LEVEL
	} else if (activeButton == LEVEL_FORWARD) {
	  if (currentLevelIndex < levelSet.getSize() - 1) {
	    currentLevelIndex++;
	    currentLevel = levelSet.getLevel(currentLevelIndex);
	    /* a) only create new level if the current one is valid
	     * b) only allow a maximum of 99 levels in a set
	     */
	  } else if (currentLevel.isValid() && currentLevelIndex < 98) {
	    currentLevel = levelSet.addLevel();
	    currentLevelIndex++;
	  } else {
	    return NONE;
	  }
	  // LEVEL MESSAGES
	} else if (activeButton == LEVEL_MESSAGES) {
	  setLevelMessages(screen);
	  // LEVEL TIME LIMIT
	} else if (activeButton == LEVEL_TIME) {
	  setLevelTimeLimit(screen);
	  // TEST-PLAY LEVEL
	} else if (activeButton == LEVEL_PLAY) {
	  if (currentLevel.isValid()) {
	    // create level set containing just the current level
	    testSet = new LevelSet("Level test", 15, 15, 5,
				   levelSet.isTwoPlayer());
	    testSet.addLevel(currentLevel);
	  } else {
	    JOptionPane.showMessageDialog(screen,
					  "This level is not valid and thus" +
					  " cannot be tested.",
					  "Tile Pile - Editor",
					  JOptionPane.ERROR_MESSAGE,
					  imgData.getIcon(Images.O_ERROR));
	    return NONE;
	  }
	  // CLEAR LEVEL
	} else if (activeButton == LEVEL_CLEAR) {
	  if (!currentLevel.isEmpty()) {
	    currentLevel.clear();
	    changedSinceLastSave = true;
	  }
	  // CLOSE EDITOR (without saving)
	} else if (activeButton == CLOSE) {
	  // warn user they might not have saved their changes
	  if (!readOnly && changedSinceLastSave &&
	      JOptionPane.showOptionDialog(screen, "One or more levels " +
					   "might have changed. Exit anyway?",
					   "Tile Pile - Editor",
					   JOptionPane.YES_NO_OPTION,
					   JOptionPane.WARNING_MESSAGE,
					   imgData.getIcon(Images.O_WARNING),
					   yesNo, yesNo[1])
	      == JOptionPane.NO_OPTION) {
	    return NONE;
	  }
	  // SAVE (& CLOSE - parent will close editor if requested)
	} else if ((activeButton == SAVE || activeButton == SAVECLOSE) &&
		   !readOnly) {
	  int[] invalidLevels;
	  if ((invalidLevels = levelSet.getInvalidLevelIndices()) != null) {
	    // create string of invalid levels for display in dialog
	    String invalidList = Integer.toString(invalidLevels[0] + 1);
	    for (int i = 1; i < (invalidLevels.length - 1); i++) {
	      invalidList += ", " + Integer.toString(invalidLevels[i] + 1);
	    }
	    /* last two level separate with an "and", also pre/append
	     * singular/plural specific noun & verb
	     */
	    if (invalidLevels.length > 1) {
	      invalidList = "Levels " + invalidList + " and " +
		Integer.toString(invalidLevels[invalidLevels.length - 1] + 1) +
		" are";
	    } else {
	      invalidList = "Level " + invalidList + " is";
	    }
	    if (JOptionPane.showOptionDialog(screen, invalidList +
					     " not valid.\nSaving now will " +
					     "cause the listed level(s) to " +
					     "be removed. Save anyway?",
					     "Tile Pile - Editor",
					     JOptionPane.YES_NO_OPTION,
					     JOptionPane.WARNING_MESSAGE,
					     imgData.getIcon(Images.O_WARNING),
					     yesNo, yesNo[1])
		== JOptionPane.NO_OPTION) {
	      return NONE;
	    }
	    levelSet.purgeInvalidLevels();
	    // make sure there are no empty levels left
	  } else {
	    levelSet.purgeEmptyLevels();
	  }
	  changedSinceLastSave = false;
	  if (levelSet.getSize() == 0) {
	    JOptionPane.showMessageDialog(screen,
					  "No levels to save.",
					  "Tile Pile - Editor",
					  JOptionPane.INFORMATION_MESSAGE,
					  imgData.getIcon(Images.O_INFO));
	  } else {
	    // write to disk
	    boolean successful;
	    try {
	      successful = LevelSet.saveToDisk(levelSet, levelSetFilename,
					       true);
	      // should never happen due to prior checks
	    } catch (IOException e) {
	      successful = false;
	    }
	    if (!successful) {
	      JOptionPane.showMessageDialog(screen, "Could not write to " +
					    levelSetFilename,
					    "Tile Pile - Editor",
					    JOptionPane.ERROR_MESSAGE,
					    imgData.getIcon(Images.O_ERROR));
	    }
	  }
	  // make sure not still displaying a purged level
	  if (activeButton == SAVE) {
	    // create a new blank level if all have been purged
	    if (levelSet.getSize() == 0) {
	      currentLevelIndex = 0;
	      currentLevel = levelSet.addLevel();
	    } else {
	      // make sure level not out of range after purging
	      if (currentLevelIndex >= levelSet.getSize()) {
		currentLevelIndex = levelSet.getSize() - 1;
	      }
	      currentLevel = levelSet.getLevel(currentLevelIndex);
	    }
	  }
	  // ARROW BUTTON SELECTED
	} else if (isArrow(activeButton)) {
	  setArrows();
	  // OTHER BUTTON SELECTED
	} else if (activeButton != NONE &&
		   !(activeButton == START2 && !levelSet.isTwoPlayer())) {
	  if ((activeButton != START && activeButton != START2) ||
	      selectedButton == activeButton) {
	    resetArrows();
	  }
	  selectedButton = activeButton;
	}
      }
      return activeButton;
      // click release button was not same as click down button
    } else {
	pressedButton = NONE;
	pressedTile = null;
	return NONE;
      }
  }

  
  /**
   * Handles mouse (down) events.
   * @param button which button has been pressed (down)
   * @return <code>true</code> if a screen update is now required.
   */
  public boolean mousePressed(int button) {
    if (mouseDown == NONE) {
      if (button == MouseEvent.BUTTON1 || button == MouseEvent.BUTTON3) {
	mouseDown = button;
	pressedButton = activeButton;
	if (activeButton == GRID) {
	  pressedTile = activeTile;
	  /* special case - right-click during teleport tile creation
	   * requires removal of TP line - request redraw
	   */
	  if (button == MouseEvent.BUTTON3 && teleportTile != null) {
	    teleportTile = null;
	    return true;
	  }
	}
      }
    }
    return false;
  }

  
  /**
   * Processes a keypress.
   * @param keyCode one of the codes as defined in the <code>KeyEvent</code>
   * class.
   * @param screen requires parent component so can draw dialogs if required
   * @return the performed action or <code>NONE</code>
   */
  public int keyResult(int keyCode, Component screen) {
    // ignore keyboard input while mouse pressed
    if (mouseDown != NONE) {
      return NONE;
    } else {
      /* "fake" mouse-down event as ultimately keypress will be handled
       * like a mouse press
       */
      mouseDown = MouseEvent.BUTTON1;
    }

    oldButton = activeButton;

    switch (keyCode) {
    case KeyEvent.VK_LEFT:
      activeButton = LEVEL_BACK;
      break;
    case KeyEvent.VK_RIGHT:
      activeButton = LEVEL_FORWARD;
      break;
    case KeyEvent.VK_UP:
      activeButton = LEVEL_LAST;
      break;
    case KeyEvent.VK_DOWN:
      activeButton = LEVEL_FIRST;
      break;
    case KeyEvent.VK_ESCAPE:
      activeButton = LEVEL_CLEAR;
      break;
    case KeyEvent.VK_ENTER:
      activeButton = LEVEL_SWAP;
      break;
    case KeyEvent.VK_P:
      activeButton = LEVEL_PLAY;
      break;
    case KeyEvent.VK_N:
      activeButton = NORMAL;
      break;
    case KeyEvent.VK_I:
      activeButton = ICE;
      break;
    case KeyEvent.VK_W:
      activeButton = ARROW_UP;
      break;
    case KeyEvent.VK_A:
      activeButton = ARROW_LEFT;
      break;
    case KeyEvent.VK_S:
      activeButton = ARROW_DOWN;
      break;
    case KeyEvent.VK_D:
      activeButton = ARROW_RIGHT;
      break;
    case KeyEvent.VK_G:
      activeButton = GRAVITY;
      break;
    case KeyEvent.VK_T:
      activeButton = TELEPORT;
      break;
    case KeyEvent.VK_1:
      activeButton = START;
      break;
    case KeyEvent.VK_2:
      activeButton = START2;
      break;
    case KeyEvent.VK_F5:
      activeButton = SAVE;
      break;
    case KeyEvent.VK_F9:
      activeButton = CLOSE;
      break;
    default:
      activeButton = NONE;
    }
    pressedButton = activeButton;
    int result = clickResult(MouseEvent.BUTTON1, screen);
    if (result == NONE) {
      activeButton = oldButton;
    } else {
      pressedButton = NONE;
    }
    return result;
  }


  /**
   * Determines whether a different button or tile is active and returns
   * clipping area (for drawing) is so.
   * @param mouseLocation current mouse position
   * @return the clipping rectangle or <code>null</code> if no screen update
   * is required.
   */
  public Rectangle updateUI(Point mouseLocation) {
    oldButton = activeButton;
    oldTile = activeTile;
    activeButton = NONE;
    activeTile = null;
    clip = null;
    
    /* this happens if editor is launched via keys only and mouse is
     * outside frame
     */
    if (mouseLocation == null) {
      activeButton = NONE;
      return null;
    }

    // determine current position
    if ((activeTile = Geom.gridPos(frames, mouseLocation)) != null) {
      activeButton = GRID;
    } else {
      for (int i = 0; i < buttonLocations.length; i++) {
	if (buttonLocations[i].contains(mouseLocation)) {
	  activeButton = i;
	  break;
	}
      }
    }

    /* pressed key - only ever highlight original mouse down button
     * Note: highlighting of selection rectangle on grid handled by
     * getGridMultiSelect via drawUpdate
     */
    if (pressedButton != NONE && activeButton != pressedButton) {
      activeButton = NONE;
    }  

    // determine clip area
    if (oldButton == activeButton) {
      // GRID -> GRID
      if (activeButton == GRID && !activeTile.equals(oldTile)) {
	clip = Geom.calcTilesClip(frames, activeTile, oldTile);
	// teleport tile construction needs larger clip
	if (teleportTile != null) {
	  /* special case - allow "mouse drag" to behave as is mouse hadn't
	   * been held down at all, only upon TP creation
	   */
	  if (mouseDown == MouseEvent.BUTTON1) {
	    pressedTile = activeTile;
	  }
	  clip.add(Geom.calcTileClip(frames, teleportTile));
	}
	addPressedButtonClip();
	addTeleportDestinationLineClip(activeTile);
	addTeleportDestinationLineClip(oldTile);
      }
    } else {
      if (oldButton == NONE) {
	// () -> GRID
	if (activeButton == GRID) {
	  clip = Geom.calcTileClip(frames, activeTile);
	  addPressedButtonClip();
	  addTeleportDestinationLineClip(activeTile);
	} else {
	  // () -> BUTTON
	  clip = buttonLocations[activeButton];
	}
      } else if (activeButton == NONE) {
	// GRID -> ()
	if (oldButton == GRID) {
	  clip = Geom.calcTileClip(frames, oldTile);
	  addPressedButtonClip();
	  addTeleportDestinationLineClip(oldTile);
	  // have to draw over teleport indicator
	  if (teleportTile != null) {
	    clip.add(Geom.calcTileClip(frames, teleportTile));
	  }
	} else {
	  // BUTTON -> ()
	  clip = buttonLocations[oldButton];
	}
	// BUTTON -> GRID
      } else if (activeButton == GRID) {
	clip = Geom.calcTileClip(frames, activeTile).
	  union(buttonLocations[oldButton]);
	addPressedButtonClip();
	// GRID -> BUTTON
      } else if (oldButton == GRID) {
	// have to draw over teleport indicator
	clip = Geom.calcTileClip(frames, oldTile).
	  union(buttonLocations[activeButton]);
	addPressedButtonClip();
	if (teleportTile != null) {
	  clip.add(Geom.calcTileClip(frames, teleportTile));
	}
      } else {
	// BUTTON -> BUTTON
	clip = buttonLocations[activeButton].
	  union(buttonLocations[oldButton]);
      }
    }
    return clip;
  }


  /**
   * Indicates whether screen should be refreshed.
   * Note: This might say the screen should be refreshed even though a call
   * to <code>updateUI</code> returned nothing.
   * @return <code>true</code> if screen should be redrawn
   */
  public boolean redrawWanted() {
    return forceRedraw;
  }


  // only draw changes since last mouse move
  /**
   * Draw any required changes after mouse movement (call after
   * <code>updateUI</code>
   * @param g graphics context used for rendering
   */
  public void drawUpdate(Graphics2D g) {
    if (oldButton == activeButton) {
      // GRID -> GRID
      if (activeButton == GRID && !activeTile.equals(oldTile)) {
	if (displayTeleportLine || teleportTile != null) {
	  frames.drawAll(true, g);
	  drawGridArea(true, g);
	  displayTeleportLine = false;
	} else {
	  drawGridArea(false, g);
	}
	// teleport construction line
	if (teleportTile != null) {
	  drawTeleportLine(teleportTile, g);
	}
	// teleport destination line
	if (currentLevel.tile(activeTile) != null &&
	    currentLevel.tile(activeTile).getType() == Tile.TELEPORT) {
	  drawTeleportLine(g);
	}
      }
    } else {
      if (oldButton == NONE) {
	// () -> GRID
	if (activeButton == GRID) {
	  drawGridArea(false, g);	  
	  if (displayTeleportLine) {
	    /* don't have to check whether there is a tile as just moved
	     * to grid & displayTeleportLine switch is active
	     */
	    drawTeleportLine(g);
	    displayTeleportLine = false;
	  }
	} else {
	  // () -> BUTTON
	  drawEditorButton(activeButton, true, g);
	}
      } else if (activeButton == NONE) {
	// GRID -> ()
	if (oldButton == GRID) {
	  // cancel teleport tile construction when moved off the grid
	  if (teleportTile != null || displayTeleportLine) {
	    teleportTile = null;
	    displayTeleportLine = false;
	    frames.drawAll(true, g);
	  }
	  drawGridArea(true, g);
	} else {
	  // BUTTON -> ()
	  drawEditorButton(oldButton, false, g);
	}
	// BUTTON -> GRID
      } else if (activeButton == GRID) {
	drawGridArea(false, g);	  
	drawEditorButton(oldButton, false, g);
	// GRID -> BUTTON
      } else if (oldButton == GRID) {
	drawAll(g);
      } else {
	// BUTTON -> BUTTON
	drawEditorButton(oldButton, false, g);
	drawEditorButton(activeButton, true, g);
      }
    }
  }


  /**
   * Draws update after mouse click (call after <code>clickResult</code>)
   * @param g graphics context used for rendering
   */
  public void drawClickUpdate(Graphics2D g) {
    if (activeButton == GRID || activeButton == LEVEL_SWAP ||
	activeButton == LEVEL_FIRST || activeButton == LEVEL_LAST ||
	activeButton == LEVEL_BACK || activeButton == LEVEL_FORWARD ||
	activeButton == LEVEL_CLEAR || activeButton == SAVE) {
      drawGrid(g);
    }
    // re-draw control bar (text & buttons)
    if (activeButton != GRID && activeButton != NONE) {
      drawControlBar(g);
    }
  }


  /**
   * Draws update required after keypress (call after <code>keyUpdate</code>)
   * @param g graphics context used for rendering
   */
  public void drawKeyUpdate(Graphics2D g) {
    Point oldTeleportTile = null;
    if (activeButton == LEVEL_BACK || activeButton == LEVEL_FORWARD ||
	activeButton == LEVEL_FIRST || activeButton == LEVEL_LAST ||
	activeButton == LEVEL_CLEAR || activeButton == LEVEL_SWAP ||
	activeButton == SAVE) {
      oldTile = activeTile;
      oldTeleportTile = teleportTile;
      activeTile = teleportTile = null;
      activeButton = NONE;
      drawGrid(g);
      if (!(activeButton == LEVEL_SWAP || activeButton == LEVEL_CLEAR)) {
	drawControlBar(g);
      }
    } else {
      // clear grid "activity" too if needed
      if (activeTile != null) {
	oldTile = activeTile;
	activeTile = teleportTile = null;
	drawGrid(g);
      }
      activeButton = NONE;
      drawControlBar(g);
    }
    activeTile = oldTile;
    oldTile = null;
    activeButton = oldButton;
    teleportTile = oldTeleportTile;
    pressedButton = oldButton = NONE;
    forceRedraw = true;
  }
  

  /**
   * Draws grid frame and contents
   */
  private void drawGrid(Graphics2D g) {
    g.setClip(raisedGridClip);
    frames.drawGridFrame(true, g);
    drawGridArea(true, g);
    // retain teleport tile destination line
    if (activeTile != null &&
	currentLevel.tile(activeTile) != null &&
	currentLevel.tile(activeTile).getType() == Tile.TELEPORT) {
      drawTeleportLine(g);
    }
  }


  /**
   * Draws internal control frame borders and content
   */
  private void drawControlBar(Graphics2D g) {
    g.setClip(allFieldClipArea);
    Geom.fillRect(frames.getEditorCtrlOffset()[0], 0, 4,
		  frames.getBgColor(), g);
    Geom.fillRect(frames.getEditorCtrlOffset()[1], 4, 4,
		  frames.getBgColor(), g);
    Geom.fillRect(frames.getEditorCtrlOffset()[2], 4, 0,
		  frames.getBgColor(), g);
    drawControlContents(g);
  }

  
  /**
   * Draws editor buttons and status elements
   */
  private void drawControlContents(Graphics2D g) {
    // set name
    g.setFont(txtFont);
    Geom.drawString(shortLevelSetName, levelSetName.getLocation(),
		    levelSetNameOffset, txtColor, g);
    // level number
    if (currentLevelIndex >= 9) {
      Geom.drawImage(imgData.getImage(Images.E_LEVEL_NUMBER_BG),
		     levelNumberBg, frames.getBgColor(), g);
      Geom.drawString(Integer.toString((currentLevelIndex + 1) / 10),
		      levelNumberBg.getLocation(), levelNumber10sOffset,
		      txtColor, g);
      Geom.drawString(Integer.toString((currentLevelIndex + 1) % 10),
		      levelNumberBg.getLocation(), levelNumber1sOffset,
		      txtColor, g);
    } else {
      Geom.drawImage(imgData.getImage(Images.E_LEVEL_NUMBER_BG, true),
		     levelNumberBg, frames.getBgColor(), g);
      Geom.drawString(Integer.toString((currentLevelIndex + 1) % 10),
		      levelNumberBg.getLocation(), levelNumberMiddleOffset,
		      txtColor, g);
    }
    // buttons
    for (int i = 0; i < buttonLocations.length; i++) {
      drawEditorButton(i, (i == activeButton), g);
    }
  }

  
  /**
   * Draws all editor components and grid. (Warning: does not check whether
   * levelset/level has been loaded)
   * @param g graphics context used for rendering
   */
  public void drawAll(Graphics2D g) {
    forceRedraw = false;
    // bg & frames
    frames.drawAll(true, g);
    // grid & buttons
    drawGridArea(true, g);
    // retain teleport tile destination line
    if (activeTile != null &&
	currentLevel.tile(activeTile) != null &&
	currentLevel.tile(activeTile).getType() == Tile.TELEPORT) {
      drawTeleportLine(g);
    }
    drawControlContents(g);
  }

  
  /**
   * Returns levelset for testing (so can be loaded into <code>GamePlayer
   * </code>.
   * @return <code>null</code> if the Play/Test button has not been pressed.
   */
  public LevelSet getTestSet() {
    return testSet;
  }


  /**
   * Indicates whether the cursor should be change (e.g. when moving to/from
   * grid)
   * @return <code>true</code> if the cursor needs to be changed
   */
  public boolean cursorChange() {
    return cursorChange(false);
  }

  
  /**
   * Indicates whether the cursor should be change (e.g. when moving to/from
   * grid)
   * @param exit indicates whether the whole game screen has been left
   * @return <code>true</code> if the cursor needs to be changed
   */  
  public boolean cursorChange(boolean exit) {
    return ((oldButton == GRID && activeButton != GRID) ||
	    (activeButton == GRID && oldButton != GRID) ||
	    (exit && activeButton == GRID));
  }


  /**
   * for teleport which has not been created yet
   */
  private void drawTeleportLine(Point start, Graphics2D g) {
    Geom.drawLineRel(frames, start,
		     (currentLevel.tile(start) == null) ?
		     0 : currentLevel.tile(start).getHeight(),
		     activeTile,
		     (currentLevel.tile(activeTile) == null) ?
		     0 : currentLevel.tile(activeTile).getHeight(),
		     gridHighlight, g);
  }


  /**
   * for existing teleport tile pairs
   */
  private void drawTeleportLine(Graphics2D g) {
    drawTeleportLine(((TptTile)currentLevel.tile(activeTile)).
		     getDestination(),
		     g);
  }


  /**
   * Draws editor grid, tiles (and highlighted coordinates)
   */
  private void drawGridArea(boolean drawGrid, Graphics2D g) {
    if (drawGrid) {
      Geom.drawGrid(frames.getGridArea(), gridSize, frames.getTileSize(),
		    gridColor, g);
    } 
    tPainter.drawTiles(currentLevel.getTiles(), getGridMultiSelect(0),
		       getGridMultiSelect(1), oldTile,
		       currentLevel.getStart2(),
		       currentLevel.hasEqualStarts(), g);
  }

  
  /**
   * returns 1st or 2nd point if multiple tiles have been selected
   * (selection rectangle's 2nd point)
   */
  private Point getGridMultiSelect(int pointIndex) {
    if (pointIndex == 0) {
      if (pressedButton == GRID && activeButton == GRID &&
	  activeTile != null && mouseDown == MouseEvent.BUTTON1 &&
	  (selectedButton == START || selectedButton == START2 ||
	   selectedButton == TELEPORT)) {
	return (activeTile.equals(pressedTile)) ? pressedTile : null;
      } else {
	return activeTile;
      }
    } else {
      if (pressedButton == GRID && activeButton == GRID &&
	  activeTile != null && (mouseDown == MouseEvent.BUTTON3 ||
				 (selectedButton != START &&
				  selectedButton != START2 &&
				  selectedButton != TELEPORT))) {
	return pressedTile;
      } else {
	return null;
      }
    }
  }

  
  /**
   * Adds a tile to the loaded level (after mouse click)
   */
  private int addTile() {
    Tile attributes = generateTile();
    if (attributes != null) {
      if (teleportTile != null) {
	currentLevel.pushTile(teleportTile, attributes, false);
	teleportTile = null;
      } else {
	currentLevel.pushTile(activeTile, attributes,
			      (selectedButton == START2));
      }
      changedSinceLastSave = true;
    } else {
      return NONE;
    }
    return activeButton;
  }


  /**
   * creates tile template based on selected button
   */
  private Tile generateTile() {
    Tile attributes = null;
    try {
      switch(selectedButton) {
      case NORMAL:
	attributes = new NrmTile(1, false);
	break;
      case GRAVITY:
	attributes = new GrvTile(1);
	break;
      case START:
      case START2:
	if (arrowCount() == 0) {
	  attributes = new NrmTile(0, true);
	} else {
	  attributes = new ArwTile(0, true, selectedArrows);
	}
	break;
      case ICE:
	attributes = new IceTile(1);
	break;
      case TELEPORT:
	if (teleportTile == null) {
	  teleportTile = activeTile;
	} else {
	  if (!activeTile.equals(teleportTile)) {
	    attributes = new TptTile(1, activeTile);
	  } else {
	    teleportTile = null;
	    return null;
	  }
	}
	break;
	// arrow tile (without start)
      default:
	attributes = new ArwTile(1, false, selectedArrows);
      }
      // should never happen
    } catch (TileException e) {
      e.printStackTrace();
    }
    return attributes;
  }


  /**
   * Draws a single button in the editor control bar
   * Note: active = mouse over
   */
  private void drawEditorButton(int index, boolean active, Graphics2D g) {
    if (index != NONE) {
      if (frames.getRenderQuality() == ScreenFrames.RENDERING_BALANCED) {
	g.setRenderingHint(RenderingHints.KEY_RENDERING,
			   RenderingHints.VALUE_RENDER_QUALITY);
	g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			   RenderingHints.VALUE_ANTIALIAS_ON);      
      }
      // draw border around selected buttons
      if (index == selectedButton ||
	  (isArrow(index) && selectedArrows[index - ARROW_UP])) {
	g.setColor(buttonBorderColor);
	g.setStroke(buttonBorderStroke);
	g.drawRoundRect(buttonLocations[index].x - (buttonBorderThickness - 1),
			buttonLocations[index].y - (buttonBorderThickness - 1),
			buttonLocations[index].width +
			((buttonBorderThickness - 1) * 2) - 1,
			buttonLocations[index].height +
			((buttonBorderThickness - 1) * 2) - 1,
			8, 8);
      }
      if (frames.getRenderQuality() == ScreenFrames.RENDERING_BALANCED) {
	g.setRenderingHint(RenderingHints.KEY_RENDERING,
			   RenderingHints.VALUE_RENDER_DEFAULT);
	g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			   RenderingHints.VALUE_ANTIALIAS_OFF);      
      }
      // use editor arrow images instead of tile ones
      if (isArrow(index)) {
	Geom.drawImage(imgData.getImage(index - ARROW_UP +
					Images.E_ARROW_UP,
					active),
		       buttonLocations[index], frames.getBgColor(), g);
      } else if (index == START2 && !levelSet.isTwoPlayer()) {
	Geom.drawImage(imgData.getImage(Images.O_DISABLED_TILE),
		       buttonLocations[index], frames.getBgColor(), g);
      } else if (!(readOnly && (index == SAVECLOSE || index == SAVE))) {
	Geom.drawImage(imgData.getImage(index, active),
		       buttonLocations[index], frames.getBgColor(), g);
      }
    }
  }

  
  /**
   * Is the given button index an arrow button
   */
  private boolean isArrow(int buttonIndex) {
    return (buttonIndex >= ARROW_UP && buttonIndex <= ARROW_RIGHT);
  }


  /**
   * Turns off all arrow buttons
   */
  private void resetArrows() {
    for (int i = 0; i < 4; i++) {
      selectedArrows[i] = false;
    }
  }


  /**
   * Number of arrow buttons selected (never more than 3)
   */
  private int arrowCount() {
    int counter = 0;
    for (int i = 0; i < 4; i++) {
      if (selectedArrows[i] == true) {
	counter++;
      }
    }
    return counter;
  }


  /**
   * Adjusts activated arrow buttons based on currently pressed arrow button.
   * Note: Calling this if an arrow button has not been pressed will most
   * possibly result in an IndexOutofBoundsException
   */
  private void setArrows() {
    int arrow = activeButton - ARROW_UP;
    selectedArrows[arrow] = !selectedArrows[arrow];
    if (selectedButton != START && selectedButton != START2) {
      selectedButton = NONE;
    }
    if (arrowCount() == 0 && selectedButton != START &&
	selectedButton != START2) {
      selectedButton = NORMAL;
    } else if (arrowCount() == 4) {
      // cannot have more than 3 active so only keep current one on
      resetArrows();
      selectedArrows[arrow] = true;
    }
  }


  /**
   * Adds the teleport destination indicator line to clipping area
   */
  private void addTeleportDestinationLineClip(Point p) {
    if (currentLevel.tile(p) != null &&
	currentLevel.tile(p).getType() == Tile.TELEPORT) {
      clip.add(Geom.calcTileClip(frames,
				 ((TptTile) currentLevel.tile(p)).
				 getDestination()));
      displayTeleportLine = true;
    }
  }


  /**
   * Adds the currently pressed (where mouse-down event occured) tile to
   * clipping area
   */
  private void addPressedButtonClip() {
    if (pressedTile != null) {
      clip.add(Geom.calcTileClip(frames, pressedTile));
    }
  }

 
  /**
   * Produces dialogs to user for start/end message input and stores them if
   * required.
   */
  private void setLevelMessages(Component c) {
    // INTRO MESSAGE
    optionPane.setMessage("Level start message:");
    optionPane.setInitialSelectionValue(currentLevel.getInText());
    optionPane.setIcon(imgData.getIcon(Images.E_LEVEL_MESSAGES));
    optionPane.createDialog(c, "Tile Pile - Editor").setVisible(true);
    if (optionPane.getValue() != null &&
	((Integer) optionPane.getValue()).equals(JOptionPane.OK_OPTION)) {
      String input = (String) optionPane.getInputValue();
      currentLevel.setInText(SplashMessage.isValidMessage(input) ?
			     input : null);
      changedSinceLastSave = true;
    } else {
      return;
    }
    // OUTRO MESSAGE
    optionPane.setMessage("Level end message:");
    optionPane.setInitialSelectionValue(currentLevel.getOutText());
    optionPane.createDialog(c, "Tile Pile - Editor").setVisible(true);
    if (optionPane.getValue() != null &&
	((Integer) optionPane.getValue()).equals(JOptionPane.OK_OPTION)) {
      String input = (String) optionPane.getInputValue();
      currentLevel.setOutText(SplashMessage.isValidMessage(input) ?
			      input : null);
      changedSinceLastSave = true;
    }
  }


  //  private boolean isValidMessage(

  /**
   * Produces dialog to user for level time limit input and stores it if
   * required.
   */
  private void setLevelTimeLimit(Component c) {
    optionPane.setMessage("Level time limit:");
    optionPane.setInitialSelectionValue(currentLevel.getTimeLimit());
    optionPane.setIcon(imgData.getIcon(Images.E_LEVEL_TIME));
    optionPane.createDialog(c, "Tile Pile - Editor").setVisible(true);
    if (optionPane.getValue() != null &&
	((Integer) optionPane.getValue()).equals(JOptionPane.OK_OPTION)) {
      if (((String) optionPane.getInputValue()).length() == 0) {
	currentLevel.setTimeLimit(0);
      } else {
	try {
	  currentLevel.setTimeLimit(Integer.parseInt((String) optionPane.
						     getInputValue()));
	  /* cannot just catch NumberFormatException because Level class
	   * can also throw if the valid Integer is negative
	   */
	  changedSinceLastSave = true;
	} catch (Exception e) {
	  JOptionPane.showMessageDialog(c, "Invalid level time limit." +
					"\n(For no time limit leave blank.)",
					"Tile Pile - Editor",
					JOptionPane.ERROR_MESSAGE);
	}
      }
    }
    optionPane.setMessage("");
  }


  /**
   * calculate all drawing element locations
   *
   *  gridFrame      ctrlFrame
   *  _____________  _____
   *  | _________  ||     |
   *  | |gridArea| ||     |
   *  | |        | ||     |
   *  | ---------- ||     |
   *  -------------  -----
   */
  private void calculateGeometry(Graphics2D g) {
    int
      tileSize = frames.getTileSize(),
      borderSpacing = frames.getBorderSpacing(),
      mediumButtonSpacing = (borderSpacing / 2),
      smallButtonSpacing = 8;
    Rectangle ctrlFrame = frames.getCtrlOffset();
    Rectangle gridArea = frames.getGridArea();

    /* BUTTON POSITIONING (-4 for arrows which are in both image sets and
     * -1 for level number background
     */
    buttonLocations = new Rectangle[imgData.getTileImageCount() +
				    imgData.getEditorImageCount() - 5];

    // make sure level set name fits on screen, otherwise shorten
    shortLevelSetName =
      Geom.getMaxString(levelSet.getName(), "...", ctrlFrame.width -
			(borderSpacing + frames.getFrameThickness()),
			txtFont, g);

    // FIRST ROW
    Rectangle tBounds = Geom.calcTextBounds(shortLevelSetName, txtFont, g);
    levelSetNameOffset = tBounds.getLocation();
    levelSetName = new Rectangle(ctrlFrame.x +
				 ((ctrlFrame.width - tBounds.width) / 2),
				 ctrlFrame.y + smallButtonSpacing,
				 tBounds.width, tBounds.height);

    // SECOND ROW
    BufferedImage bI = imgData.getImage(LEVEL_FIRST);
    buttonLocations[LEVEL_FIRST] =
      new Rectangle(ctrlFrame.x + (borderSpacing / 2),
		    levelSetName.y + levelSetName.height +
		    (smallButtonSpacing * 2),
		    bI.getWidth(), bI.getHeight());

    bI = imgData.getImage(LEVEL_BACK);
    buttonLocations[LEVEL_BACK] =
      new Rectangle(buttonLocations[LEVEL_FIRST].x +
		    buttonLocations[LEVEL_FIRST].width + smallButtonSpacing,
		    buttonLocations[LEVEL_FIRST].y,
		    bI.getWidth(), bI.getHeight());

    bI = imgData.getImage(Images.E_LEVEL_NUMBER_BG);
    tBounds = Geom.calcTextBounds("1", txtFont, g);
    levelNumberBg =
      new Rectangle(ctrlFrame.x + ((ctrlFrame.width - bI.getWidth())/ 2),
		    buttonLocations[LEVEL_BACK].y,
		    bI.getWidth(), bI.getHeight());
    levelNumber10sOffset = new Point(tBounds.getLocation());
    levelNumber10sOffset.translate(-15, -8);
    levelNumber1sOffset = new Point(tBounds.getLocation());
    levelNumber1sOffset.translate(-47, -8);
    levelNumberMiddleOffset = new Point(tBounds.getLocation());
    levelNumberMiddleOffset.translate(-30, -8);

    bI = imgData.getImage(LEVEL_LAST);
    buttonLocations[LEVEL_LAST] =
      new Rectangle(ctrlFrame.x + ctrlFrame.width -
		    ((borderSpacing / 2) + bI.getWidth()),
		    buttonLocations[LEVEL_FIRST].y,
		    bI.getWidth(), bI.getHeight());

    bI = imgData.getImage(LEVEL_FORWARD);
    buttonLocations[LEVEL_FORWARD] =
      new Rectangle(buttonLocations[LEVEL_LAST].x -
		    (buttonLocations[LEVEL_LAST].width +
		     smallButtonSpacing),
		    buttonLocations[LEVEL_FIRST].y,
		    bI.getWidth(), bI.getHeight());

    // THIRD ROW
    buttonLocations[LEVEL_SWAP] =
      new Square(ctrlFrame.x + borderSpacing,
		 buttonLocations[LEVEL_FIRST].y +
		 buttonLocations[LEVEL_FIRST].height +
		 (smallButtonSpacing * 2),
		 tileSize);

    buttonLocations[LEVEL_MESSAGES] =
      new Square(buttonLocations[LEVEL_SWAP].x + tileSize + buttonSpacing,
		 buttonLocations[LEVEL_SWAP].y,
		 tileSize);

    buttonLocations[LEVEL_TIME] =
      new Square(buttonLocations[LEVEL_MESSAGES].x + tileSize + buttonSpacing,
		 buttonLocations[LEVEL_MESSAGES].y,
		 tileSize);


    // FOURTH ROW
    buttonLocations[NORMAL] =
      new Square(buttonLocations[LEVEL_SWAP].x,
		 buttonLocations[LEVEL_SWAP].y + tileSize + buttonSpacing + 17,
		 tileSize);

    buttonLocations[ICE] =
      new Square(buttonLocations[NORMAL].x + tileSize + buttonSpacing,
		 buttonLocations[NORMAL].y,
		 tileSize);

    buttonLocations[GRAVITY] =
      new Square(buttonLocations[ICE].x + tileSize + buttonSpacing,
		 buttonLocations[NORMAL].y,
		 tileSize);

    // FIFTH ROW
    buttonLocations[TELEPORT] =
      new Square(buttonLocations[NORMAL].x,
		 buttonLocations[NORMAL].y + tileSize + mediumButtonSpacing,
		 tileSize);

    buttonLocations[START] =
      new Square(buttonLocations[TELEPORT].x + tileSize + buttonSpacing,
		 buttonLocations[TELEPORT].y,
		 tileSize);

    buttonLocations[START2] =
      new Square(buttonLocations[START].x + tileSize + buttonSpacing,
		 buttonLocations[START].y,
		 tileSize);


    // special layout for arrow buttons
    buttonLocations[ARROW_UP] =
      new Square(buttonLocations[START].x,
		 buttonLocations[START].y + tileSize + mediumButtonSpacing,
		 tileSize);

    buttonLocations[ARROW_LEFT] =
      new Square(buttonLocations[ARROW_UP].x -
		 (tileSize + buttonBorderThickness + 1),
		 buttonLocations[ARROW_UP].y + tileSize +
		 buttonBorderThickness + 1,
		 tileSize);

    buttonLocations[ARROW_RIGHT] =
      new Square(buttonLocations[ARROW_UP].x +
		 tileSize + buttonBorderThickness + 1,
		 buttonLocations[ARROW_LEFT].y,
		 tileSize);

    buttonLocations[ARROW_DOWN] =
      new Square(buttonLocations[ARROW_UP].x,
		 buttonLocations[ARROW_LEFT].y + tileSize +
		 buttonBorderThickness + 1,
		 tileSize);

    // clear & close buttons
    bI = imgData.getImage(Images.E_SAVECLOSE);
    buttonLocations[SAVECLOSE] =
      new Rectangle(ctrlBorderSpacing(bI.getWidth()),
		    ctrlFrame.y + ctrlFrame.height -
		    ((buttonSpacing / 2) + bI.getHeight()),
		    bI.getWidth(), bI.getHeight());


    bI = imgData.getImage(Images.E_CLOSE);
    buttonLocations[CLOSE] = !readOnly ?
      new Rectangle(ctrlBorderSpacing(bI.getWidth()),
		    buttonLocations[SAVECLOSE].y -
		    (bI.getHeight() + smallButtonSpacing),
		    bI.getWidth(), bI.getHeight()) :
      // read only = put this button in place of save&close button
      new Rectangle(ctrlBorderSpacing(bI.getWidth()),
		    ctrlFrame.y + ctrlFrame.height -
		    ((buttonSpacing / 2) + bI.getHeight()),
		    bI.getWidth(), bI.getHeight());

    bI = imgData.getImage(Images.E_SAVE);
    buttonLocations[SAVE] = !readOnly ?
      new Rectangle(ctrlBorderSpacing(bI.getWidth()),
		    buttonLocations[CLOSE].y -
		    (bI.getHeight() + smallButtonSpacing),
		    bI.getWidth(), bI.getHeight()) :
      buttonLocations[SAVECLOSE];
    
    bI = imgData.getImage(Images.E_PLAY);
    buttonLocations[LEVEL_PLAY] =
      new Rectangle(ctrlBorderSpacing(bI.getWidth()),
		    buttonLocations[readOnly ? CLOSE : SAVE].y -
		    (smallButtonSpacing + bI.getHeight()),
		    bI.getWidth(), bI.getHeight());

    bI = imgData.getImage(Images.E_CLEAR);
    buttonLocations[LEVEL_CLEAR] =
      new Rectangle(ctrlBorderSpacing(bI.getWidth()),
		    buttonLocations[LEVEL_PLAY].y -
		    (smallButtonSpacing + bI.getHeight()),
		    bI.getWidth(), bI.getHeight());

    

    int maxSlant = tileHeightLimit * frames.getHeightPerLevel();
    raisedGridClip =
      new Rectangle (gridArea.x, gridArea.y - maxSlant,
		     gridArea.width + maxSlant, gridArea.height + maxSlant);

    // set clipping area for all non-grid components (apart from borders)
    allFieldClipArea = new Rectangle(levelSetName);
    allFieldClipArea.add(levelNumberBg);
    for (int i = 0; i < buttonLocations.length; i++) {
      allFieldClipArea.add(buttonLocations[i]);
    }
    // extend button clip to include button borders
    allFieldClipArea.x -= (buttonBorderThickness);
    allFieldClipArea.y -= (buttonBorderThickness);
    allFieldClipArea.width += (buttonBorderThickness * 2);
    allFieldClipArea.height += (buttonBorderThickness * 2);

    // set unused locations to value which are impossible for user to reach
    if (readOnly) {
      buttonLocations[SAVECLOSE] = buttonLocations[SAVE] =
	new Rectangle(-1, -1, 0, 0);
    }
  }


  private int ctrlBorderSpacing(int buttonWidth) {
    return ((frames.getCtrlOffset().width - buttonWidth) / 2) +
      frames.getCtrlOffset().x;
  }

}