/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.event.*;
import javax.swing.JOptionPane;
import java.io.*;
import java.net.URL;
import java.util.Random;


/**
 * Defines function and rendering of the gameplay component.
 * @author V.Termanis
 */
public class GamePlayer {
  
  // player modes
  /**
   * Single player mode
   */
  public static final int SINGLE = 0;
  /**
   * Cooperative two player mode
   */
  public static final int COOP = 1;
  /**
   * Versus two player mode
   */
  public static final int VS = 2;
  /**
   * Versus alternate two player mode
   */
  public static final int VS_TURN_BASED = 3;
  /**
   * Versus mode with choice between simulatenous and alternate modes
   */
  public static final int VS_CHOOSE = 4;

  // levelset load results
  /**
   * <code>loadLevelSet</code> method return value
   */
  public static final int LOAD_OK = 0;
  /**
   * <code>loadLevelSet</code> method return value
   */
  public static final int LOAD_NOT2P = 1;
  /**
   * <code>loadLevelSet</code> method return value
   */
  public static final int LOAD_CANCELLED = 2;

  // movement directions
  public static final int NONE = -1;

  public static final int UP = 0;
  public static final int DOWN = 1;
  public static final int LEFT = 2;
  public static final int RIGHT = 3;

  // only used to start automatic levels
  public static final int AUTOSTART = 4;
  
  public static final int MOVED = 5;
  public static final int GAMEOVER = 6;
  public static final int FINISHED_SET = 7;
  public static final int FINISHED_LEVEL = 8;
  public static final int INVALID_MOVE = 9;
  // also indicates level failure with infinite lives
  public static final int LIFE_LOST = 10;
  public static final int RESET = 11;
  public static final int QUIT = 12;
  // dev version only flag to indicate level has been skipped
  public static final int SKIPPED_LEVEL = 13;

  // level (from 0) on which to show MediaMerge tile instead of normal ones
  public static final int MEDIAMERGE_LEVEL = -5; // (was 28 in v0.986)
  // level (from 0) on which to show numbered tiles
  public static final int NUMBERED_LEVEL = 57;
  // tutorial level with custom death message
  public static final int CUSTOM_DEATH_MESSAGE_LEVEL = 10;

  /**
   * ID of first player
   */
  public static final int P1 = 0;
  /**
   * ID of second player
   */
  public static final int P2 = 1;
  // constant below is only used internally
  private static final int BOTH = 2;

  // level end messages
  private static final String[] levelEndComments =
  {"Good", "Nice one", "OK", "Level finished", "Well done", "Lovely", "Superb",
   "Excellent", "You're clever", "Top skills", "That's it"};


  private Images imgData;
  private Color txtColor;
  private Font txtFont;
  private TilePainter tPainter;
  private SplashMessage message;
  private ScreenFrames frames;
  // locations of various individual elements which are rendered
  private Rectangle setName, levelNumber, tileText, tileTextNum, livesLeft,
    livesLeftNum, timeLeft, timeLeft2, timeLeftImg, timeLeft2Img, p1Tiles,
    p1TileImg, p2Tiles, p2TileImg, closeButton, scoreString, scoreStringNum;
  private Point setNameOffset,levelNumberOffset, tileTextOffset,
    tileTextNumOffset, livesLeftOffset, livesLeftNumOffset, timeLeftOffset,
    timeLeft2Offset, p1TilesOffset, p2TilesOffset, scoreStringOffset,
    scoreStringNumOffset;
  private boolean closeButtonActive, closeButtonPressed, scoringEnabled,
    unlockable, suddenDeath, vs_alt_round2, mediaMergeRequired, testMode,
    drawBothTimers;
  private LevelSet levelSet;
  private String shortLevelSetName;
  private int stringSpacing, numSpacing, lives, maxLives, currentLevelIndex,
    updateState, lastDirectionP1, lastDirectionP2, gameMode, originalGameMode,
    p1Wins, p2Wins, score;
  private LevelPlayer lp;
  private GameTimer gameTimer;
  // used to show different level end messages
  private Random random;
  // do not show same message twice
  private int lastMessage;

  /**
   * Creates a new <code>GamePlayer</code> instance with the given attributes
   * @param imgData global image resources
   * @param frames for drawing borders and background
   * @param tPainter for drawing tiles
   * @param txtColor colour of all text (except tile numbers)
   * @param txtFont style used for all text (except tile numbers)
   * @param stringSpacing vertical spacing (in pixels) between text elements
   * @param numSpacing horizontal spacing (in pixels from left) for status
   * numbers (e.g. lives left)
   * @param g graphics context for drawing
   * @param lvlSetFilename levelset to initially load (in single player mode),
   * can also be set to <code>null</code> if a levelset is to be loaded later
   * @param beepSoundName full path to countdown "beep" sound
   * @param message popup message handler
   * @param screen parent component
   */
  public GamePlayer(Images imgData, ScreenFrames frames,
		    TilePainter tPainter, Color txtColor, Font txtFont,
		    int stringSpacing, int numSpacing, Graphics2D g,
		    String lvlSetFilename, String beepSoundName,
		    SplashMessage message, Screen screen)
    throws NullPointerException, IllegalArgumentException, IOException {
    if (imgData == null || frames == null || tPainter == null ||
	txtColor == null || txtFont == null || message == null) {
      throw new NullPointerException("One of the parameters is null");
    }
    if (stringSpacing < 0) {
      throw new IllegalArgumentException("String spacing is less than zero");
    }
    this.imgData = imgData;
    this.frames = frames;
    this.tPainter = tPainter;
    this.txtColor = txtColor;
    this.txtFont = txtFont;
    this.stringSpacing = stringSpacing;
    this.numSpacing = numSpacing;
    this.message = message;
    // start off with infinte number of lives
    maxLives = -1;
    if (lvlSetFilename != null) {
      loadLevelSet(lvlSetFilename, g);
    }
    originalGameMode = gameMode = SINGLE;
    gameTimer = new GameTimer(beepSoundName, screen, this);
  }


  /**
   * Loads a levelset from a file.
   * @param filename absolute path to levelset file
   * @param g graphics context for drawing
   */
  public void loadLevelSet(String filename, Graphics2D g) throws IOException {
    loadLevelSet((new File(filename)).toURI().toURL(), SINGLE, null, g);
  }


  /**
   * Loads a levelset from a file with the specified game mode.
   * @param filename absolute path to levelset file
   * @param gameMode e.g. <code>SINGLE</code>
   * @param c parent component for drawing dialogs
   * @param g graphics context for drawing
   * @return e.g. <code>LOAD_OK</code>
   */
  public int loadLevelSet(String filename, int gameMode, Component c,
			  Graphics2D g)
    throws IOException, IllegalArgumentException {
    return loadLevelSet((new File(filename)).toURI().toURL(), gameMode, c, g);
  }

 
  /**
   * Loads a levelset from the given location
   * @param fileURL absolute location of levelset file
   * @param g graphics context for drawing
   * @return e.g. <code>LOAD_OK</code>
   */
  public int loadLevelSet(URL fileURL, Graphics2D g) throws IOException {
    return loadLevelSet(fileURL, SINGLE, null, g);
  }


  /**
   * Loads a levelset from a location with the specified game mode.
   * @param fileURL absolute path to levelset file
   * @param gameMode e.g. <code>SINGLE</code>
   * @param c parent component for drawing dialogs
   * @param g graphics context for drawing
   * @return e.g. <code>LOAD_OK</code>
   */
  public int loadLevelSet(URL fileURL, int gameMode, Component c, Graphics2D g)
    throws IOException, IllegalArgumentException, NullPointerException {
    if (gameMode < 0 || gameMode > 4) {
      throw new IllegalArgumentException("Invalid gameMode supplied.");
    }
    try {
      levelSet = LevelSet.load(fileURL);
    } catch (ClassNotFoundException e) {
      throw new IOException("Invalid levelset file.");
    }

    // can the levelset be played in two player mode
    if (gameMode != SINGLE && !levelSet.isTwoPlayer()) {
      return LOAD_NOT2P;
    }

    // display a dialog for user to decide which VS mode they want
    if (gameMode == VS_CHOOSE) {
      if (c == null) {
	throw new NullPointerException("No component specified so cannot " +
				       "create dialog.");
      }
      gameMode = vsModeChooserResult(c);
      if (gameMode == -1) {
	return LOAD_CANCELLED;
      }
    }

    setLevelSet(levelSet, gameMode, false, g);
    return LOAD_OK;
  }


  /**
   * Returns the active levelset
   * @return the loaded levelset
   */
  protected LevelSet getLevelSet() {
    return levelSet;
  }


  /**
   * Indicate that this levelset's levels have to be unlocked
   * Note: Unlock status is reset each time a new levelset is loaded
   */
  public void enableUnlocking() {
    unlockable = true;
  }


  /**
   * Indicates whether scoring is turned on
   * Note: Scoring status is reset each time a new levelset is loaded
   * @return <code>true</code> is scoring system it turned on
   */
  public boolean scoringEnabled() {
    return scoringEnabled;
  }


  /**
   * Returns the currently set game mode
   * @return e.g. <code>SINGLE</code>
   */
  public int gameMode() {
    return originalGameMode;
  }


  /**
   * Returns total score up to now
   * Note: Scoring status is reset each time a new levelset is loaded
   * @return the current score. Note: If scoring is off this will return
   * <code>0</code>
   */
  public int getScore() {
    return score;
  }


  /**
   * Choose which level to play
   * @param index the level number
   */
  public void setLevel(int index) {
    if (index >= levelSet.getSize()) {
      currentLevelIndex = levelSet.getSize() - 1;
    } else if (index > 0) {
      currentLevelIndex = index;
    }
    setLevelStartMessage();
    setLevel();
  }


  // score - indicates whether to calculate & display score or not
  /**
   * Loads a levelset straight from its native format.
   * @param ls the levelset to load
   * @param gameMode e.g. <code>SINGLE</code>
   * @param score whether to enable scoring or not
   * @param g graphics context to use for drawing
   */
  public void setLevelSet(LevelSet ls, int gameMode, boolean score,
			  Graphics2D g)
    throws NullPointerException, IllegalArgumentException {
    if (ls == null) {
      throw new NullPointerException("ls is null");
    }
    if (gameMode < 0 || gameMode > 3) {
      throw new IllegalArgumentException("Invalid gameMode supplied.");
    }
    mediaMergeRequired = vs_alt_round2 = testMode = suddenDeath = false;
    this.score = 0;
    originalGameMode = this.gameMode = gameMode;
    originalGameMode = this.gameMode = gameMode;
    currentLevelIndex = p1Wins = p2Wins = 0;
    scoringEnabled = score;

    gameTimer.reset();
    gameTimer.resetAccumulatedTime();

    levelSet = ls;
    setLevel();
    setLevelStartMessage();
    lives = maxLives = -1;
    unlockable = false;
    calculateGeometry(g);
  }
  

  /**
   * tells levelplayer to perform move, changes level or resets if necessary
   */
  private int move(int direction) throws IllegalArgumentException {
    int result;
    boolean ranOutOfTime = false;

    /* as this method is not called from outside this class this should
     * never happen
     */
    if ((direction < UP || direction > RIGHT) && direction != AUTOSTART) {
      throw new IllegalArgumentException("direction is invalid");
    }

    // store direction player last moved in (required for moving e.g. over ice)
    if (lp.secondPlayerActive()) {
      lastDirectionP2 = direction;
    } else {
      lastDirectionP1 = direction;
    }

    // time limit handling
    if (lp.hasNotStarted() && !gameTimer.hasStarted()) {
      gameTimer.start();
    }
    if (!gameTimer.isActive() || gameTimer.timeLeft() > 0) {
      result = lp.moveOne(direction);
    } else {
      result = lp.setActivePlayerDead();
      ranOutOfTime = (result == LevelPlayer.LOST);
    }

    switch(result) {
    case LevelPlayer.MOVED:
      // swap players if necessary (but not when autostarting level)
      if (gameMode == VS_TURN_BASED && direction != AUTOSTART) {
	lp.setActivePlayerTo(!lp.secondPlayerActive());
      }
      return MOVED;
    case LevelPlayer.STUCK:
      // make sure mode not set to turn based anymore
      if (gameMode == VS_TURN_BASED) {
	/* switch timer here if necessary because in VS mode won't be able to
	 * switch again
	 */
	if (lp.activePlayerDead()) {
	  stopAndToggleTimer();
	  unpauseTimer();
	}
	gameMode = VS;
	drawBothTimers = true;
      }
      return MOVED;
    case LevelPlayer.FINISHED:
      addScore();
      gameTimer.stopAll();
      // use custom end message, if exists (only COOP & single player)
      String msg = (gameMode == SINGLE || gameMode == COOP) ?
	levelSet.getLevel(currentLevelIndex).getOutText() : null;
      /* don't want to display empty end-messages - these could have been
       * introduced in levels made prior to v0.972
       */
      if (msg != null && msg.length() == 0) {
	msg = null;
      }
      // END OF WHOLE SET
      if (currentLevelIndex == levelSet.getSize() - 1) {
	// endlessly repeat set in test mode
	if (testMode) {
	  message.request("Completed", SplashMessage.DEFAULT);
	  return (updateState = FINISHED_LEVEL);
	}
	if (gameMode == SINGLE || gameMode == COOP) {
	  if (scoringEnabled) {
	    gameTimer.incrementAccumulatedTime();
	    score = ((int) gameTimer.getAccumulatedTimeSeconds() *
		     (gameMode == SINGLE ? 1 : 2)) +
	      ((lives == -1) ? 0 : (lives * 50));
	    message.request("Total score: " + score, SplashMessage.LONG);
	  } else {
	    if (msg == null) {
	      message.request("Congratulations!", SplashMessage.LONG);
	    } else {
	      message.request(msg);
	    }
	  }
	} else {
	  message.request(p1Wins, p2Wins, SplashMessage.DEFAULT);
	  if (originalGameMode == VS_TURN_BASED && !vs_alt_round2) {
	    vs_alt_round2 = true;
	    return (updateState = FINISHED_LEVEL);
	    // sudden death - keep playing until have a winner
	  } else if (gameMode == VS && originalGameMode == VS &&
		     p1Wins == p2Wins) {
	    if (!suddenDeath) {
	      message.request("Sudden death!", SplashMessage.LONG);
	      suddenDeath = true;
	    }
	    return (updateState = FINISHED_LEVEL);
	  } else {
	    message.setDisplayLength(SplashMessage.LONG);
	  }
	}
	return (updateState = FINISHED_SET);
	// END OF LEVEL
      } else {
	if (gameMode == SINGLE || gameMode == COOP) {
	  if (unlockable) {
	    if (ProgressHandler.setReachedLevel(currentLevelIndex + 1,
						(gameMode == COOP))
		== ProgressHandler.UNLOCKED) {
	      msg = ProgressHandler.getUnlockMessage(currentLevelIndex + 1);
	    }
	  }
	  if (msg == null) {
	    message.request(getLevelEndMessage(), SplashMessage.SHORT);
	  } else {
	    message.request(msg);
	  }
	} else {
	  message.request(p1Wins, p2Wins, SplashMessage.DEFAULT);
	  // end game as soon as one player is ahead
	  if (suddenDeath && p1Wins != p2Wins) {
	    message.setDisplayLength(SplashMessage.LONG);
	    return (updateState = FINISHED_SET);
	  }
	}
	if (scoringEnabled) {
	  gameTimer.incrementAccumulatedTime();
	  // in two player challenge mode seconds left give twice the score
	  score = (int) gameTimer.getAccumulatedTimeSeconds() *
	    (gameMode == SINGLE ? 1 : 2);
	}
	return (updateState = FINISHED_LEVEL);
      }
    case LevelPlayer.INVALID_MOVE:
      return INVALID_MOVE;
    case LevelPlayer.LOST:
      gameTimer.stopAll();
      if (lives > 0) {
	lives--;
	if (lives == 0) {
	  if (ranOutOfTime) {
	    message.request("Game Over (Out of time)", SplashMessage.DEFAULT);
	  } else {
	    message.request("Game Over", SplashMessage.DEFAULT);
	  }
	  return (updateState = GAMEOVER);
	} else {
	  if (ranOutOfTime) {
	    message.request("Life Lost (Out of time)", SplashMessage.DEFAULT);
	  } else {
	    message.request("Life Lost", SplashMessage.DEFAULT);
	  }
	  return (updateState = LIFE_LOST);
	}
	// keep repeating level in test mode
      } else if (testMode) {
	message.request("Out of moves" +
			(ranOutOfTime ? " (Out of time)" : ""),
			SplashMessage.DEFAULT);
	return (updateState = FINISHED_LEVEL);
	// infinite number of lives
      } else if (gameMode == SINGLE || gameMode == COOP) {
	// only required for tutorial level
	if (mediaMergeRequired &&
	    currentLevelIndex == CUSTOM_DEATH_MESSAGE_LEVEL &&
	    lp.teleportDeath == true) {
	  message.request("You teleported here   " +
			  "\u00BB\u00BB\u00BB\u00BB\u00BB\u00BB\u00BB\u00BB");
	} else {
	  message.request("Try again" + (ranOutOfTime ? " (Out of time)" : ""),
			  SplashMessage.DEFAULT);
	}
	return (updateState = LIFE_LOST);
	// go to next level - "life loss" counts as point for other player
      } else {
	lp.setActivePlayerTo(lp.secondPlayerActive());
	addScore(); // update VS score
	message.request(p1Wins, p2Wins, SplashMessage.DEFAULT);
	if (currentLevelIndex == levelSet.getSize() - 1) {
	  if (originalGameMode == VS_TURN_BASED && !vs_alt_round2) {
	    vs_alt_round2 = true;
	    return (updateState = FINISHED_LEVEL);
	    // sudden death - keep playing until have a winner
	  } else if (gameMode == VS && originalGameMode == VS &&
		     p1Wins == p2Wins) {
	    if (!suddenDeath) {
	      message.request("Sudden death!", SplashMessage.LONG);
	      suddenDeath = true;
	    }
	    return (updateState = FINISHED_LEVEL);
	  } else {
	    message.setDisplayLength(SplashMessage.LONG);
	    return (updateState = FINISHED_SET);
	  }
	} else {
	  if (suddenDeath & p1Wins != p2Wins) {
	    message.setDisplayLength(SplashMessage.LONG);
	    return (updateState = FINISHED_SET);
	  } else {
	    return (updateState = FINISHED_LEVEL);
	  }
	}
      }
      // should never happen
    default:
      return INVALID_MOVE;
    }
  }


  /**
   * Moves one player by one (or performs automatic move)
   */
  private int move(int direction, int playerIndex, boolean auto) {
    int result = INVALID_MOVE;
    if ((playerIndex == P1 && lp.secondPlayerActive()) ||
	(playerIndex == P2 && !lp.secondPlayerActive())) {
      if (!(gameMode == VS_TURN_BASED && !auto && direction != AUTOSTART)) {
	stopAndToggleTimer();
	lp.setActivePlayerTo(playerIndex == P2);
	unpauseTimer();
	result = move(direction);
	stopAndToggleTimer();
	lp.setActivePlayerTo(!(playerIndex == P2));
	unpauseTimer();
      }
    } else {
      result = move(direction);
    }
    return result;
  }


  /**
   * Moves one or both players (or executes control command, e.g. F9 key)
   */
  public int performMoves(int keyP1, int keyP2) {
    int resultP1 = INVALID_MOVE, resultP2 = INVALID_MOVE;

    switch(keyP1) {
      // *** CONTROL KEYS ***
      // quit
    case KeyEvent.VK_F9:
      gameTimer.stopAll();
      return QUIT;
      /* Level skip (disabled unless in development mode - see constructor
       * in keyHandler class)
       */
    case KeyEvent.VK_END:
      if (currentLevelIndex < levelSet.getSize() -1) {
	gameTimer.stopAll();
	updateState = FINISHED_LEVEL;
	updateState(false);
	return SKIPPED_LEVEL;
      } else {
	return INVALID_MOVE;
      }
      // restart level
    case KeyEvent.VK_ESCAPE:
      if (gameMode == SINGLE || gameMode == COOP || testMode == true) {
	if (lives > 1 || lives == -1) {
	  if (lives > 1) {
	    lives--;
	  }
	  lp.reset();
	  gameTimer.reset();
	  return RESET;
	} else {
	  message.request("Game Over", SplashMessage.LONG);
	  gameTimer.stopAll();
	  return (updateState = GAMEOVER);
	}
      } else {
	return INVALID_MOVE;
      }
    default:
      // *** PLAYER 1 KEYS ***
      if (keyP1 != AUTOSTART && automaticMovesLeft(false)) {
	resultP1 = move(lastDirectionP1, P1, true);
      } else if (keyP1 != PlayerKeys.NONE) {
	resultP1 = move(keyCodeToP1Direction(keyP1), P1, false);
      }
    }

    // *** PLAYER 2 KEYS ***
    // no need to continue life lost / end of level / gameover
    if (gameMode != SINGLE &&
	(resultP1 == MOVED || resultP1 == INVALID_MOVE)) {
      if (keyP1 != AUTOSTART && automaticMovesLeft(true)) {
	resultP2 = move(lastDirectionP2, P2, true);
      } else if (keyP2 != PlayerKeys.NONE) {
	resultP2 = move(keyCodeToP2Direction(keyP2), P2, false);
      }
    }

    return (resultP2 != INVALID_MOVE) ? resultP2 : resultP1;
  }


  /**
   * Maps <code>KeyEvent</code> keycode to <code>GamePlayer</code> direction
   */
  private static int keyCodeToP1Direction(int code) {
    switch(code) {
    case KeyEvent.VK_UP:
      return UP;
    case KeyEvent.VK_DOWN:
      return DOWN;
    case KeyEvent.VK_LEFT:
      return LEFT;
    case KeyEvent.VK_RIGHT:
      return RIGHT;
    case AUTOSTART:
      return AUTOSTART;
    }
    // should never happen
    throw new IllegalArgumentException("Invalid P1 direction code: " + code);
  }


  /**
   * Maps <code>KeyEvent</code> keycode to <code>GamePlayer</code> direction
   */
  private static int keyCodeToP2Direction(int code) {
    switch(code) {
    case KeyEvent.VK_W:
      return UP;
    case KeyEvent.VK_S:
      return DOWN;
    case KeyEvent.VK_A:
      return LEFT;
    case KeyEvent.VK_D:
      return RIGHT;
    case AUTOSTART:
      return AUTOSTART;
    }
    // should never happen
    throw new IllegalArgumentException("Invalid P2 direction code: " + code);
  }


  /**
   * Returns true if there are automatic moves left for a player (e.g. gravity)
   */
  private boolean automaticMovesLeft(boolean playerTwo) {
    if (lp.secondPlayerActive() != playerTwo) {
      lp.setActivePlayerTo(playerTwo);
      boolean result = lp.automaticMovesLeft(playerTwo ? lastDirectionP2 :
					     lastDirectionP1);
      lp.setActivePlayerTo(!playerTwo);
      return result;
    } else {
      return lp.automaticMovesLeft(playerTwo ? lastDirectionP2 :
				   lastDirectionP1);
    }
  }


  /**
   * Returns true if there are any automatic moves left
   */
  public boolean automaticMovesLeft() {
    if (lp == null) {
      return false;
    }
    if (gameTimer.isActive() && gameTimer.timeLeft() <= 0) {
      return false;
    }
    if (gameMode == SINGLE) {
      return automaticMovesLeft(false);
    } else {
      return (automaticMovesLeft(false) || automaticMovesLeft(true));
    }
  }


  /**
   * Like in chess - stops one timer and activates it for the other player
   */
  public void stopAndToggleTimer() {
    if (gameMode == VS_TURN_BASED) {
      gameTimer.stop();
      gameTimer.swapActiveTimer();
    }
  }


  public void unpauseTimer() {
    if (gameMode == VS_TURN_BASED) {
      gameTimer.resume();
    }
  }


  /**
   * Adds score to grand total
   */
  private void addScore() {
    if (lp.tilesClearedByP1() >= lp.tilesClearedByP2()) {
      p1Wins++;
    }
    if (lp.tilesClearedByP1() <= lp.tilesClearedByP2()) {
      p2Wins++;
    }
  }
  

  /* has to be called after keyresult to make sure new level is loaded
   * or current one reset
   */
  public void updateState(boolean requestMessages) {
    if (updateState == FINISHED_LEVEL) {
      currentLevelIndex++;
      if ((suddenDeath || vs_alt_round2 || testMode) &&
	  currentLevelIndex == levelSet.getSize()) {
	currentLevelIndex = 0;
      }
      setLevel();
      // display custom level start message (if not in test mode)
      if (requestMessages && !testMode) {
	setLevelStartMessage();
      }
    } else if (updateState == LIFE_LOST && (lives > 0 || lives == -1)) {
      lp.reset();
    }
    gameTimer.reset(lp.secondPlayerActive() ? P2 : P1);
    updateState = NONE;
  }


  /**
   * Sets message to be displayed upon start of level
   */
  private void setLevelStartMessage() {
    switch (gameMode) {
    case SINGLE:
    case COOP:
      int area = -1;
      if (unlockable && gameMode == SINGLE) {
	area = ProgressHandler.levelWillUnlockMode(currentLevelIndex + 1);
      }
      // indicate to player that completing this level will unlock something
      if (area != -1) {
	message.request(ProgressHandler.getRequestMessage(area, true));
      } else {
	message.request(levelSet.getLevel(currentLevelIndex).getInText());
      }
      break;
    case VS:
      message.request("Get ready!", SplashMessage.DEFAULT);
      break;
    case VS_TURN_BASED:
      message.request(lp.secondPlayerActive(), SplashMessage.DEFAULT);
    }
  }


  public boolean updateUI(Point mouseLocation) {
    if (mouseLocation != null && closeButton.contains(mouseLocation)) {
      if (!closeButtonActive) {
	closeButtonActive = true;
	return true;
      }
    } else if (closeButtonActive) {
      closeButtonActive = false;
      return true;
    }
    return false;
  }


  /**
   * Indicate a mouse button has been pressed. Note: Use a call to <code>
   * updateUIb</code> first to set the correct mouse location.
   * @param button the mouse button
   * @return <code>QUIT</code> if the close button has been pressed
   */
  public int clickResult(int button) {
    if (button == MouseEvent.BUTTON1 && closeButtonActive) {
      gameTimer.stopAll();
      return QUIT;
    } else {
      return NONE;
    }
  }


  /**
   * Returns the current game state (e.g. <code>GAMEOVER</code>)
   */
  public int getState() {
    return updateState;
  }


  /**
   * Checks whether the level has been started
   * @return <code>true</code> if the level has been started
   */
  public boolean levelHasNotStarted() {
    return lp.hasNotStarted();
  }


  // gets reset when new levelset loaded
  /**
   * Indicate that the MediaMerge tile should be shown in this levelset.
   * Note: This is reset each time a new levelset is loaded
   */
  public void activateMediaMerge() {
    mediaMergeRequired = true;
  }


  /**
   * Indicate that the MediaMerge tile should not be used anymore
   */
  public void revokeMediaMergeTile() {
    if (currentLevelIndex == MEDIAMERGE_LEVEL) {
      imgData.toggleMediaMerge();
    }
  }


  // gets reset when new levelset loaded
  /**
   * Indicate that the editor originating test mode is active
   * Note: This is reset each time a new levelset is loaded
   */
  public void activateTestMode() {
    testMode = true;
  }


  /**
   * Checks whether the (editor) test mode is on
   * @return <code>true</code> if test mode is active
   */
  public boolean testModeActive() {
    return testMode;
  }


  /**
   * Returns the current game mode
   * @return e.g. <code>SINGLE</code>
   */
  public int getGameMode() {
    return gameMode;
  }


  /**
   * Sets number of lives player has at beginning of levelset. This also
   * affects lives player currently has also accepts zero for infinite lives
   * Note: number of lives is reset when a new level is loaded
   * @param amount number of lives to give one or both players
   * @param g graphics context to use for rendering
   */
  public void setLives(int amount, Graphics2D g)
    throws IllegalArgumentException {
    if (amount < 0) {
      throw new IllegalArgumentException("amount is less than zero");
    }
    lives = maxLives = (amount == 0) ? -1 : amount;
    calculateGeometry(g);
  }


  /**
   * Depending on situation this will either render only the timer (i.e. if
   * the <code>GameTimer</code> instance has requested a refresh) or everything
   */
  public void draw(Graphics2D g) {
    if (!gameTimer.timeUpdateRequested()) {
      drawAll(g);
    } else {
      drawTimersOnly(g);
    }
  }

  public void drawAll(Graphics2D g) {
    frames.drawAll(true, gameTimer.isActive(), g);
    tPainter.drawGTiles(lp.getGameTiles(), lp.getP1Position(),
			lp.getP2Position(), levelRequiresNumbers(), g);
    g.setFont(txtFont);
    // LEVEL SET NAME
    Geom.drawString(shortLevelSetName, setName.getLocation(), setNameOffset,
		      txtColor, g);
    // LEVEL NUMBER
    if (!testMode) {
      Geom.drawString("Level    " + (currentLevelIndex + 1),
		      levelNumber.getLocation(), levelNumberOffset, txtColor,
		      g);
    }
    // SCORE
    if (scoringEnabled) {
      Geom.drawString("Score", scoreString.getLocation(),
		      scoreStringOffset, txtColor, g);
      Geom.drawString(Integer.toString(score), scoreStringNum.getLocation(),
		      scoreStringNumOffset, txtColor, g);
    }
    // LIVES REMAINING
    if (lives > -1) {
      Geom.drawString("Lives", livesLeft.getLocation(),
		      livesLeftOffset, txtColor, g);
      Geom.drawString(Integer.toString(lives), livesLeftNum.getLocation(),
		      livesLeftNumOffset, txtColor, g);
      }
    // TILES REMAINING / TILES COLLECTED
    if (gameMode == SINGLE || gameMode == COOP) {
      Geom.drawString("Tiles left", tileText.getLocation(), tileTextOffset,
		      txtColor, g);
      Geom.drawString(Integer.toString(lp.tilesLeft()),
		      tileTextNum.getLocation(), tileTextNumOffset, txtColor,
		      g);
    } else {
      Geom.drawString("Tiles collected", tileText.getLocation(),
		      tileTextOffset, txtColor, g);
      Geom.drawImage(imgData.getImage(Images.T_START), p1TileImg, null, g);
      Geom.drawString("    " + lp.tilesClearedByP1(),
		      p1Tiles.getLocation(), p1TilesOffset, txtColor, g);
      Geom.drawImage(imgData.getImage(Images.T_START2), p2TileImg, null, g);
      Geom.drawString("    " + lp.tilesClearedByP2(),
		      p2Tiles.getLocation(), p2TilesOffset, txtColor, g);
    }
    // TIME REMAINING
    if (gameTimer.isActive()) {
      drawTimers(g);
    }
    // CLOSE
    Geom.drawImage(imgData.getImage(Images.O_CLOSE, closeButtonActive),
		   closeButton, null, g);
  }


  /**
   * Indicates whether the current level should be rendered with numbers on
   * tiles. Note: This will only return <code>true</code> if mediaMerge has
   * also been activated (that way it knows this is the JTP levelset)
   */
  private boolean levelRequiresNumbers() {
    return (mediaMergeRequired && (currentLevelIndex == NUMBERED_LEVEL));
  }


  /**
   * Renders both timers, clearing background first
   */
  private void drawTimersOnly(Graphics2D g) {
    Geom.fillRectMax(timeLeft, frames.getBgColor(), g);
    if (gameMode == VS_TURN_BASED || drawBothTimers) {
      Geom.fillRectMax(timeLeft2, frames.getBgColor(), g);
    }
    g.setFont(txtFont);
    drawTimers(g);
  }


  /**
   * Renders one or both timers
   */
  private void drawTimers(Graphics2D g) {
    Geom.drawImage(imgData.getImage((gameMode != VS_TURN_BASED &&
				     !drawBothTimers) ?
				    Images.E_LEVEL_TIME :
				    Images.O_TIME),
		   timeLeftImg, frames.getBgColor(), g);
    Geom.drawString("    " + gameTimer.toString(P1),
		    timeLeft.getLocation(), timeLeftOffset,
		    gameTimer.timeLeft(P1) > 10000 ?
		    txtColor : Color.red, g);

    if (gameMode == VS_TURN_BASED || drawBothTimers) {
      Geom.drawImage(imgData.getImage(Images.O_TIME, true), timeLeft2Img,
		     frames.getBgColor(), g);
      Geom.drawString("    " + gameTimer.toString(P2),
		      timeLeft2.getLocation(), timeLeft2Offset,
		      gameTimer.timeLeft(P2) > 10000 ?
		      txtColor : Color.red, g);
    }
  }
  

  /**
   * Advances game to next level
   */
  private void setLevel() {
    Level level = levelSet.getLevel(currentLevelIndex);

    // next level will be level with advert, if required
    if (mediaMergeRequired) {
      // ENABLE
      if (currentLevelIndex == MEDIAMERGE_LEVEL) {
	imgData.toggleMediaMerge();
	// DISABLE
      } else if (currentLevelIndex == (MEDIAMERGE_LEVEL + 1)) {
	imgData.toggleMediaMerge();
      }
    }
    gameMode = originalGameMode;
    if (lp == null) {
      /* LevelPlayer only understands two modes: single & two-player
       * i.e. anything other than SINGLE mode in GamePlayer is two-player
       * mode in LevelPlayer
       */
      lp = new LevelPlayer(level, (gameMode != SINGLE));
    } else {
      lp.setLevel(level, (gameMode != SINGLE), vs_alt_round2);
    }
    gameTimer.setTimeLimit(level.getTimeLimit());
    drawBothTimers = false;
    if (gameMode == VS_TURN_BASED) {
      // P1 starts even levels ...
      lp.setActivePlayerTo(!((currentLevelIndex % 2) == 0));
      if (vs_alt_round2) {
	// ... unless it's the 2nd run
	lp.setActivePlayerTo(!lp.secondPlayerActive());
      }
    }
  }

  
  /**
   * Generates a level end message (either random or loads custom one)
   */
  private String getLevelEndMessage() {
    if (random == null) {
      random = new Random();
    }
    int newMessage;
    while((newMessage = random.nextInt(levelEndComments.length ))
	  == lastMessage);
    lastMessage = newMessage;
    return levelEndComments[newMessage];
  }


  /**
   * Returns what borders (frames) are being rendered
   * @return see <code>ScreenFrames</code> class for possible return values
   */
  public int currentFrameMode() {
    if (scoringEnabled) {
      return ScreenFrames.CTRL_P_CHALLENGE;
    } else if (gameMode == VS) {
      return ScreenFrames.CTRL_P_SIMULTANEOUS;
    } else if (gameMode == VS_TURN_BASED) {
      return ScreenFrames.CTRL_P_ALTERNATE;
    } else {
      return ScreenFrames.CTRL_P_NORMAL;
    }
  }


  /**
   * produces dialog where user can select alternate or simultaneous modes
   */
  private int vsModeChooserResult(Component c) {
    String[] buttonStrings = {"Alternate", "Simultaneous"};
    int result = JOptionPane.
      showOptionDialog(c, "\tChoose versus mode:", "Tile Pile - Game",
		       JOptionPane.PLAIN_MESSAGE, JOptionPane.DEFAULT_OPTION,
		       imgData.getIcon(Images.O_START_BOTH),
		       buttonStrings, buttonStrings[0]);
    if (result == 0) {
      return GamePlayer.VS_TURN_BASED;
    } else if (result == 1) {
      return GamePlayer.VS;
    } else {
      return -1;
    }
  }


  /**
   * calculates string and other component locations on screen
   */
  private void calculateGeometry(Graphics2D g) {
    // temporarily set frame set so can calculate status element positions
    int originalFrameType = frames.getCtrlFrameType();
    frames.setCtrlFrameType(currentFrameMode());

    // make sure level set name fits on screen, otherwise shorten
    shortLevelSetName =
      Geom.getMaxString(levelSet.getName(), "...",
			frames.getCtrlOffset().width - 
			(frames.getBorderSpacing() +
			 frames.getFrameThickness()),
			txtFont, g);

    // ************ BOX 1 ************

    // LEVEL SET NAME
    Rectangle tBounds = Geom.calcTextBounds(shortLevelSetName, txtFont, g);
    setNameOffset = tBounds.getLocation();
    setName =
      new Rectangle(frames.getGameCtrlOffset(0).x + frames.getBorderSpacing(),
		    frames.getGameCtrlOffset(0).y + frames.getBorderSpacing(),
		    tBounds.width, tBounds.height);

    // LEVEL NUMBER
    tBounds = Geom.calcTextBounds("Level  " +
				  Integer.toString(levelSet.getSize()),
				  txtFont, g);
    levelNumberOffset = tBounds.getLocation();
    levelNumber =
      new Rectangle(setName.x, setName.y + setName.height + stringSpacing,
		    tBounds.width, tBounds.height);
    
    
    // ************ BOX 2 ************

    // SCORE
    tBounds = Geom.calcTextBounds("Score", txtFont, g);
    scoreStringOffset = tBounds.getLocation();
    scoreString =
      new Rectangle(frames.getGameCtrlOffset(1).x + frames.getBorderSpacing(),
		    frames.getGameCtrlOffset(1).y + frames.getBorderSpacing(),
		    tBounds.width, tBounds.height);
    tBounds = Geom.calcTextBounds("000000", txtFont, g);
    scoreStringNumOffset = tBounds.getLocation();
    scoreStringNum =
      new Rectangle(frames.getGameCtrlOffset(1).x +
		    frames.getBorderSpacing() + numSpacing,
		    scoreString.y, tBounds.width, tBounds.height);
    // convert to placeholder
    if (!scoringEnabled) {
      scoreString.setSize(0, -stringSpacing);
    }

    // LIVES REMAINING
    if (maxLives == -1) {
      livesLeft = scoreString;
    } else {
      tBounds =
	Geom.calcTextBounds("Lives", txtFont, g);
      livesLeftOffset = tBounds.getLocation();
      livesLeft =
	new Rectangle(scoreString.x,
		      scoreString.y + scoreString.height + stringSpacing,
		      tBounds.width, tBounds.height);

      tBounds = Geom.calcTextBounds(maxLives + " ", txtFont, g);
      livesLeftNumOffset = tBounds.getLocation();
      livesLeftNum =
	new Rectangle(frames.getGameCtrlOffset(1).x +
		      frames.getBorderSpacing() + numSpacing,
		      livesLeft.y, tBounds.width, tBounds.height);
    }

    // TILES REMAINING
    if (gameMode == SINGLE || gameMode == COOP) {
      tBounds = Geom.calcTextBounds("Tiles left", txtFont, g);
      tileText =
	new Rectangle(livesLeft.x,
		      livesLeft.y + livesLeft.height + stringSpacing,
		      tBounds.width, tBounds.height);
      tileTextOffset = tBounds.getLocation();
      tBounds = Geom.calcTextBounds("0000", txtFont, g);
      tileTextNum =
	new Rectangle(frames.getGameCtrlOffset(1).x +
		      frames.getBorderSpacing() + numSpacing,
		      tileText.y, tBounds.width, tBounds.height);
      tileTextNumOffset = tBounds.getLocation();
    // TILES COLLECTED
    } else {
      tBounds = Geom.calcTextBounds("Tiles collected", txtFont, g);
      tileText =
	new Rectangle(livesLeft.x,
		      livesLeft.y + livesLeft.height + stringSpacing,
		      tBounds.width, tBounds.height);
      tileTextOffset = tBounds.getLocation();

      p1TileImg = new Square(tileText.x,
			     tileText.y + tileText.height + stringSpacing,
			     imgData.getImage(Images.T_START).getWidth());
      tBounds = Geom.calcTextBounds("    00002", txtFont, g);
      // p1Tiles
      p1TilesOffset = p2TilesOffset = tBounds.getLocation();
      p1Tiles =
	new Rectangle(p1TileImg.x + p1TileImg.width,
		      p1TileImg.y + ((p1TileImg.height - tBounds.height) / 2),
		      tBounds.width, tBounds.height);
      // p2Tiles
      p2TileImg = new Square(p1TileImg.x,
			     p1TileImg.y + p1TileImg.height + stringSpacing,
			     imgData.getImage(Images.T_START2).getWidth());
      p2Tiles =
	new Rectangle(p1Tiles.x,
		      p2TileImg.y + ((p2TileImg.height - tBounds.height) / 2),
		      tBounds.width, tBounds.height);
    }

    // ************ BOX 3 ************

    // TIME REMAINING (will never be active together with above)
    timeLeftImg =
      new Square(frames.getGameCtrlOffset(2).x + frames.getBorderSpacing(),
		 frames.getGameCtrlOffset(2).y + frames.getBorderSpacing(),
		 imgData.getImage(Images.E_LEVEL_TIME).getWidth());
    tBounds = Geom.calcTextBounds("    00 : 002", txtFont, g);
    timeLeftOffset = tBounds.getLocation();
    timeLeft = new Rectangle(timeLeftImg.x + timeLeftImg.width,
			     timeLeftImg.y +
			     ((timeLeftImg.height - tBounds.height) / 2),
			     tBounds.width, tBounds.height);
    
    if (gameMode == VS_TURN_BASED) {
      timeLeft2Img =
	new Square(timeLeftImg.x,
		   timeLeftImg.y + timeLeftImg.height + stringSpacing,
		   imgData.getImage(Images.O_TIME).getWidth());
      tBounds = Geom.calcTextBounds("    00 : 002", txtFont, g);
      timeLeft2Offset = tBounds.getLocation();
      timeLeft2 = new Rectangle(timeLeft2Img.x + timeLeft2Img.width,
				timeLeft2Img.y +
				((timeLeft2Img.height - tBounds.height) / 2),
				tBounds.width, tBounds.height);      
    }
    
    // ************ BOX 4 ************

    // close button
    BufferedImage bI = imgData.getImage(Images.O_CLOSE);
    closeButton =
      new Rectangle(frames.getGameCtrlOffset(3).x + frames.getBorderSpacing(),
		    frames.getGameCtrlOffset(3).y + frames.getBorderSpacing(),
		    bI.getWidth(), bI.getHeight());
	
    frames.setCtrlFrameType(originalFrameType);
  }

}