/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import javax.swing.Timer;
import java.awt.event.*;
import javax.sound.sampled.*;
import java.io.IOException;

/**
 * Class for controlling timed levels.
 * Uses Timer class provide by swing framework.
 * @author V.Termanis
 */
public class GameTimer {

  // needed so that screen updates etc. can be carried out
  private Screen screen;
  // required to update gamestate when timer runs out
  private GamePlayer gamePlayer;
  // underlying swing timer
  private Timer timer;
  private TimeStore[] times;
  private int activeTimer;
  // for deciding whether to redraw whole screen or just the time
  private boolean drawTimeUpdate;
  // sound played when countdown reaches < 10s (each second)
  private Clip beep;
  // action performed when timer expires (once every second)
  private ActionListener levelTimeTicker =
    new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
	if (timeLeft() <= 0) {
	  playBeep();
	  if (gamePlayer.performMoves(GamePlayer.AUTOSTART,
				      GamePlayer.AUTOSTART)
	      == GamePlayer.GAMEOVER) {
	    stopAll();
	    screen.returnToMenu();
	  } else {
	    screen.updateScreen();
	  }
	} else {
	  if (timeLeft() <= 10000) {
	    playBeep();
	  }
	  drawTimeUpdate = true;
	  screen.updateScreen();
	  drawTimeUpdate = false;
	}
      }
    };
  

  /**
   * Constructs a new <code>GameTimer</code> instance with the given
   * attributes. Note: The timer is not started upon construction.
   * @param s parent rendering instance
   * @param gp gameplay instance
   */
  public GameTimer(String beepSoundName, Screen s, GamePlayer gp) {
    times = new TimeStore[2];
    times[0] = new TimeStore();
    times[1] = new TimeStore();
    drawTimeUpdate = false;
    screen = s;
    gamePlayer = gp;
    timer = new Timer(1000, levelTimeTicker);
    try {
      AudioInputStream aIn =
	AudioSystem.getAudioInputStream(getClass().getResource(beepSoundName));
      beep = AudioSystem.getClip();
      beep.open(aIn);
    } catch (NullPointerException e) {
      System.out.println("Warning: Skipping " + beepSoundName +
			 " (File not found)");
      beep = null;
    } catch (IOException e) {
      System.out.println("Warning: Skipping " + beepSoundName +
			 " (File could not be read)");
      beep = null;
    } catch (UnsupportedAudioFileException e) {
      System.out.println("Warning: Skipping " + beepSoundName +
			 " (Unsupported format)");
      beep = null;
    } catch (LineUnavailableException e) {
      System.out.println("Warning: Could not initialise sound");
      beep = null;
    }
  }


  /**
   * Plays beep sound (if possible), stopping any active sounds first.
   */
  private void playBeep() {
    // do not overlap two beeps
    if (beep != null && !beep.isRunning()) {
      beep.stop();
      beep.flush();
      beep.setFramePosition(0);
      beep.start();
    }
  }


  /**
   * Returns currently active timer
   */
  private TimeStore activeTimer() {
    return times[activeTimer];
  }


  /**
   * Starts the active timer if a time limit has been set.
   * If the timer has already been started the started it will be reset and
   * restarted.
   */
  public void start() {
    if (activeTimer().getTimeLimit() > 0) {
      activeTimer().start();
      timer.setInitialDelay(0);
      timer.restart();
    }
  }


  /**
   * Indicates whether <code>start()</code> method has been called.
   * @return <code>true</code> if <code>start()</code> has been called since
   * construction or the last <code>reset()</code> call.
   */
  public boolean hasStarted() {
    return activeTimer().hasStarted();
  }


  /**
   * Stops the active timer.
   */
  public void stop() {
    timer.stop();
    activeTimer().stop();
  }


  /**
   * Stops all (both) timers.
   */
  public void stopAll() {
    timer.stop();
    times[0].stop();
    times[1].stop();
  }


  /**
   * Continues countdown of a timer stopped by <code>stop()</code>.
   */
  public void resume() {
    if (activeTimer().getTimeLimit() > 0) {
      activeTimer().resume();
      // possibly want to refresh screen sooner than 1 second
      timer.setInitialDelay((int) (activeTimer().timeLeft() % 1000));
      timer.restart();
    }
  }


  /**
   * Sets other timer to active one.
   */
  public void swapActiveTimer() {
    activeTimer = (activeTimer + 1) % 2;
  }


  /**
   * Resets start and stop times for all timers as well as setting active
   * timer back to the default.
   */
  public void reset() {
    times[0].reset();
    times[1].reset();
    activeTimer = 0;
  }


  /**
   * Resets start and stop times for all timers as well as setting active
   * timer back to specified one.
   * @param activeTimer either <code>GamePlayer.P1</code> or
   * <code>GamePlayer.P2</code>
   * @throws IllegalArgumentException if <code>activeTimer</code> is not one
   * of <code>GamePlayer.P1</code> or <code>GamePlayer.P2</code>.
   */
  public void reset(int activeTimer) throws IllegalArgumentException {
    if (activeTimer != GamePlayer.P1 && activeTimer != GamePlayer.P2) {
      throw new IllegalArgumentException("activeTimer parameter invalid");
    }
    reset();
    this.activeTimer = activeTimer;
  }


  /**
   * Resets the time added up by using <code>incrementAccumulatedTime()</code>
   * method back to zero for active timer.
   */
  public void resetAccumulatedTime() {
    activeTimer().resetAccumulatedTime();
  }


  /**
   * Sets time limit to use.
   * Note: Same time limit will be used for both timers.
   * @param timeLimit time in seconds. Set this to zero for no time limit.
   * @throws IllegalArgumentException if <code>timeLimit < 0</code>.
   */
  public void setTimeLimit(int timeLimit) throws IllegalArgumentException {
    times[0].setTimeLimit(timeLimit);
    times[1].setTimeLimit(timeLimit);
    reset();
  }


  /**
   * Adds the currently leftover time to the total for active timer.
   * Note: It only makes sense to call this after the timer has been stopped
   * and before it has been reset.
   */
  public void incrementAccumulatedTime() {
    activeTimer().incrementAccumulatedTime();
  }


  /**
   * Returns time added up by using <code>incrementAccumulatedTime()</code>
   * method for active timer.
   * @return accumulated time in milliseconds.
   */
  public long getAccumulatedTime() {
    return activeTimer().getAccumulatedTime();
  }

  
  /**
   * Returns time added up by using <code>incrementAccumulatedTime()</code>
   * method, in seconds, for active timer. Note: The returned value will have
   * been rounded up to the nearest second.
   * @return accumulated time in seconds.
   */
  public long getAccumulatedTimeSeconds() {
    return activeTimer().getAccumulatedTimeSeconds();
  }


  /**
   * Used to decide whether the whole screen or just timer display has to be
   * rendered.
   */
  public boolean timeUpdateRequested() {
    return drawTimeUpdate;
  }


  /**
   * Indicates whether a time limit has been set. Note: This only means a time
   * limit has been set but not whether it is currently counting down.
   * @return <code>true</code> if a time limit has been set.
   */
  public boolean isActive() {
    return (activeTimer().getTimeLimit() > 0);
  }

  
  /**
   * Returns time remaining for active timer.
   * @return time in milliseconds
   */
  public long timeLeft() {
    return activeTimer().timeLeft();
  }


  /**
   * Returns time remaining.
   * @param index either <code>GamePlayer.P1</code> or
   * <code>GamePlayer.P2</code>.
   * @return time in milliseconds
   */
  public long timeLeft(int index) {
    if (index != GamePlayer.P1 && index != GamePlayer.P2) {
      throw new IllegalArgumentException("index is not one of GamePlayer.P1 " +
					 "or GamePlayer.P2");
    }
    return times[index].timeLeft();
  }


  /**
   * Returns time remaining for active timer in <code>mm:ss</code> format.
   * @return time in <code>String</code> representation.
   */
  public String toString() {
    return activeTimer().toString();
  }


  /**
   * Returns time remaining in <code>mm:ss</code> format.
   * @param index either <code>GamePlayer.P1</code> or
   * <code>GamePlayer.P2</code>.
   * @return time in <code>String</code> representation.
   */
  public String toString(int index) throws IllegalArgumentException {
    if (index != GamePlayer.P1 && index != GamePlayer.P2) {
      throw new IllegalArgumentException("index is not one of GamePlayer.P1 " +
					 "or GamePlayer.P2");
    }
    return times[index].toString();
  }

}