/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.awt.font.TextLayout;
import java.awt.image.*;

/**
 * Geometry related support methods
 * @author V.Termanis
 */
public class Geom {

  /**
   * Draws a line between two points (pixel positioning). Note:
   * No line will be drawn in <code>p1</code> and <code>p2</code> are equal.
   * @param p1 the 1st point
   * @param p2 the 2nd point
   * @param c color of the line
   * @param g graphics context for drawing
   */
  private static void drawLineAbs(Point p1, Point p2, Color c, Graphics2D g) {
    if (!p1.equals(p2)) {
      g.setColor(c);
      g.drawLine(p1.x, p1.y, p2.x, p2.y);
    }
  }


  /**
   * Draws a line between tile locations on grid (relative positioning
   * dependant on tile height)
   * @param frames dimensions required for calculating absolute locations
   * @param relP1 relative position of 1st point
   * @param p1Height height of pile for 1st location
   * @param relP2 relative position of 2nd point
   * @param p2Height height of pile for 2st location
   * @param c color of the line
   * @param g graphics context for drawing   
   */
  public static void drawLineRel(ScreenFrames frames, Point relP1,
				 int p1Height, Point relP2, int p2Height,
				 Color c, Graphics2D g) {
    drawLineAbs(absPos(frames, relP1, p1Height),
		absPos(frames, relP2, p2Height),
		c, g);
  }


  /**
   * Draws a rectangle.
   * Note: The resulting rectangle will be 1 pixel smaller in both directions
   * than if <code>Graphics2D</code> version of drawRect were called.
   * @param r size and location of rectangle to draw
   * @param c color of rectangle outline
   * @param g graphics context for drawing   
   */
  public static void drawRect(Rectangle r, Color c, Graphics2D g) {
    g.setColor(c);
    g.drawRect(r.x, r.y, r.width - 1, r.height - 1);
  }


  /**
   * Draws a filled rectangle.
   * @param r size and location of rectangle to draw
   * @param c color of rectangle
   * @param g graphics context for drawing   
   */  public static void fillRect(Rectangle r, Color c, Graphics2D g) {
    fillRect(r, 0, 0, c, g);
  }


  /**
   * Modified version of fillRect() for use in editor for correct cropping
   */
  protected static void fillRect(Rectangle r, int topCrop, int bottomCrop,
				 Color c, Graphics2D g) {
    g.setColor(c);
    g.fillRect(r.x, r.y + topCrop,
	       r.width - 1, r.height - (1 + topCrop + bottomCrop));
  }


  /**
   * Fills a rectangle exactly as specified in
   * <code>Graphics2D.fillRect()</code>
   * @param r size and location of rectangle to draw
   * @param c color of rectangle outline
   * @param g graphics context for drawing
   */
  public static void fillRectMax(Rectangle r, Color c, Graphics2D g) {
    g.setColor(c);
    g.fillRect(r.x, r.y, r.width, r.height);
  }


  /**
   * Like <code>fillRectMax</code> except offset by one pixel horizontally
   * and vertically, i.e.: x - 1, y - 1, width + 2, height + 2.
   * @param r size and location of rectangle to draw
   * @param c color of rectangle outline
   * @param g graphics context for drawing
   */
  public static void fillRectOversize(Rectangle r, Color c, Graphics2D g) {
    g.setColor(c);
    g.fillRect(r.x - 1, r.y - 1, r.width + 2, r.height + 2);
  }


  /**
   * Fills a rectangle with rounded edges.
   * <code>Graphics2D.fillRect()</code>
   * @param r size and location of rectangle to draw
   * @param arc amount of curvature
   * (see <code>Graphics2D.fillRoundRect()</code> for more details)
   * @param c color of rectangle
   * @param g graphics context for drawing
   */
  public static void fillRoundRect(Rectangle r, int arc, Color c,
				   Graphics2D g) {
    g.setColor(c);
    g.fillRoundRect(r.x, r.y, r.width - 1, r.height - 1, arc, arc);
  }


  /**
   * Draws a rectangle with rounded edges.
   * <code>Graphics2D.fillRect()</code>
   * @param r size and location of rectangle to draw
   * @param arc amount of curvature
   * (see <code>Graphics2D.fillRoundRect()</code> for more details)
   * @param c color of rectangle outline
   * @param g graphics context for drawing
   */
  public static void drawRoundRect(Rectangle r, int arc, Color c, Stroke s,
				   Graphics2D g) {
    Stroke oldStroke = null;

    g.setColor(c);
    if (s != null) {
      oldStroke = g.getStroke();
      g.setStroke(s);
    }
    g.drawRoundRect(r.x, r.y, r.width - 1, r.height - 1, arc, arc);
    if (s != null) {
      g.setStroke(oldStroke);
    }
  }


  /**
   * Draws a character at the given location in the specified color
   * @param c character to draw
   * @param x absolute horizontal position
   * @param y absolute vertical position
   * @param clr color of character
   * @param g graphics context for drawing
   */
  public static void drawChar(char c, int x, int y, Color clr, Graphics2D g) {
    drawString(String.valueOf(c), x, y, clr, g);
  }


  /**
   * Draws a number at the given location in the specified color
   * @param n number to draw
   * @param x absolute horizontal position
   * @param y absolute vertical position
   * @param c color of character
   * @param g graphics context for drawing
   */ 
  public static void drawNumber(int n, int x, int y, Color c, Graphics2D g) {
    drawString(Integer.toString(n), x, y, c, g);
  }


  /**
   * Draws a string at the given location in the specified color
   * @param s string to draw
   * @param x absolute horizontal position
   * @param y absolute vertical position
   * @param c color of character
   * @param g graphics context for drawing
   */ 
  public static void drawString(String s, int x, int y, Color c,
				Graphics2D g) {
    g.setColor(c);
    g.drawString(s, x, y);
  }


  /**
   * Draws a string at the given location in the specified color
   * @param s string to draw
   * @param p absolute position
   * @param c color of character
   * @param g graphics context for drawing
   */ 
  public static void drawString(String s, Point p, Color c,
				Graphics2D g) {
    drawString(s, p.x, p.y, c, g);
  }


  /**
   * Draws a string at the given location with an offset.
   * This is useful once the offset has been calculated
   * (see <code>Geom.calcTextBounds()</code>)
   * @param s string to draw
   * @param p absolute position
   * @param offset deviation from absolute position
   * @param c color of character
   * @param g graphics context for drawing
   */ 
  public static void drawString(String s, Point p, Point offset, Color c,
				Graphics2D g) {
    drawString(s, p.x - offset.x, p.y - offset.y, c, g);
  }


  /**
   * Renders an image at the specified location, clearing the background
   * first with the specified color.
   * @param img image to render
   * @param r size of image and absolute position for rendering
   * @param bgColor color with which to clear area first
   * @param g graphics context for drawing
   */ 
  public static void drawImage(BufferedImage img, Rectangle r, Color bgColor,
			       Graphics2D g) {
    if (bgColor != null) {
      fillRectMax(r, bgColor, g);
    }
    g.drawImage(img, null, r.x, r.y);    
  }



  /**
   * Draws grid in editor
   * @param gridArea where grid starts on screen
   * @param gridSize specifies number of rows and columns to draw
   * @param tileSize size of each grid square
   * @param c color of the grid
   * @param g graphics context for drawing
   */
  public static void drawGrid(Rectangle gridArea, int gridSize,
			      int tileSize, Color c, Graphics2D g) {
    g.setColor(c);
    for (int i = 0; i < gridSize; i++) {
      for (int j = 0; j < gridSize; j++) {
	g.drawRect(gridArea.x + i * tileSize, gridArea.y + j * tileSize,
		   tileSize - 1, tileSize - 1);
      }
    }
  }


  /**
   * Calculates absolute locations and size of a tile
   * @param frames dimensions required for calculating absolute locations
   * @param tile relative position on grid
   * @return the absolute size and position of the tile on the screen
   * @throws NullPointerException if either of the parameters is
   * <code>null</code>.
   */
  public static Square tilePosToArea(ScreenFrames frames, Point tile)
    throws NullPointerException {
    if (frames == null || tile == null) {
      throw new NullPointerException("frames or tile is null");
    }
    return new Square(tile.x * frames.getTileSize() + frames.getGridArea().x,
		      tile.y * frames.getTileSize() + frames.getGridArea().y,
		      frames.getTileSize());
  }


  /**
   * Calculates maximum clip area for one tile
   * @param frames dimensions required for calculating absolute locations
   * @param tile relative position on grid
   * @return absolute clip area on the screen
   */
  public static Square calcTileClip(ScreenFrames frames, Point tile) {
    return new Square(tile.x * frames.getTileSize() + frames.getGridArea().x,
		      (tile.y - 1) * frames.getTileSize() +
		      frames.getGridArea().y,
		      frames.getTileSize() * 2);
  }

  
  /**
   * Calculates maximum clip area for two tiles
   * @param frames dimensions required for calculating absolute locations
   * @param tile1 1st tile
   * @param tile2 2nd tile
   * @return absolute clip area on the screen
   */
  public static Rectangle calcTilesClip(ScreenFrames frames, Point tile1,
					  Point tile2) {
    return calcTileClip(frames, tile1).union(calcTileClip(frames, tile2));
  }


  /**
   * Calculates relative grid position given absolute mouse location
   * @param frames dimensions required for calculating absolute locations
   * @param mouseLocation absolute location of mouse cursor
   * @return relative position on grid, or <code>null</code> if mouse
   * not on grid
   */
  public static Point gridPos(ScreenFrames frames, Point mouseLocation) {
    if (frames.getGridArea().contains(mouseLocation)) {
      return new Point((mouseLocation.x - frames.getGridArea().x) /
		       frames.getTileSize(),
		       (mouseLocation.y - frames.getGridArea().y) /
		       frames.getTileSize());
    } else {
      return null;
    }
  }


  /**
   * Calculates absolute location (in pixels) of centre of tile face.
   * This is used to draw teleport lines in editor
   * @param frames dimensions required for calculating absolute locations
   * @param p relative grid position
   * @param tileHeight height of pile at given position
   * @return absolute location of centre of tile face
   */
  public static Point absPos(ScreenFrames frames, Point p, int tileHeight) {
    // centre of tile pile base
    Point abs =
      new Point(frames.getGridArea().x + (p.x * frames.getTileSize()) +
		(frames.getTileSize() / 2),
		frames.getGridArea().y + (p.y * frames.getTileSize()) +
		(frames.getTileSize() / 2));
    // centre of tile pile with given height
    abs.translate(frames.getHeightPerLevel() * tileHeight,
		  - (frames.getHeightPerLevel() * tileHeight));
    return abs;
  }


  /**
   * Calculates space taken up by given string and its offset
   * @param s string to calculate required space for
   * @param f font to calculate space usage for
   * @param g graphics context for off-screen rendering
   */
  public static Rectangle calcTextBounds(String s, Font f, Graphics2D g) {
    return (new TextLayout(s, f, g.getFontRenderContext())).getBounds().
      getBounds();
  }


  /**
   * Calculates offset for drawing a single character on the face of a tile
   * @param c the character for which to calculate the offset
   * @param f font to calculate offset for
   * @param frames dimensions required for calculating absolute locations
   * @param g graphics context for off-screen rendering
   */
  public static Point calcCharOnTileOffset(char c, Font f,
					   ScreenFrames frames, Graphics2D g) {
    Rectangle charBounds = calcTextBounds(String.valueOf(c), f, g);
    
    return new Point((frames.getTileSize() - charBounds.width) / 2 -
		     charBounds.x,
		     (frames.getTileSize() - charBounds.height) / 2 -
		     charBounds.y);
  }


  /**
   * Calculates maximum substring which can fit in the given absolute width
   * @param s the string for which to calculate fitting substring
   * @param addition string to append to substring if all will not fit
   * @param maxWidth maximum width of whole string in pixels
   * @param f font to calculate dimensions for
   * @param g graphics context for off-screen rendering
   */
  public static String getMaxString(String s, String addition, int maxWidth,
				    Font f, Graphics2D g) {
    int sLength, additionLength = 0;
    Rectangle tBounds;

    // no change needed
    if (calcTextBounds(s, f, g).width < maxWidth) {
      return s;
    }
    
    // work out space required for additional ending
    if (addition != null) {
      tBounds = calcTextBounds(addition, f, g);
      // impossible if the additino itself is too long
      if (tBounds.width >= maxWidth) {
	return null;
      }
      additionLength = tBounds.width;
    }
    
    String newS = s;

    do {
      newS = newS.substring(0, newS.length() - 1);
      sLength = calcTextBounds(newS, f, g).width;
    } while ((sLength + additionLength > maxWidth) && newS.length() > 1);
    
    // if the string is still too long
    if (sLength + additionLength > maxWidth) {
      return null;
    } else {
      return newS + addition;
    }
  }


  /**
   * Calculates width required for a button rendered with tiles
   * @param s string which will be rendered on button
   * @param is3D indicates whether button is raised or not
   * @param frames dimensions required for calculating absolute locations
   * @return width in pixels of button
   */
  public static int getTileButtonWidth(String s, boolean is3D,
				       ScreenFrames frames) {
    return ((s.length() * frames.getTileSize()) + (is3D ? 10 : 0));
  }


  // maximum line height of a number of lines in the given graphics context
  /**
   * Calculates maximum line height and width for a set lines of text.
   * Useful for spacing lines with the exact same spacing
   * @param lines text to analyze
   * @param f font with which to calculate text dimensions
   * @return maximum dimensions for each line of text
   */
  public static Dimension calcMaxLineDimensions(String[] lines, Font f,
						Graphics2D g) {
    Dimension d = new Dimension();
    Rectangle lineBounds;

    for (int i = 0; i < lines.length; i++) {
      lineBounds = calcTextBounds(lines[i], f, g);
      d.width = Math.max(d.width, lineBounds.width);
      d.height = Math.max(d.height, lineBounds.height);
    }
    return d;
  }

}