/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.IOException;
import javax.swing.JOptionPane;

/**
 * Defines function and rendering of the highscore UI component
 * @author V.Termanis
 */
public class Highscores {

  // possble mouseover states
  public static final int NONE = -1;
  public static final int CLOSE = 0;
  public static final int SINGLE = 1;
  public static final int COOP = 2;
  public static final int SCORE_SET = 3;
  public static final int NAME_CHANGED = 4;

  private Rectangle buttons[];
  private ScreenFrames frames;
  private TilePainter tPainter;
  /* active - mouse over, press - depressed button, location of new highscore
   * in list so know where to draw it
   */
  private int
    activeButton, pressedButton, oldButton,
    newScoreIndex, newScore,
    maxChars,
    buttonBorderThickness;
  private Color buttonBorderColor;
  private Stroke buttonBorderStroke;
  private Images imgData;
  private boolean showing2PScores, addScoreMode;
  // used temporarily when adding a highscore
  private Progress progress;
  // editable highscore name
  private String enteredName;


  public Highscores(Images imgData, ScreenFrames frames,
		    TilePainter tilePainter, Color buttonBorderColor,
		    float buttonBorderThickness, int buttonSpacing,
		    int maxChars) {
    activeButton = pressedButton = oldButton = NONE;
    this.imgData = imgData;
    this.frames = frames;
    this.tPainter = tilePainter;
    this.buttonBorderColor = buttonBorderColor;
    this.buttonBorderThickness = Math.round(buttonBorderThickness);
    buttonBorderStroke = new BasicStroke(buttonBorderThickness);
    this.maxChars = maxChars;
    calculateGeometry(buttonSpacing);
  }


  public Rectangle updateUI(Point mouseLocation) {
    if (mouseLocation == null) {
      return null;
    }

    oldButton = activeButton;
    activeButton = NONE;
    
    // determine current position
    for (int i = 0; i < buttons.length; i++) {
      if (buttons[i].contains(mouseLocation)) {
	activeButton = i;
	break;
      }
    }
    // no view change buttons in "edit" mode
    if (addScoreMode && activeButton != CLOSE) {
      activeButton = NONE;
    }

    // pressed key - only ever highlight original mouse down button
    if (pressedButton != NONE && activeButton != pressedButton) {
      activeButton = NONE;
    }

    // calculate clip area
    if (oldButton == activeButton) {
      return null;
    } else {
      // only need to redraw one button
      if (oldButton == NONE) {
	return buttons[activeButton];
      } else if (activeButton == NONE) {
	return buttons[oldButton];
      } else {
	// clipping area for old and new button
	return buttons[oldButton].union(buttons[activeButton]);
      }
    }
  }


  /**
   * Choose which highscore table should be shown
   * @param b set to <code>true</code> to show the two player table
   */
  public void set2PTableAsActive(boolean b) {
    showing2PScores = b;
  }

  
  public int clickResult(Point mouseLocation) {
    int buttonResult = pressedButton;
    // only a click if still over the same button as before
    if (pressedButton == NONE || pressedButton != activeButton) {
      buttonResult = NONE;
    }
    switch(buttonResult) {
    case SINGLE:
      showing2PScores = false;
      break;
    case COOP:
      showing2PScores = true;
      break;
    case CLOSE:
      resetHighscoreState();
    }
    pressedButton = NONE;
    return buttonResult;
  }

  
  /**
   * Resets highscore settings to default
   */
  private void resetHighscoreState() {
    progress = null;
    addScoreMode = showing2PScores = false;
  }


  // note the mouseLocation parameter is not handled in this class
  public void mousePressed(int button, Point mouseLocation) {
    if (button == MouseEvent.BUTTON1) {
      pressedButton = activeButton;
    }
  }


  public int keyResult(int keyCode, Component c) {
    if (addScoreMode) {
      switch(keyCode) {
	// DELETE ONE CHAR
      case KeyEvent.VK_BACK_SPACE:
      case KeyEvent.VK_DELETE:
	if (enteredName.length() > 0) {
	  enteredName = enteredName.substring(0, enteredName.length() - 1);
	  return NAME_CHANGED;
	} else {
	  return NONE;
	}
	// SAVE HIGHSCORE
      case KeyEvent.VK_ENTER:
	enteredName = enteredName.trim();
	if (enteredName.length() == 0) {
	  enteredName = "Anonymous";
	}
	progress = ProgressHandler.loadProgress();
	progress.addHighScore(showing2PScores, newScore, enteredName);
	try {
	  ProgressHandler.saveProgress(progress);
	} catch (IOException e) {
	  JOptionPane.
	    showMessageDialog(c, "Could not save highscores (to " +
			      "progress.dat)", "Tile Pile - Highscore error",
			      JOptionPane.ERROR_MESSAGE,
			      imgData.getIcon(Images.O_ERROR));
	}
	addScoreMode = false;
	return SCORE_SET;
	// CANCEL HIGHSCORE
      case KeyEvent.VK_ESCAPE:
	addScoreMode = false;
	progress = null;
	return SCORE_SET;
	// QUIT HIGHSCORE MENU
      case KeyEvent.VK_F9:
	resetHighscoreState();
	return CLOSE;
      default:
	return NONE;
      }
    } else {
      if (keyCode == KeyEvent.VK_ENTER) {
	showing2PScores ^= true;
	return (showing2PScores ? COOP : SINGLE);
      } else if (keyCode == KeyEvent.VK_ESCAPE || keyCode == KeyEvent.VK_F9) {
	resetHighscoreState();
	return CLOSE;
      } else {
	return NONE;
      }
    }
  }
  

  /**
   * Used to process key events for entering highscore name
   */
  public int keyTypedResult(char c) {
    if (addScoreMode) {
      if (isValidHighscoreChar(c) && enteredName.length() < maxChars) {
	enteredName += c;
	return NAME_CHANGED;
      } else {
	return NONE;
      }
    }
    return NONE;
  }


  /**
   * 
   */
  private static boolean isValidHighscoreChar(char c) {
    int type = Character.getType(c);

    return (type == Character.CONNECTOR_PUNCTUATION ||
	    type == Character.CURRENCY_SYMBOL ||
	    type == Character.DASH_PUNCTUATION ||
	    type == Character.DECIMAL_DIGIT_NUMBER ||
	    type == Character.START_PUNCTUATION ||
	    type == Character.END_PUNCTUATION ||
	    type == Character.INITIAL_QUOTE_PUNCTUATION ||
	    type == Character.FINAL_QUOTE_PUNCTUATION ||
	    type == Character.LETTER_NUMBER ||
	    type == Character.LOWERCASE_LETTER ||
	    type == Character.MATH_SYMBOL ||
	    type == Character.MODIFIER_SYMBOL ||
	    type == Character.UPPERCASE_LETTER ||
	    type == Character.TITLECASE_LETTER ||
	    type == Character.OTHER_LETTER ||
	    type == Character.OTHER_PUNCTUATION ||
	    c == ' ');
  }


  // only draw changes since last mouse move
  public void drawUpdate(Graphics2D g) {
    if (oldButton != activeButton) {
      drawButton(activeButton, true, g);
      drawButton(oldButton, false, g);
    }
  }


  /**
   * Decide whether the given score needs to be enterd in the highscore table
   * @param coop indicates whether the score is from cooperative mode or not
   * @param score number of points made
   */
  public void dealWithScore(boolean coop, int score) {
    // show correct table for given mode
    showing2PScores = coop;
    progress = ProgressHandler.loadProgress();
    if ((newScoreIndex = progress.isHighScoreWhere(coop, score)) >= 0) {
      addScoreMode = true;
      // placeholder will be discarded by reloading highscores from disk
      progress.addHighScore(showing2PScores, score, "PLACEHOLDER");
    }
    enteredName = new String();
    newScore = score;
  }


  public void drawAll(Graphics2D g) {
    frames.drawAll(true, g);
    drawHighscores(g);
    if (!addScoreMode) {
      // draw border around selected button
      if (frames.getRenderQuality() == ScreenFrames.RENDERING_BALANCED) {
	g.setRenderingHint(RenderingHints.KEY_RENDERING,
			   RenderingHints.VALUE_RENDER_QUALITY);
	g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			   RenderingHints.VALUE_ANTIALIAS_ON);      
      }
      g.setColor(buttonBorderColor);
      g.setStroke(buttonBorderStroke);
      int index = showing2PScores ? 2 : 1;
      g.drawRoundRect(buttons[index].x - (buttonBorderThickness + 2),
		      buttons[index].y - (buttonBorderThickness + 2),
		      buttons[index].width +
		      ((buttonBorderThickness - 1) * 2) + 5,
		      buttons[index].height +
		      ((buttonBorderThickness - 1) * 2) + 5,
		      8, 8);
      if (frames.getRenderQuality() == ScreenFrames.RENDERING_BALANCED) {
	g.setRenderingHint(RenderingHints.KEY_RENDERING,
			   RenderingHints.VALUE_RENDER_DEFAULT);
	g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			   RenderingHints.VALUE_ANTIALIAS_OFF);      
      }
      drawButton(SINGLE, (activeButton == SINGLE), g);
      drawButton(COOP, (activeButton == COOP), g);
    }
    drawButton(CLOSE, (activeButton == CLOSE), g);
  }


  private void drawButton(int index, boolean active, Graphics2D g) {
    if (index != NONE) {
      Geom.drawImage(imgData.getOtherImage(index, active),
		     buttons[index], null, g);
    }
  }


  private void drawHighscores(Graphics2D g) {
    int score;
    String name;
    boolean editable, first = true;
    
    if (progress == null) {
      progress = ProgressHandler.loadProgress();
    }

    for (int i = (Progress.HIGH_SCORE_COUNT - 1); i >= 0; i--) {
      name = progress.getName(showing2PScores, i);
      score = progress.getScore(showing2PScores, i);
      
      /* editing new highscore - replace existing entry with what will be
       * shown there instead
       */
      if (addScoreMode && (i == newScoreIndex)) {
	name = enteredName;
	score = newScore;
	editable = true;
      } else {
	editable = false;
      }
      
      if (name != null) {
	tPainter.drawScoreString(name, editable, first,
				 frames.getGridArea().x,
				 frames.getGridArea().y + 5 +
				 (((Progress.HIGH_SCORE_COUNT - 1) - i) * 60),
				 g);
	tPainter.drawScoreString(String.valueOf(score), false, first,
				 frames.getCtrlOffset().x +
				 frames.getCtrlOffset().width -
				 (frames.getBorderSpacing() * 2) - 
				 (String.valueOf(score).length() *
				  frames.getTileSize()),
				 frames.getGridArea().y + 5 +
				 (((Progress.HIGH_SCORE_COUNT - 1) - i) * 60),
				 g);
      }
      first = false;
    }
  }


  /**
   * Works out position of all components draw as part of highscore UI
   */
  private void calculateGeometry(int buttonSpacing) {
    buttons = new Rectangle[3];

    // close button
    BufferedImage bI = imgData.getImage(Images.O_CLOSE);
    buttons[0] =
      new Rectangle(frames.getCtrlOffset().x + frames.getCtrlOffset().width -
		    frames.getBorderSpacing() - bI.getWidth(),
		    frames.getCtrlOffset().y + frames.getCtrlOffset().height -
		    frames.getBorderSpacing() - bI.getHeight(),
		    bI.getWidth(), bI.getHeight());
    // coop button
    bI = imgData.getImage(Images.O_2P);
    buttons[2] =
      new Rectangle(buttons[0].x - (bI.getWidth() + buttonSpacing),
		    buttons[0].y, bI.getWidth(), bI.getHeight());
    
   // single player button
    bI = imgData.getImage(Images.O_1P);
    buttons[1] =
      new Rectangle(buttons[2].x - (bI.getWidth() + buttonSpacing),
		    buttons[2].y, bI.getWidth(), bI.getHeight());
  }

}