/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.io.*;
import java.awt.*;
import java.awt.image.BufferStrategy;
import javax.swing.*;
import java.net.*;

/**
 * Set of methods for loading levelsets with appropriate the error handling.
 * @author V.Termanis
 */
public class IO {

  /**
   * <code>LevelSet</code> will be loaded for play.
   */
  public static final int GAME_LOAD = 0;
  /**
   * <code>LevelSet</code> will be created in editor.
   */
  public static final int EDITOR_NEW = 1;
  /**
   * <code>LevelSet</code> will be loaded in editor.
   */
  public static final int EDITOR_LOAD = 2;

  // error messages used in more than one method
  private static final String
    eMessageP1 = "Could not open level set file (",
    eMessageP2 = ").\nThe file is either corrupt, not a level set, does " +
    "not exist or you do not have the necessary permissions to open it.";


  /**
   * Tries to load the given file for editing.
   * @param editor instance of <code>Editor</code> used for editing.
   * @param filename absolute or relative location of file to load.
   * @param op instance of <code>JOptionPane</code> to use for displaying
   * error messages to the user.
   * @param c parent rendering class (required by <code>JOptionPane</code>).
   * @param g graphics context for updating editor layout.
   * @return <code>true</code> if the levelset was loaded successfully.
   */
  public static boolean editorLoadLevelSet(Editor editor, String filename,
					   JOptionPane op, Screen c,
					   Graphics2D g) {
    String
      wMessageP1 = "You will not be able to save changes made to this level " +
      "set because the file (",
      wMessageP2 = ") is read-only.";

    File f = new File(filename);
    // test writeability
    if (f.exists() && !f.canWrite()) {
      op.setIcon(c.getImageData().getIcon(Images.O_INFO));
      op.setMessage(wMessageP1 + f.getName() + wMessageP2);
      op.createDialog(c, "Tile Pile - Editor").setVisible(true);
    }
    try {
      editor.loadLevelSet(filename, c, g);
    } catch (Exception e) {
      op.setIcon(c.getImageData().getIcon(Images.O_ERROR));
      op.setMessage(eMessageP1 + f.getName() + eMessageP2);
      op.createDialog(c, "Tile Pile - Editor").setVisible(true);
      return false;
    }
    return true;
  }


  /**
   * Tries to load the given file for play.
   * @param gamePlayer instance of <code>GamePlayer</code> used for playing.
   * @param filename absolute or relative location of file to load.
   * @param op instance of <code>JOptionPane</code> to use for displaying
   * error messages to the user.
   * @param c parent rendering class (required by <code>JOptionPane</code>).
   * @param gameMode type of game to set up. See <code>GamePlayer</code> for
   * possible modes.
   * @param g graphics context for updating in-game layout.
   * @return <code>true</code> if the levelset was loaded successfully.
   */
  public static boolean gameLoadLevelSet(GamePlayer gamePlayer,
					 String filename, JOptionPane op,
					 Screen c, int gameMode,
					 Graphics2D g) {
    try {
      return gameLoadLevelSet(gamePlayer, (new File(filename).toURI().toURL()),
			      op, c, gameMode, g);
      // should never happen
    } catch (Exception e) {
      op.setIcon(c.getImageData().getIcon(Images.O_ERROR));
      op.setMessage("Internal error: failed to convert filename (" + filename +
		    ") to URL (via URI)");
      op.createDialog(c, "Tile Pile - Game").setVisible(true);
      return false;
    }
  }
      

  /**
   * Tries to load the given file for play.
   * Note: This method can be used to load a set from:<ul>
   * <li>local or remote disk</li>
   * <li>website</li>
   * <li>JAR</li></ul>
   * @param gamePlayer instance of <code>GamePlayer</code> used for playing.
   * @param fileURL location of file to load.
   * @param op instance of <code>JOptionPane</code> to use for displaying
   * error messages to the user.
   * @param c parent rendering class (required by <code>JOptionPane</code>).
   * @param gameMode type of game to set up. See <code>GamePlayer</code> for
   * possible modes.
   * @param g graphics context for updating in-game layout.
   * @return <code>true</code> if the levelset was loaded successfully.
   */
  public static boolean gameLoadLevelSet(GamePlayer gamePlayer, URL fileURL,
					 JOptionPane op, Screen c,
					 int gameMode, Graphics2D g) {
    String
      mMessageP1 = "This level set (",
      mMessageP2 = ") does not support two player modes.";

    try {
      int result;
      if ((result = gamePlayer.loadLevelSet(fileURL, gameMode, c, g))
	  != GamePlayer.LOAD_OK) {
	if (result == GamePlayer.LOAD_NOT2P) {
	  op.setIcon(c.getImageData().getIcon(Images.O_INFO));
	  op.setMessage(mMessageP1 + fileURL.getFile() + mMessageP2);
	  op.createDialog(c, "Tile Pile - Game").setVisible(true);
	}
	return false;
      } else {
	return true;
      }
      // should never happen
    } catch (Exception e) {
      op.setIcon(c.getImageData().getIcon(Images.O_ERROR));
      op.setMessage(eMessageP1 + fileURL.getFile() + eMessageP2);
      op.createDialog(c, "Tile Pile - Game").setVisible(true);
      return false;
    }
  }


  /**
   * Prompts user for level set file and returns the absolute path to the
   * chosen file. The returned path is guaranteed to end in
   * <code>fileExtension</code>.
   * @param fc instance of <code>JFileChooser</code> for user to select file
   * with.
   * @param requestType one of <code>GAME_LOAD</code>, <code>EDITOR_NEW</code>
   * or <code>EDITOR_LOAD</code>. This dictates the type of error messages
   * shown to the user.
   * @param fileExtension to add to filename, if not already present.
   * @param c parent rendering class (required by <code>JFileChooser</code>).
   * @return absolute path to chosen file or <code>null</code>.
   * @throws IllegalArgumentException if <code>requestType</code> is not one of
   * the types specified in this class.
   * @throws NullPointerException if any of <code>fc</code>,
   * <code>fileExtension</code> or <code>c</code> are <code>null</code>.
   */
  public static String getUserFile(JFileChooser fc, int requestType,
				   String fileExtension, Screen c)
    throws IllegalArgumentException, NullPointerException {
    String[][] dialogStrings =
      {
	{"Tile Pile - Choose a level set to play", "Play",
	 "Play selected level set"},
	{"Tile Pile - Choose name and location for new level set", "Create",
	 "Create a new level set with given name"},
	{"Tile Pile - Please choose your Tile Pile file", "Open",
	 "Opens selected level set for editing"}
      };

    if (fc == null || fileExtension == null || c == null) {
      throw new NullPointerException("one of the parameters is null");
    }

    if (requestType != GAME_LOAD && requestType != EDITOR_NEW &&
	requestType != EDITOR_LOAD) {
      throw new IllegalArgumentException("requestType invalid");
    }
    
    fc.setDialogTitle(dialogStrings[requestType][0]);
    fc.setApproveButtonText(dialogStrings[requestType][1]);
    fc.setApproveButtonToolTipText(dialogStrings[requestType][2]);

    if (fc.showOpenDialog(c) == JFileChooser.APPROVE_OPTION) {
      String filepath = fc.getSelectedFile().getPath();
      // add extension to filename if necessary (only if does not exist)
      if (!fc.getSelectedFile().exists()) {
	int i = filepath.lastIndexOf('.');
	if (!(i > 0 && i == filepath.length() - 4 &&
	      filepath.substring(i + 1).toLowerCase().equals(fileExtension))) {
	  filepath += "." + fileExtension;
	}
      }
      File f = new File(filepath);
      if (requestType == EDITOR_NEW && f.exists()) {
	if (JOptionPane.
	    showOptionDialog(c, "A file with the chosen name exists already." +
			     " Do you want overwrite this file?",
			     "Tile Pile - Editor", JOptionPane.YES_NO_OPTION,
			     JOptionPane.WARNING_MESSAGE,
			     c.getImageData().getIcon(Images.O_WARNING),
			     null, "No")
	    == JOptionPane.YES_OPTION) {
	  if (!f.delete()) {
	    JOptionPane.
	      showMessageDialog(c, "Could not overwrite " +
				filepath, "Tile Pile - Editor",
				JOptionPane.ERROR_MESSAGE,
				c.getImageData().getIcon(Images.O_ERROR));
	    filepath = null;
	  }
	} else {
	  filepath = null;
	}
      } else if (requestType == EDITOR_LOAD && !f.exists()) {
	JOptionPane.
	  showMessageDialog(c, "Cannot find " + filepath,
			    "Tile Pile - Editor",
			    JOptionPane.ERROR_MESSAGE,
			    c.getImageData().getIcon(Images.O_ERROR));
	filepath = null;
      }

      return filepath;
    } else {
      return null;
    }
  }

}