/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import javax.imageio.ImageIO;
import java.io.IOException;
import java.awt.*;
import java.awt.image.*;
import java.net.URL;

/**
 * This class is used to load the various image resources used by the game.
 * @author V.Termanis
 */
public class ImageLoader {

  
  /**
   * Loads a set of images.
   * File names are combined as follows:<br>
   * <code>imgPrefix + imgName + imgPostfix + '.' + imgExt</code>
   * @param cl instance of <code>Class</code> which is needed to access
   * image files via the <code>getResource()</code> method.
   * @param imgPath relative or absolute path to the images
   * @param imgPrefix prefix used by all image files
   * @param imgExt the extension (without '.') used by all image files
   * @param imgPostfix different endings used by each image group
   * @param imgNames names of image groups to load
   * @param requiredImagePostfixLimit minimum number of images that have to
   * exist (if they do not then <code>IOException</code> will be thrown)
   * @return An array of groups containing images with all different endings
   * @throws IndexOutOfBoundsException if
   * <code>requiredImagePostfixLimit</code> > <code>imgPostfix.length</code>
   * @throws NullPointerException if any of the parameters are <code>null
   * </code>
   * @throws IOException if the image cannot be loaded
   */
  public static BufferedImage[][] getImages(Class cl, String imgPath,
					    String imgPrefix, String imgExt,
					    String[] imgPostfix,
					    String[] imgNames,
					    int requiredImagePostfixLimit)
    throws IndexOutOfBoundsException, NullPointerException, IOException {

    if (cl == null || imgPath == null || imgPrefix == null || imgExt == null ||
	imgPostfix == null || imgNames == null) {
      throw new NullPointerException("one of the parameters is null");
    }
    if (requiredImagePostfixLimit > imgPostfix.length) {
      throw new IndexOutOfBoundsException("required image number greater " +
					  "than provided name array size");
    }

    BufferedImage[][] tmpImages;
    BufferedImage[][] images;
    
    // request loading of all images
    tmpImages = new BufferedImage[imgNames.length][];
    for (int i = 0; i < imgNames.length; i++) {
      tmpImages[i] = new BufferedImage[imgPostfix.length];
      for (int j = 0; j < imgPostfix.length; j++) {
	String filename =
	  imgPath + imgPrefix + imgNames[i]+ imgPostfix[j] + imgExt;
	// file existance check
	URL imgLocation;
	if ((imgLocation = cl.getResource(filename)) == null) {
	  if (j < requiredImagePostfixLimit) {
	    throw new IOException("Cannot find " + filename);
	  } else {
	    // substitute for missing image
	  tmpImages[i][j] = tmpImages[i][j - 1];
	  }
	} else {
	  try {
	    tmpImages[i][j] = ImageIO.read(imgLocation);
	  } catch (Exception e) {
	    throw new IOException("Failed to load " + filename);
	  }
	}
      }
    }
    // convert fully loaded images to buffered
    images = new BufferedImage[imgNames.length][];
    for (int i = 0; i < imgNames.length; i++) {
      images[i] = new BufferedImage[imgPostfix.length];
      for (int j = 0; j < imgPostfix.length; j++) {
	images[i][j] =
	  GraphicsEnvironment.getLocalGraphicsEnvironment().
	  getDefaultScreenDevice().getDefaultConfiguration().
	  createCompatibleImage(tmpImages[i][0].getWidth(),
				tmpImages[i][0].getHeight(),
				Transparency.TRANSLUCENT);
	//Transparency.OPAQUE, Transparency.BITMASK, Transparency.TRANSLUCENT
	Graphics2D g = images[i][j].createGraphics();
	g.drawImage(tmpImages[i][j], 0, 0, null);
      }
    }
    return images;
  }


  /**
   * Loads an image from the given location.
   * @param imgLocation full path or link to the image to load
   * @return the loaded image
   * @throws IOException if the image could not be read succesfully
   * @throws NullPointerException if <code>imgLocation</code> is
   * <code>null</code>
   */
  public static BufferedImage loadImage(URL imgLocation)
    throws IOException, NullPointerException {
    if (imgLocation == null) {
      throw new NullPointerException("imgLocation is null");
    }
    return ImageIO.read(imgLocation);
  }

}