/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.image.BufferedImage;
import javax.swing.ImageIcon;

/**
 * Defines and loads game graphics resources (staged).
 * Note: Images are not loaded in constructor to enable (more granular) loading
 * progress to be shown on program startup. (Hence there is no custom
 * constructor method.)
 * @author V.Termanis
 */
public class Images {

  // tiles
  public static final int T_NORMAL = 0;
  public static final int T_ICE = 1;
  public static final int T_TELEPORT = 2;
  public static final int T_GRAVITY = 3;
  public static final int T_ARROW_UP = 4;
  public static final int T_ARROW_DOWN = 5;
  public static final int T_ARROW_LEFT = 6;
  public static final int T_ARROW_RIGHT = 7;
  public static final int T_START = 8;
  public static final int T_START2 = 9;

  // editor
  public static final int E_LEVEL_SWAP = 10;
  public static final int E_LEVEL_FIRST = 11;
  public static final int E_LEVEL_LAST = 12;
  public static final int E_LEVEL_MESSAGES = 13;
  public static final int E_LEVEL_BACK = 14;
  public static final int E_LEVEL_FORWARD = 15;
  public static final int E_LEVEL_TIME = 16;
  public static final int E_CLEAR = 17;
  public static final int E_PLAY = 18;
  public static final int E_SAVE = 19;
  public static final int E_CLOSE = 20;
  public static final int E_SAVECLOSE = 21;
  public static final int E_ARROW_UP = 22; 
  public static final int E_ARROW_DOWN = 23;
  public static final int E_ARROW_LEFT = 24;
  public static final int E_ARROW_RIGHT = 25;
  public static final int E_LEVEL_NUMBER_BG = 26;

  // misc
  public static final int O_CLOSE = 27;
  public static final int O_1P = 28;
  public static final int O_2P = 29;
  public static final int O_3D = 30;
  public static final int O_3D_GRAVITY = 31;
  public static final int O_3D_ICE = 32;
  public static final int O_3D_TELEPORT = 33;
  public static final int O_START_BOTH = 34;
  public static final int O_TIME = 35;
  // level chooser buttons
  public static final int O_FIRST = 36;
  public static final int O_BACK = 37;
  public static final int O_FORWARD = 38;
  public static final int O_LAST = 39;
  // Andrej's company logo to feature in one level
  public static final int O_MEDIAMERGE = 40;
  public static final int O_DISABLED_TILE = 41;
  public static final int O_DISABLED_TILE_P1 = 42;
  public static final int O_DISABLED_TILE_P2 = 43;
  // dialog replacement "icons"
  public static final int O_INFO = 44;
  public static final int O_QUESTION = 45;
  public static final int O_WARNING = 46;
  public static final int O_ERROR = 47;

  // do not require images with a postfix with index over 0 (e.g. _in)
  private static final int requiredImagePostfixLimit = 1;

  // relative path from JTP root
  private static final String imgPath = "img/";
  // strings to append to image names
  private static final String imgPostfix[] = {"", "_in"};
  // extension to append to full image name
  private static final String imgExt = ".png";
  // strings to prepend to relevant image names (see tNames etc. below)
  private static final String tPrefix = "tile_";
  private static final String ePrefix = "editor_";
  private static final String oPrefix = "misc_";


  private static final String[] tNames = {"normal", "ice", "teleport",
					  "gravity", "arrowup", "arrowdown", 
					  "arrowleft", "arrowright", "start",
					  "start2"};
  private static final String[] eNames = {"swap", "first", "last", "messages",
					  "back", "forward", "time", "clear",
					  "play", "save", "close", "saveclose",
					  "arrowup", "arrowdown", "arrowleft",
					  "arrowright", "level_number_bg"};
  private static final String[] oNames = {"close", "1P", "2P", "3D",
					  "3D_gravity", "3D_ice",
					  "3D_teleport", "start_both", "time",
					  "first", "back", "forward", "last", 
					  "mediamerge", "disabled_tile",
					  "disabled_tile_P1",
					  "disabled_tile_P2", "info",
					  "question", "warning", "error"};
  
  // final image resource storage
  private BufferedImage[][] images;
  // temporary image storage for staged loading process
  private BufferedImage[][] tImages, eImages, oImages;


  /**
   * Indicates whether more images still have to be loaded.
   * @return <code>false</code> if all image resources have been loaded into
   * memory, <code>true</code> otherwise.
   */
  public boolean moreToLoad() {
    return (images == null &&
	    (tImages == null || eImages == null ||
	     oImages == null || images == null));
  }


  /**
   * Loads the next set of image resources into memory.
   * @return <code>false</code> if there is nothing left to load, <code>true
   * </code>otherwise.
   */
  public boolean loadNextSet() throws java.io.IOException {
    // tile images
    if (tImages == null) {
      tImages =
	ImageLoader.getImages(getClass(), imgPath, tPrefix, imgExt, imgPostfix,
			      tNames, requiredImagePostfixLimit);
      // editor images
    } else if (eImages == null) {
      eImages =
	ImageLoader.getImages(getClass(), imgPath, ePrefix, imgExt, imgPostfix,
			      eNames, requiredImagePostfixLimit);
      // miscellaneous images
    } else if (oImages == null) {
      oImages =
	ImageLoader.getImages(getClass(), imgPath, oPrefix, imgExt, imgPostfix,
			      oNames, requiredImagePostfixLimit);
      // combine separate resource arrays into one
    } else if (images == null) {
      images = 
	new BufferedImage[tImages.length + eImages.length + oImages.length][];
      int i;
      for (i = 0; i < tImages.length; i++) {
	images[i] = tImages[i];
      }
      for (i = 0; i < eImages.length; i++) {
	images[i + tImages.length] = eImages[i];
      }
      for (i = 0; i < oImages.length; i++) {
	images[i + tImages.length + eImages.length] =
	  oImages[i];
      }
      /* not needed anymore so make sure Java knows these can be
       * garbage collected now
       */
      tImages = eImages = oImages = null;
    } else {
      return false;
    }
    return true;
  }

 
  /**
   * Returns the image resource associated with the given index.
   * @param index of the wanted image
   * @param in set to <code>true</code> to request the "_in" version of this
   * resource
   * @return the requested image resource
   */
  public BufferedImage getImage(int index, boolean in) {
    return images[index][in ? 1 : 0];
  }


  /**
   * Returns the image resource associated with the given index.
   * @param index of the wanted image
   * @return the requested image resource
   */
  public BufferedImage getImage(int index) {
    return getImage(index, false);
  }


  /**
   * Returns the image resource assicated with the given index (relative to
   * the "other" image group).
   * @param index relative index of the wanted image
   * @param in set to <code>true</code> to request the "_in" version of this
   * resource
   * @return the requested image resource
   */
  public BufferedImage getOtherImage(int index, boolean in) {
    return getImage(index + getOtherImageOffset(), in); 
  }


  /**
   * Returns the matching "3D" wall for a tile of the given index. Note:
   * see the <code>Tile</code> class for a list of tile indices.
   * @param tileIndex index of the wanted "3D" image
   * @param in set to <code>true</code> to request the "_in" version of this
   * resource
   * @return the requested image resource
   */
  public BufferedImage get3DImage(int tileIndex, boolean in) {
    switch (tileIndex) {
    case T_GRAVITY:
      return getImage(O_3D_GRAVITY, in);
    case T_ICE:
      return getImage(O_3D_ICE, in);
    case T_TELEPORT:
      return getImage(O_3D_TELEPORT, in);
    default:
      return getImage(O_3D, in);
    }
  }


  /**
   * Returns the total number of tile image resources (excluding "3D" walls)
   * @return the number of images
   */
  public int getTileImageCount() {
    return tNames.length;
  }

  
  /**
   * Returns starting offset for the tile image group.
   * @return the offset
   */
  public int getTileImageOffset() {
    return 0;
  }


  /**
   * Returns the total number of editor image resources (excluding common ones)
   * @return the number of images
   */
  public int getEditorImageCount() {
    return eNames.length;
  }

  
  /**
   * Returns starting offset for the editor image group.
   * @return the offset
   */
  public int getEditorImageOffset() {
    return tNames.length;
  }


  /**
   * Returns the total number of other (common) image resources.
   * @return the number of images
   */
  public int getOtherImageCount() {
    return oNames.length;
  }


  /**
   * Returns starting offset for the other (common) image group.
   * @return the offset
   */
  public int getOtherImageOffset() {
    return getEditorImageOffset() + eNames.length;
  }


  /**
   * Returns the image resource with the given (absolute) index in icon form.
   * @param index the absolute index of this image
   * @return the converted image
   */
  public ImageIcon getIcon(int index) {
    return new ImageIcon(getImage(index));
  }


  /**
   * Swaps the <code>T_NORMAL</code> image resource with <code>O_MEDIAMERGE
   * </code>.
   */
  public void toggleMediaMerge() {
    BufferedImage[] tmp = images[T_NORMAL];
    images[T_NORMAL] = images[O_MEDIAMERGE];
    images[O_MEDIAMERGE] = tmp;
  }

}