/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.event.*;
import javax.swing.Timer;

/**
 * Class for handling keyboard control in-game.
 * @author V.Termanis
 */
public class KeyHandler implements KeyListener {
  
  /**
   * Handling mode: Bypassed (handling performed by <code>Screen</code> class)
   */
  public static final int NORMAL = 0;
  /**
   * Handling mode: Only Player 1 keys allowed
   */
  public static final int GAME_1P = 1;
  /**
   * Handling mode: Both Player keys allowed
   */
  public static final int GAME_2P = 2;
  
  // Keys used by first player.
  private static final int[] keysP1 = {KeyEvent.VK_UP, KeyEvent.VK_DOWN,
				       KeyEvent.VK_LEFT, KeyEvent.VK_RIGHT};
  // Keys used by second player.
  private static final int[] keysP2 = {KeyEvent.VK_W, KeyEvent.VK_S,
				       KeyEvent.VK_A, KeyEvent.VK_D};
  /* Keys used to control game (e.g. quit)
   * Note: This array is not final because END key might be added to the list
   * depending on whether development mode is switched on or not (see
   * constructor for this class below)
   */
  private static int[] keysCtrl = {KeyEvent.VK_F9, KeyEvent.VK_ESCAPE};
  // Delay before keyboard repeats start (in milliseconds)
  private static final int preRepeatDelay = 250;
  // Delay between repeats (in milliseconds)   1000/45 ~ 22fps
  private static final int repeatDelay = 45;

  // one of NORMAL, GAME_1P or GAME_2P - defines how key events are handled
  private int keyMode;
  // initial delay before repeats start
  private Timer[] delay;
  // repeat timer - used for both players
  private Timer repeater;
  /* keeps track of which gameplay keys are being held down and which is the
   * active one
   */
  private PlayerKeys[] pressedKeys;
  /* key action to only be performed once, e.g. control or when holding down
   * starts but before repeats
   */
  // indicates whether repeat mode is active for either player
  private boolean[] repeatsActive;
  // parent class to forward events to
  private Screen screen;
  /* indicates whether to reacto to SPACE key (only for discarding popup
   * messages inbetween levels
   * repeatsAllowed - should holding down a key produce repeated events or not
   */
  private boolean spaceKeyActive, repeatsAllowed;

  /**
   * Constructs a new instance of <code>KeyHandler</code>.
   * @param s the parent processing class.
   */
  public KeyHandler(Screen s) {
    keyMode = NORMAL;
    repeatsAllowed = true;
    delay = new Timer[2];
    delay[GamePlayer.P1] = new Timer(preRepeatDelay, new RepeatActivator(0));
    delay[GamePlayer.P2] = new Timer(preRepeatDelay, new RepeatActivator(1));
    delay[GamePlayer.P1].setRepeats(false);
    delay[GamePlayer.P2].setRepeats(false);
    repeater = new Timer(repeatDelay, new RepeatHandler());
    repeater.setInitialDelay(0);
    pressedKeys = new PlayerKeys[2];
    pressedKeys[GamePlayer.P1] = new PlayerKeys();
    pressedKeys[GamePlayer.P2] = new PlayerKeys();
    screen = s;
    // Include level skip key if in development mode
    if (Main.develModeOn()) {
      int[] tmp = {KeyEvent.VK_F9, KeyEvent.VK_ESCAPE, KeyEvent.VK_END};
      keysCtrl = tmp;
    }
  }


  /**
   * Sets key input handling mode.
   * @param mode the mode to use. Available modes:
   * <ul><li><code>NORMAL</code> - Simply forwards all events to the <code>
   * Screen</code> class</li>
   * <li><code>GAME_1P</code> - Handling of 1st player keys only</li>
   * <li><code>GAME_2P</code> - Handling of both players' keys</li></ul>
   * @param repeatsAllowed indicates whether holding down a key should produce
   * multiple events or not. Note: This parameter has no effect in <code>NORMAL
   * </code> mode.
   * @throws IllegalArgumentException if <code>mode</code> is not one of
   * <code>NORMAL</code>, <code>GAME_1P</code> or <code>GAME_2P</code>.
   */
  public void setMode(int mode, boolean repeatsAllowed)
    throws IllegalArgumentException {
    if (mode < NORMAL || mode > GAME_2P) {
      throw new IllegalArgumentException();
    }
    keyMode = mode;
    this.repeatsAllowed = repeatsAllowed;
    repeater.stop();
    reset();
  }


  /**
   * Resets state of this <code>KeyHandler</code> instance.
   * Note: The <code>repeatsAllowed</code> flag is not reset.
   */
  public void reset() {
    synchronized(this) {
      delay[GamePlayer.P1].stop();
      delay[GamePlayer.P2].stop();
      pressedKeys[GamePlayer.P1].clear();
      pressedKeys[GamePlayer.P2].clear();
      spaceKeyActive = false;
    }
  }


  /**
   * To be called at start of level (to perform automatic moves, if any)
   */
  public void autoStart() {
    synchronized(this) {
      reset();
      pressedKeys[0].setOneOffKey(GamePlayer.AUTOSTART);
      if (keyMode == GAME_2P) {
	pressedKeys[1].setOneOffKey(GamePlayer.AUTOSTART);
      }
      repeater.start();
    }
  }

  
  /**
   * To be called when a popup message is shown in-game so it can be cancelled
   * via keyboard.
   */
  public void enableSpaceKey() {
    spaceKeyActive = true;
  }


  /**
   * Handles key-down events.
   * @param e the incoming keyboard event.
   */
  public void keyPressed(KeyEvent e) {
    // bypass this class if requested (i.e. when not in game)
    if (keyMode == NORMAL) {
      screen.keyPressedNormal(e);
    } else {
      synchronized(this) {
	int keyCode = e.getKeyCode();
	int playerIndex;
	// PLAYER KEYS
	if ((playerIndex = getPlayer(keyCode, keyMode)) != -1) {
	  // key already being held down, ignore
	  if (pressedKeys[playerIndex].keyDown(keyCode) == false) {
	    return;
	  }
	  // CONTROL KEYS (IN-GAME)
	} else if (isOneOf(keyCode, keysCtrl) ||
		   (spaceKeyActive && keyCode == KeyEvent.VK_SPACE)) {
	  /* it does not matter which player's key-cache is used for control
	   * and control keys take priority over everything else
	   */
	  pressedKeys[0].setOneOffKey(keyCode);
	  // discard all other keypresses
	} else {
	  return;
	}
	/* initialise repeat delay if the key is a player key and repeats
	 * are allowed
	 */
	if (playerIndex != -1) {
	  pressedKeys[playerIndex].setOneOffKey(keyCode);
	  if (repeatsAllowed) {
	    delay[playerIndex].restart();
	  }
	}
	/* the repeater timer is responsible for all moves, so turn it on
	 * if it is not already
	 */
	if (!repeater.isRunning()) {
	  repeater.start();
	}
      }
    }
  }


  /**
   * Handles key-up events.
   * @param e the incoming keyboard event.
   */
  public void keyReleased(KeyEvent e) {
    // bypass this class if requested (i.e. when not in game)
    if (keyMode == NORMAL) {
      screen.keyReleased(e);
    } else {
      int playerIndex;
      synchronized(this) {
	/* if the released key is a player key which was the active held-down
	 * key then cancel the (possibly) active repeat delay
	 */
	if ((playerIndex = getPlayer(e.getKeyCode(), keyMode)) != -1 &&
	    pressedKeys[playerIndex].keyUp(e.getKeyCode())) {
	  delay[playerIndex].stop();
	}
      }
    }
  }


  /**
   * Used by <code>Screen</code> class to determine whether a key is being
   * held down in-game.
   * @return <code>true</code> is a player or control key is in the down state,
   * given the current key handling mode. Note: In <code>NORMAL</code> mode
   * this will always return <code>false</code>.
   */
  public boolean keyDown() {
    synchronized(this) {
      if (keyMode != NORMAL) {
	return (pressedKeys[GamePlayer.P1].activeKey() != PlayerKeys.NONE ||
		pressedKeys[GamePlayer.P1].oneOffKeySet() ||
		pressedKeys[GamePlayer.P2].activeKey() != PlayerKeys.NONE ||
		pressedKeys[GamePlayer.P2].oneOffKeySet());
      } else {
	return false;
      }
    }
  }


  /**
   * Handles key-typed events.
   * Note: This event is always bypassed (i.e. handled by <code>Scene</code>
   * class because it is only used when entering characters as part of a new
   * highscore.
   * @param e the incoming keyboard event.
   */
  public void keyTyped(KeyEvent e) {
    if (keyMode == NORMAL) {
      screen.keyTyped(e);
    }
  }

  
  /**
   * Either returns number of player or -1 it control key.
   * Note: If keyMode is not set to GAME_2P player two keys wil result in
   * a return value of -1 also.
   */
  private static int getPlayer(int keyIndex, int keyMode) {
    if (isOneOf(keyIndex, keysP1)) {
      return GamePlayer.P1;
    } else if (keyMode == GAME_2P && isOneOf(keyIndex, keysP2)) {
      return GamePlayer.P2;
    } else {
      return -1;
    }
  }


  /**
   * Should be synchronised when called (as this method itself is not)!
   */
  private boolean playerIsActive(int player) {
    return (pressedKeys[player].oneOffKeySet() ||
	    pressedKeys[player].isRepeating());
  }


  /**
   * Should be synchronised when called (as this method itself is not)!
   */
  private int getPlayerActiveKey(int player) {
    if (pressedKeys[player].oneOffKeySet()) {
      return pressedKeys[player].getAndClearOneOffKey();
    } else if (pressedKeys[player].isRepeating()) {
      return pressedKeys[player].activeKey();
    } else {
      return PlayerKeys.NONE;
    }
  }


  /**
   * quick check for whether a keycode is in keysP1, keysP2, etc.
   */
  private static boolean isOneOf(int keyIndex, int[] group) {
    for (int i = 0; i < group.length; i++) {
      if (keyIndex == group[i]) {
	return true;
      }
    }	
    return false;
  }
  
  
  /**
   * Initiates repeats for one player
   */
  private void startRepeats(int playerIndex) {
    synchronized(this) {
      // make sure repeat key has not been cancelled in the mean time
      if (pressedKeys[playerIndex].activeKey() != PlayerKeys.NONE) {
	pressedKeys[playerIndex].setRepeats(true);
	repeater.start();
      }
    }
  }


  /**
   * handles simulated keyboard repeat events
   */
  private class RepeatHandler implements ActionListener {

    public void actionPerformed(ActionEvent e) {
      int p1Key, p2Key;

      synchronized(this) {
	// determine which keys need to be processed
	p1Key = getPlayerActiveKey(GamePlayer.P1);
	if (keyMode == GAME_2P) {
	  p2Key = getPlayerActiveKey(GamePlayer.P2);
	} else {
	  p2Key = PlayerKeys.NONE;
	}
      }

      screen.keyPressedGame(p1Key, p2Key, e.getWhen());

      // determine whether to stop (no keys held down and no automoves)
      synchronized(this) {
	if (!pressedKeys[0].isRepeating() &&
	    !pressedKeys[0].oneOffKeySet() &&
	    !pressedKeys[1].isRepeating() &&
	    !pressedKeys[1].oneOffKeySet()) {
	  if (!screen.automaticMovesLeft()) {
	    repeater.stop();
	    return;
	  }
	}
      }
    }
  }
  

  /**
   * simple handler for activating keyboard repeats
   */
  private class RepeatActivator implements ActionListener {
    int playerIndex;

    public RepeatActivator(int i) {
      playerIndex = i;
    }

    public void actionPerformed(ActionEvent evt) {
      startRepeats(playerIndex);
    }
  }

}