/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.Point;
import java.awt.Rectangle;
import java.util.ArrayList;

/**
 * Deals with handling and storage of individual levels.
 * @author V.Termanis
 */
public class Level implements java.io.Serializable {

  // had to be introduced after removing unnecessary method in v0.981
  static final long serialVersionUID = 6051174117516761959L;

  private Point start, start2;
  // tile stack limit
  private int maxTileHeight;
  // tiles[y][x], i.e. column first
  private Tile[][] tiles;
  // Messages to be displayed before starting / after finishing level
  private String inText, outText;
  // time limit, 0 = no limit
  private int timeLimit;
  // list of teleport start/destination pairs
  private ArrayList <Point[]> tpLocations;
  // two player level? (if two player, can also be used as single player
  private boolean twoPlayer;

  
  /**
   * Constructs a new <code>Level</code> instance with the given attributes.
   * @param width in tiles of the level
   * @param height in tiles of the level
   * @param maxTileHeight maximum height allowed for any one tile pile
   * @param timeLimit in which player has to complete this level. Set to 0 for
   * no time limit
   * @param twoPlayer specifies whether this a single or two-player level
   * @throws IllegalArgumentException if any of the following are satisfied:
   * <ul>
   * <li><code>width</code> < 2</li>
   * <li><code>height</code> < 2</li>
   * <li><code>maxTileHeight</code> < 1</li>
   * <li><code>timeLimit</code> < 0</li></ul>
   */
  public Level(int width, int height, int maxTileHeight, int timeLimit,
	       boolean twoPlayer) {
    initLevel(width, height, maxTileHeight, timeLimit, twoPlayer);
  }


  /**
   * Constructs a new <code>Level</code> instance with the given attributes and
   * no time limit.
   * @param width in tiles of the level
   * @param height in tiles of the level
   * @param maxTileHeight maximum height allowed for any one tile pile
   * @param twoPlayer specifies whether this a single or two-player level
   * @throws IllegalArgumentException if any of the following are satisfied:
   * <ul>
   * <li><code>width</code> < 2</li>
   * <li><code>height</code> < 2</li>
   * <li><code>maxTileHeight</code> < 1</li></ul>
   */
  public Level(int width, int height, int maxTileHeight, boolean twoPlayer) {
    initLevel(width, height, maxTileHeight, 0, twoPlayer);
  }


  /**
   * Constructs a new single-player <code>Level</code> instance with the given
   * attributes.
   * @param width in tiles of the level
   * @param height in tiles of the level
   * @param maxTileHeight maximum height allowed for any one tile pile
   * @param timeLimit in which player has to complete this level. Set to 0 for
   * no time limit
   * @throws IllegalArgumentException if any of the following are satisfied:
   * <ul>
   * <li><code>width</code> < 2</li>
   * <li><code>height</code> < 2</li>
   * <li><code>maxTileHeight</code> < 1</li>
   * <li><code>timeLimit</code> < 0</li></ul>
   */
  public Level(int width, int height, int maxTileHeight, int timeLimit) {
    initLevel(width, height, maxTileHeight, timeLimit, false);
  }


  /**
   * Constructs a new single-player <code>Level</code> instance with the given
   * attributes and no time limit.
   * @param width in tiles of the level
   * @param height in tiles of the level
   * @param maxTileHeight maximum height allowed for any one tile pile
   * @throws IllegalArgumentException if any of the following are satisfied:
   * <ul>
   * <li><code>width</code> < 2</li>
   * <li><code>height</code> < 2</li>
   * <li><code>maxTileHeight</code> < 1</li></ul>
   */
  public Level(int width, int height, int maxTileHeight) {
    initLevel(width, height, maxTileHeight, 0, false);
  }


  /**
   * Used by all constructors to validate and set the level attributes.
   */
  private void initLevel(int width, int height, int maxTileHeight,
			 int timeLimit, boolean twoPlayer)
    throws IllegalArgumentException {
    if (maxTileHeight < 1 || height < 2 || width < 2 || timeLimit < 0) {
      throw new IllegalArgumentException("one of the parameters is invalid");
    }
    tiles = new Tile[height][];
    for (int i = 0; i < height; i++) {
      tiles[i] = new Tile[width];
    }
    this.maxTileHeight = maxTileHeight;
    this.timeLimit = timeLimit;
    this.twoPlayer = twoPlayer;
  }


  /**
   * Pushes a tile onto the grid (works a bit like a stack - hence the name).
   * If the given tile has the same parameters (apart from height),
   * the tile at location gets a height increase. Otherwise the
   * tile's type and attributes are changed but height stays the
   * same (unless changing from zero height start tile).
   * If attributes == null also just increase height (or create normal tile)
   * @param location on grid where to push the given tile on
   * @param p2Start if <code>attributeTile</code> is a start tile then
   * this indicates whether it is the start of the 1st or 2nd player
   * @throws NullPointerException if <code>location</code> is <code>null</code>
   * @throws IllegalArgumentException if <code>location</code> is invalid
   */
  public void pushTile(Point location, Tile attributeTile, boolean p2Start)
    throws NullPointerException, IllegalArgumentException {
    if (location == null) {
      throw new NullPointerException("location is null");
    } else if (!validCoordinates(location)) {
      throw new IllegalArgumentException("location is not a valid coordinate");
    }
    // no attributes given
    if (attributeTile == null) {
      if (tile(location) != null) {
	incTileHeight(location);
	// if no tile present, create normal one
      } else {
	try {
	  setTile(location, new NrmTile(1, false));
	  // should never happen
	} catch (TileException e) {
	  e.printStackTrace();
	}
      }
      // no tile present
    } else if (tile(location) == null) {
      if (isStart(attributeTile)) {
	setStart(location, p2Start);
      }
      setTile(location, attributeTile);
      // no attribute change
    } else if (tile(location).equalsApartFromHeight(attributeTile)) {
      /* all attributes might be the same, but is the other player's start to
       * be added (in which case do not increase height)
       */
      if (!hasEqualStarts()) {
	if ((start != null && location.equals(start) && p2Start) ||
	    (start2 != null && location.equals(start2) && !p2Start)) {
	  setStart(location, p2Start);
	} else {
	  incTileHeight(location);
	}
      } else {
	incTileHeight(location);
      }
    } else {
      // preserve height of tile
      attributeTile.setHeight(tile(location).getHeight());
      // becomes start
      if (isStart(attributeTile)) {
	setStart(location, p2Start);
	// cancel start
      } else if (isStart(location)) {
	cancelStart(location);
      }
      setTile(location, attributeTile);
    }
  }


  /**
   * Pops a tile from the grid.
   * @param location on grid where to pop tile from
   * @throws NullPointerException if <code>location</code> is <code>null</code>
   * @throws IllegalArgumentException if <code>location</code> is invalid
   */
  public void popTile(Point location) throws NullPointerException,
					     IllegalArgumentException {
    if (location == null) {
      throw new NullPointerException("location is null");
    } else if (!validCoordinates(location)) {
      throw new IllegalArgumentException("location is not a valid coordinate");
    }
    Tile t = tile(location);
    if (t != null) {
      // remove start
      if (t.getHeight() == 0) {
	cancelStart(location);
      } else if (t.getHeight() == 1) {
	ArrayList <Point> startPs;
	// check whether teleport destination - in which case remove teleport
	if ((startPs = findTeleportStarts(location)) != null) {
	  replaceTeleports(startPs);
	}

	// if a teleport tile, remove from tparray
	if (t.getType() == Tile.TELEPORT) {
	  removeTeleportLocation(location);
	}

	// allow height to be decremented to 0 for start tile only
	if (isStart(t)) {
	  t.decHeight();
	} else {
	  setTile(location, null);
	}
      } else {
	t.decHeight();
      }
    }
  }


  /* exclude START, TELEPORT & teleport destinations
   * setting attributeTile to null will remove a tile
   */
  /**
   * Push or pop multiple tiles.
   * Notes:<ul>
   * <li>This method utilises Level.push() and Level.pop()</li>
   * <li><code>START</code> and <code>TELEPORT</code> type tiles will not be
   * altered by this operation.</li></ul>
   * @param corner1 1st corner for specifying push/pop area
   * @param corner2 2nd corner for specifying push/pop area
   * @param attributeTile tile to push, or set to <code>null</code> to pop
   */
  public void manipulateMultiple(Point corner1, Point corner2,
				 Tile attributeTile) {
    Rectangle area = pointsToArea(corner1, corner2);
    Tile t;
    Point tLocation = new Point(0, 0);

    for (int x = area.x; x <= (area.x + area.width); x++) {
      for (int y = area.y; y <= (area.y + area.height); y++) {
	tLocation.setLocation(x,y);
	t = tile(tLocation);
	// PUSH
	if (attributeTile != null &&
	    (t == null ||
	     (t != null && t.getType() != Tile.TELEPORT &&
	      !(start != null && isStart(tLocation))))) {
	  pushTile(tLocation, attributeTile.copy(), false);
	  // POP
	} else if (t != null && t.getType() != Tile.TELEPORT &&
		   !(start != null && isStart(tLocation)) &&
		   countTeleportTilesTo(tLocation) < t.getHeight()) {
	  popTile(tLocation);
	}
      }
    }
  }


  /**
   * Used in manipulateMultiple() to tranform input parameters.
   */
  private Rectangle pointsToArea(Point p1, Point p2) {
    Rectangle r = new Rectangle(p1);
    r.add(p2);
    return r;
  }


  /**
   * Returns type of tile at given location
   * @param location of tile
   * @throws NullPointerException if <code>location</code> is null
   * @throws IllegalArgumentException if <code>location</code> is invalid
   * @return the type of the tile as specified in the <code>Tile</code> class
   */
  public int tileType(Point location) throws NullPointerException,
					     IllegalArgumentException {
    if (location == null) {
      throw new NullPointerException("location is null");
    } else if (!validCoordinates(location)) {
      throw new IllegalArgumentException("location is not a valid coordinate");
    }
    return tile(location).getType();
  }


  /**
   * Returns 1st player start location
   * @return start location of 1st player.
   */
  public Point getStart() {
    return start;
  }


  /**
   * Returns 2nd player start location
   * @return start location of 2nd player.
   */
  public Point getStart2() {
    return start2;
  }


  /**
   * Determine whether a particular location of the grid is a start
   * @param location to query
   * @return <code>true</code> if tile at <code>location</code> is a start
   */
  private boolean isStart(Point location) {
    if (location != null) {
      if ((start != null && start.equals(location)) ||
	  (start2 != null && start2.equals(location))) {
	return true;
      } else
	return false;
    } else {
      return false;
    }
  }


  /**
   * Clears a level. This returns it to the state of a newly create level.
   */
  public void clear() {
    start = start2 = null;
    if (tpLocations != null) {
      tpLocations.clear();
    }
    for (int y = 0; y < lvlHeight(); y++) {
      for (int x = 0; x < lvlWidth(); x++) {
	setTile(x, y, null);
      }
    }
  }


  /**
   * Validates a level.
   * This method does not determine whether a level is soluble. Instead it uses
   * the following rules:<br>
   * A level needs a start position and has to contain at least one
   * tile adjacent to the start position (or both in case of two player mode)
   * @return <code>true</code> if the level meets the above criteria
   */
  public boolean isValid() {
    if (isValid(start) && (!twoPlayer || (twoPlayer && isValid(start2)))) {
      return true;
    } else {
      return false;
    }
  }


  /**
   * Determines whether the current point is a valid start (i.e. whether
   * there is at least one move possible from this location
   */
  private boolean isValid(Point s) {
    if (s != null &&
	((s.x > 0 && tile(s.x - 1, s.y) != null &&
	  tile(s.x - 1, s.y).getHeight() >= 1) ||
	 (s.x < lvlWidth() - 1 && tile(s.x + 1, s.y) != null &&
	  tile(s.x + 1, s.y).getHeight() >= 1)  ||
	 (s.y > 0 && tile(s.x, s.y - 1) != null &&
	  tile(s.x, s.y - 1).getHeight() >= 1) ||
	 (s.y < lvlHeight() - 1 && tile(s.x, s.y + 1) != null &&
	  tile(s.x, s.y + 1).getHeight() >= 1))) {
      return true;
    } else {
      return false;
    }
  }

  
  /**
   * Determines whether the level is empty / blank.
   * @return <code>true</code> if no tiles have been placed
   */
  public boolean isEmpty() {
    for (int y = 0; y < lvlHeight(); y++) {
      for (int x = 0; x < lvlWidth(); x++) {
	if (tile(x, y) != null) {
	  return false;
	}
      }
    }
    return true;
  }
  
  
  /**
   * Sets start of either player to a particular location.
   * Note: This method only compliments the push()/pop() methods
   */
  private void setStart(Point location, boolean p2Start)
    throws IllegalArgumentException {
    // should never happen
    if (!twoPlayer && p2Start) {
      throw new IllegalArgumentException("No 2P start in 1P level");
    }
    // do nothing if start already in place
    if (!((!p2Start && start != null && start.equals(location)) ||
	  (p2Start && start2 != null && start2.equals(location)))) {
      if (!p2Start) {
	if (start != null) {
	  cancelStart(start, false);
	}
	start = (location == null) ? null : new Point(location);
      } else {
	if (start2 != null) {
	  cancelStart(start2, true);
	}
	start2 = (location == null) ? null : new Point(location);
      }
    }
  }


  /**
   * Removes start status from a tile (for either of both players)
   * Note: This method only compliments the push()/pop() methods
   */
  private void cancelStart(Point location) {
    if (start != null && location.equals(start)) {
      cancelStart(location, false);
    }
    if (start2 != null && location.equals(start2)) {
      cancelStart(location, true);
    }
  }


  /**
   * Removes start for one player only
   * Note: This method only compliments the push()/pop() methods
   */
  private void cancelStart(Point location, boolean p2Start) {
    if (isStart(location)) {
      if (!hasEqualStarts()) {
	if (tile(location).getHeight() == 0) {
	  setTile(location, null);
	} else {
	  try {
	    ((StartableTile) tile(location)).setStart(false);
	    // should never happen
	  } catch (TileException e) {
	    e.printStackTrace();
	  }
	}
      }
      if (!p2Start) {
	start = null;
      } else {
	start2 = null;
      }
    }
  }


  /**
   * Adds a teleport start and destination pair to list of teleports
   */
  private void addTeleportLocation(Point tpStart, Point destination) {
    if (tpLocations == null) {
      tpLocations = new ArrayList <Point[]> (1);
    }
    Point[] tmp = new Point[2];
    // will not be altered, so can just be referenced
    tmp[0] = new Point(tpStart);
    tmp[1] = new Point(destination);
    tpLocations.add(tmp);
  }

  
  /**
   * Removes teleport pair with the given start position.
   */
  private void removeTeleportLocation(Point tpStart) {
    if (tpLocations != null) {
      for (int i = 0; i < tpLocations.size(); i++) {
	if (tpLocations.get(i)[0].equals(tpStart)) {
	  tpLocations.remove(i);
	  return;
	}
      }
    }
  }


  /**
   * Removes a set of teleports (given their start positions) and
   * changes the affected tile to <code>NORMAL</code> type.
   */
  private void replaceTeleports(ArrayList <Point> tpStarts) {
    try {
      for (int i = 0; i < tpStarts.size(); i++) {
	pushTile(tpStarts.get(i), new NrmTile(1, false), false);
	removeTeleportLocation(tpStarts.get(i));
      }
    } catch (TileException e) {
      e.printStackTrace();
    }
  }
  

  /**
   * returns teleport tile which has given tile as destination
   */
  private ArrayList <Point> findTeleportStarts(Point destination) {
    if (tpLocations != null) {
      ArrayList <Point> tpStarts = new ArrayList <Point> (3);
      for (int i = 0; i < tpLocations.size(); i++) {
	if (tpLocations.get(i)[1].equals(destination)) {
	  tpStarts.add(tpLocations.get(i)[0]);
	}
      }
      if (!tpStarts.isEmpty()) {
	return tpStarts;
      }
    }
    return null;
  }


  /**
   * returns total number of teleport tiles (not piles) to given location
   */
  private int countTeleportTilesTo(Point destination) {
    int total = 0;
    if (tpLocations != null) {
      for (int i = 0; i < tpLocations.size(); i++) {
	if (tpLocations.get(i)[1].equals(destination)) {
	  if (tile(tpLocations.get(i)[0]) != null) {
	    total += tile(tpLocations.get(i)[0]).getHeight();
	  }
	}
      }
    }
    return total;
  }


  /**
   * Unlike push() this method simply sets the given grid location to the
   * specified tile (and removes associated teleport if one existed)
   */
  private void setTile(Point location, Tile t) {
    /* if the tile to be set is a teleport destination, might have to remove
     * if it now becomes an invalid destination
     */
    if (t != null) {
      ArrayList <Point> startPs;
      if ((startPs = findTeleportStarts(location)) != null &&
	  (t.getType() == Tile.ICE || t.getHeight() == 0)) {
	replaceTeleports(startPs);
      }
      // additional checks for teleport tile
      if (t.getType() == Tile.TELEPORT &&
	  !teleportSetupSuccessful(location, t)) {
	return;
      }
    }
    setTile(location.x, location.y, t);
  }

  
  /**
   *processes teleport tile addition - returns true if placement possible
   */
  private boolean teleportSetupSuccessful(Point location, Tile t) {
    Point dest = ((TptTile) t).getDestination();
    if (validCoordinates(dest)) {
      // empty destination
      if (tile(dest) == null) {
	try {
	  setTile(dest, new NrmTile(t.getHeight(), false));
	  // should never happen
	} catch (TileException e) {
	  e.printStackTrace();
	}
	// invalid destination tile
      } else if (tile(dest).getType() == Tile.ICE) {
	try {
	  setTile(dest, new NrmTile(tile(dest).getHeight(), false));
	  // should never happen
	} catch (TileException e) {
	  e.printStackTrace();
	}
	// start tile with 0 height cannot be teleport destination
      } else if (tile(dest).getHeight() == 0) {
	incTileHeight(dest);
      } else {
	/* Make sure destination has minimum number of tiles necessary
	 * to allow all "incoming" teleports to work, unless of course
	 * tile height limit is reached
	 * Have to include current tile as it has not been added
	 * to the teleport array, so countTeleportTilesTo will not
	 * take it into consideration
	 */
	int requiredTileHeight =
	  countTeleportTilesTo(dest) + t.getHeight();
	if (requiredTileHeight > tile(dest).getHeight()) {
	  // only allow addition of teleport if height limit not reached
	  if (requiredTileHeight <= maxTileHeight) {
	    tile(dest).setHeight(requiredTileHeight);
	  } else {
	    return false;
	  }
	}
      }
      // remove old destination before adding new one
      removeTeleportLocation(location);
      addTeleportLocation(location, dest);
      //should never happen
    } else {
      return false;
    }
    return true;
  }


  /**
   * Only adds normal tiles and starts.
   * This method is used in the <code>Challenge</code> class.
   */
  protected boolean pushTile(Point location, boolean isStart) {
    if (tile(location) == null) {
      if (isStart) {
	setStart(location, false);
      }
      try {
	setTile(location.x, location.y, new NrmTile(isStart ? 0 : 1, isStart));
	// should never happen
      } catch (TileException e) {
	e.printStackTrace();
      }
      return true;
    } else {
      if (isStart) {
	if (!isStart(location)) {
	  setStart(location, (start != null));
	  try {
	    ((StartableTile) tile(location)).setStart(true);
	    // uncommented
	    tile(location).decHeight();
	    return true;
	    // should never happen
	  } catch (TileException e) {
	    e.printStackTrace();
	  }
	} else {
	  // dual start
	  setStart(location, true);
	  try {
	    ((StartableTile) tile(location)).setStart(true);
	    // uncommented
	    tile(location).decHeight();
	    return true;
	  } catch (TileException e) {
	    e.printStackTrace();
	  }
	  return true;
	}
      }
      return incTileHeight(location);
    }
  }

  
  /**
   * Checks whether both players have equal starts
   * @return <code>true</code> if both players start at the same location
   */
  public boolean hasEqualStarts() {
    return (start != null && start2 != null && start.equals(start2));
  }


  /**
   * Provides direct access to a tile at the given location
   */
  protected Tile tile(Point location) {
    return tile(location.x, location.y);
  }

  
  /**
   * Provides direct access to a tile at the given location
   */
  protected Tile tile(int x, int y) {
    return tiles[y][x];
  }

  
  /**
   * Sets grid location to the tile specified
   */
  private void setTile(int x, int y, Tile t) {
    tiles[y][x] = t;
  }


  /**
   * increments height of tile, and also destination if teleport tile
   * checks for height limit
   */
  private boolean incTileHeight(Point location) {
    if (tile(location).getHeight() < maxTileHeight) {
      // if a teleport, check destination can be increased, otherwise cancel
      if (tile(location).getType() == Tile.TELEPORT) {
	if (tile(((TptTile) tile(location)).getDestination()).getHeight()
	    < maxTileHeight) {
	  tile(((TptTile) tile(location)).getDestination()).incHeight();
	} else {
	  return false;
	}
      }
      tile(location).incHeight();
      return true;
    } else {
      return false;
    }
  }


  /**
   * checks whether the given tile is a start (can be used with any tile)
   */
  private boolean isStart(Tile t) {
    return (t.isStartable() &&
	    ((StartableTile) t).isStart());
  }


  /**
   * width of level
   */
  private int lvlWidth() {
    return tiles[0].length;
  }


  /**
   * height of level
   */
  private int lvlHeight() {
    return tiles.length;
  }

  
  /**
   * returns true if the coordinates provided fall within limits of this level
   */
  private boolean validCoordinates(Point p) {
    if (p.x < 0 || p.x >= lvlWidth() || p.y < 0 || p.y >= lvlHeight()) {
      return false;
    } else {
      return true;
    }
  }


  /**
   * Direct access to tile array.
   * Used by <code>TilePainter</code> class for quick access during rendering
   */
  protected Tile[][] getTiles() {
    return tiles;
  }

  
  /**
   * Counts number of tiles (not piles) in this level
   * @return the total number of tiles in this level
   */
  public int size() {
    int total = 0;
    
    for (int i = 0; i < tiles.length; i++) {
      for (int j = 0; j < tiles[i].length; j++) {
	if (tiles[i][j] != null) {
	  total += tiles[i][j].getHeight();
	}
      }
    }
    return total;
  }

  
  /**
   * Sets level start message to specified string
   * @param s message to use for level start
   */
  public void setInText(String s) {
    inText = s;
  }


  /**
   * Sets level end message to specified string
   * @param s message to use for level end
   */  public void setOutText(String s) {
    outText = s;
  }


  /**
   * Get the level start message
   * @return the level start message
   */
  public String getInText() {
    return inText;
  }


  /**
   * Get the level end message
   * @return the level end message
   */
  public String getOutText() {
    return outText;
  }


  /**
   * Get the time limit of this level
   * @return time limit in seconds
   */
  public int getTimeLimit() {
    return timeLimit;
  }


  /**
   * Sets time limit for this level
   * @param seconds time limit
   * @throws IllegalArgumentException if either<ul>
   * <li><code>seconds</code> < 0 or</li>
   * <li><code>seconds</code> > 3599 (59m, 59s)</li></ul>
   */
  public void setTimeLimit(int seconds) throws IllegalArgumentException {
    if (seconds < 0 || seconds > 3599) {
      throw new IllegalArgumentException("seconds parameter invalid");
    }
    timeLimit = seconds;
  }

}