/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.IOException;
import java.net.URL;

/**
 * Defines function and rendering of the level chooser component.
 * @author V.Termanis
 */
public class LevelChooser {

  // mouseover states
  public static final int NONE = -1;
  public static final int FIRST = 0;
  public static final int PREVIOUS = 1;
  public static final int NEXT = 2;
  public static final int LAST = 3;
  public static final int PLAY = 4;
  public static final int CLOSE = 5;

  // actions
  public static final int LEVEL_CHANGED = 0;
  public static final int LEVEL_CHOSEN = 1;
  public static final int EXIT = 2;

  // levelsets
  public static final int SINGLE = 0;
  public static final int COOP = 1;
  public static final int CUSTOM = 2;

  private static final Rectangle
    levelBig = new Rectangle(0, 0, 625, 625),
    levelBigOffset = new Rectangle(0, 25, 0, 0),
    levelSmall = new Rectangle(0, 0, 375, 375),
    levelMicro = new Rectangle(0, 0, 125, 125);

  private static final int
    buttonSpacing = 20,
    previewSpacing = 50,
    previewHorizontalOffset = -85, // from middle of screen
    previewBorderOffset = 10,
    previewBorderOffsetSub = 15,
    previewBorderCurvature = 15;

  private static final float previewBorderThickness = 3f;

  private static final Color previewBorderColor = new Color(245, 186, 92);

  // all default levelsets (only single/coop sets will ever be used)
  private LevelSet levelSet[];
  // specifies currently active set/level and limits
  private int levelSetIndex, levelIndex, oldLevelIndex, maxAllowedLevel;
  // buffered level previews
  private BufferedImage[] levelPreview, scaledLevelPreview;
  private TilePainter tPainter;
  private ScreenFrames frames;
  private Images imgData;
  // button locations for quick mouseover lookup
  private Rectangle[] buttonLocations, levelPreviewLocations;
  private Rectangle levelString, previewBorder;
  private Point levelStringOffset;
  private Font txtFont;
  private Color txtColor;
  private int activeButton, pressedButton, oldButton;
  private boolean lastUpdateWasKeyPress, customSetAsTwoPlayer, twoPlayer;
  private Stroke previewBorderStroke;


  /**
   * Create a new <code>LevelChooser</code> instance with the given attributes.
   * @param painter used to draw previews and buttons
   * @param imgData provides access to global resources
   * @param frames used to draw frames
   * @param defaultLevelSets holds single, cooperative, versus and versus
   * alternative levelsets (at most single and cooperative sets are required
   * in that order)
   * @param txtFont font style to use (not for buttons)
   * @param txtColor colour to use for text
   * @param g the graphics context for drawing
   */
  public LevelChooser(TilePainter painter, Images imgData, ScreenFrames frames,
		      LevelSet[] defaultLevelSets, Font txtFont,
		      Color txtColor, Graphics2D g) {
    levelSetIndex = SINGLE;
    this.tPainter = painter;
    this.imgData = imgData;
    this.frames = frames;
    levelSet = defaultLevelSets;
    this.txtFont = txtFont;
    this.txtColor = txtColor;
    activeButton = pressedButton = oldButton = NONE;
    lastUpdateWasKeyPress = customSetAsTwoPlayer = twoPlayer = false;
    reset();
    calculateGeometry(g);
    levelPreview = new BufferedImage[3];
    for (int i = 0; i < 3; i++) {
      levelPreview[i] = new BufferedImage(levelBig.width, levelBig.height,
					  BufferedImage.TYPE_INT_ARGB);
    }
    scaledLevelPreview = new BufferedImage[3];
    scaledLevelPreview[0] = new BufferedImage(levelMicro.width,
					      levelMicro.height,
					      BufferedImage.TYPE_INT_ARGB);
    scaledLevelPreview[1] = new BufferedImage(levelSmall.width,
					      levelSmall.height,
					      BufferedImage.TYPE_INT_ARGB);
    scaledLevelPreview[2] = new BufferedImage(levelMicro.width,
					      levelMicro.height,
					      BufferedImage.TYPE_INT_ARGB);
    this.previewBorderStroke = new BasicStroke(previewBorderThickness);
  }


  /**
   * Specify a levelset which can be previewed in <code>CUSTOM</code> mode.
   * @param ls the levelset to load
   * @param twoPlayer if <code>false</code> then two player levelsets are
   * rendered without 2nd start (when choosing to play 2P levelset in 1P mode)
   * @throws NullPointerException if <code>ls</code> is <code>null</code>.
   */
  public void setCustomLevelSet(LevelSet ls, boolean twoPlayer)
    throws NullPointerException {
    if (ls == null) {
      throw new NullPointerException("ls is null");
    }
    levelSet[2] = ls;
    customSetAsTwoPlayer = twoPlayer;
  }


  /**
   * Specify which levelset to display.
   * @param type one of <code>SINGLE</code>, <code>COOP</code> or <code>CUSTOM
   * </code>.
   * @throws IllegalArgumentException if type is invalid
   * @throws NullPointerException if <code>CUSTOM</code> type has been chosen
   * without previously choosing a custom set via the <code>setCustomLevelSet
   * </code> method.
   */
  public void setLevelSet(int type)
    throws IllegalArgumentException, NullPointerException {
    if (levelSetIndex < 0 || levelSetIndex > 2) {
      throw new IllegalArgumentException("Invalid levelset index");
    }
    if (levelSetIndex == 2 && levelSet[2] == null) {
      throw new NullPointerException("Custom levelset does not exist");
    }
    levelSetIndex = type;
    reset();
  }


  public void drawAll(Graphics2D g) {
    frames.drawAll(true, g);
    frames.setRenderingHints(g, true);
    Geom.drawRoundRect(previewBorder, previewBorderCurvature,
		       previewBorderColor, previewBorderStroke, g);
    frames.setRenderingHints(g, false);
    drawPreviews(g);
    drawButtons(g);
    g.setFont(txtFont);
    Geom.drawString("Level " + Integer.toString(levelIndex + 1),
		    levelString.getLocation(), levelStringOffset, txtColor, g);
  }


  /**
   * Generates updated level previews if highlighted level has changed
   */
  private void getPreviews() {
    boolean requireNext = true, requireCurrent = true, requirePrevious = true;
    BufferedImage tmp;

    if (levelIndex != oldLevelIndex) {
      // special case - only one level needs to be rendered
      if (levelIndex == (oldLevelIndex - 1)) {
	tmp = levelPreview[2];
	levelPreview[2] = levelPreview[1];
	levelPreview[1] = levelPreview[0];
	levelPreview[0] = tmp;
	requireNext = false;
	requireCurrent = false;
      } else if (levelIndex == (oldLevelIndex + 1)) {
	tmp = levelPreview[0];
	levelPreview[0] = levelPreview[1];
	levelPreview[1] = levelPreview[2];
	levelPreview[2] = tmp;
	requirePrevious = false;
	requireCurrent = false;
      }

      // previous
      if (requirePrevious && levelIndex > 0) {
	levelPreview[0] = getPreview(levelPreview[0], levelIndex - 1);
      }
      // current
      if (requireCurrent) {
	levelPreview[1] = getPreview(levelPreview[1], levelIndex);
      }
      // next
      if (requireNext && levelIndex < maxAllowedLevel) {
	levelPreview[2] = getPreview(levelPreview[2], levelIndex + 1);
      }
      // next time will not get previews if they have not changed
      oldLevelIndex = levelIndex;
    }
  }


  /**
   * Generates level preview to be rendered to screen later
   */
  private BufferedImage getPreview(BufferedImage target, int level) {
    Graphics2D g = target.createGraphics();

    g.setComposite(AlphaComposite.Clear);
    g.fill(levelBig);
    g.setComposite(AlphaComposite.SrcOver);
    
    // also show special tile in preview
    if (level == GamePlayer.MEDIAMERGE_LEVEL && levelSetIndex == 0) {
      imgData.toggleMediaMerge();
    }
    Level l = levelSet[levelSetIndex].getLevel(level);
    tPainter.drawTiles(l.getTiles(), levelBigOffset, frames.getTileSize(),
		       null, null, null, l.getStart2(), twoPlayer,
		       l.hasEqualStarts(), false, g);
    if (level == GamePlayer.MEDIAMERGE_LEVEL && levelSetIndex == 0) {
      imgData.toggleMediaMerge();
    }
    return target;
  }


  public void drawUpdate(Graphics2D g) {
    if (oldButton != activeButton) {
      drawButton(activeButton, true, g);
      drawButton(oldButton, true, g);
    }
  }
  
  
  /**
   * Renders all three previews to screen
   */
  private void drawPreviews(Graphics2D g) {
    getPreviews();

    // previous
    if (levelIndex > 0) {
      drawPreview(0, g);
    }
    // current
    drawPreview(1, g);
    // next
    if (levelIndex < maxAllowedLevel) {
      drawPreview(2, g);
    }
  }


  /**
   * Scales and draws a single preview to screen
   */
  private void drawPreview(int index, Graphics2D g) {
    Graphics2D scaledLevel = scaledLevelPreview[index].createGraphics();

    scaledLevel.setComposite(AlphaComposite.Clear);
    scaledLevel.fill(levelSmall);
    scaledLevel.setComposite(AlphaComposite.SrcOver);
    scaledLevel.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
				 RenderingHints.VALUE_INTERPOLATION_BILINEAR);
    scaledLevel.drawImage(levelPreview[index], 0, 0,
			  levelPreviewLocations[index].width,
			  levelPreviewLocations[index].height, null);
    Geom.drawImage(scaledLevelPreview[index],
		   levelPreviewLocations[index], null, g);
  }


  /**
   * Draws all buttons part of level chooser
   */
  private void drawButtons(Graphics2D g) {
    for (int i = 0; i < buttonLocations.length; i++) {
      drawButton(i, (i == activeButton), g);
    }
  }


  /**
   * Draws a single level chooser button
   */
  private void drawButton(int buttonIndex, boolean active, Graphics2D g) {
    // level first/previous/next/last buttons
    if (buttonIndex < 4) {
      g.drawImage(imgData.getImage(Images.O_3D, active), null,
		  buttonLocations[buttonIndex].x,
		  buttonLocations[buttonIndex].y + 6);
      if (!active) {
	g.drawImage(imgData.getImage(Images.O_3D, active), null,
		    buttonLocations[buttonIndex].x + 5,
		    buttonLocations[buttonIndex].y + 1);
      }
      g.drawImage(imgData.getImage(Images.O_FIRST + buttonIndex, active),
		  null, buttonLocations[buttonIndex].x + (active ? 5 : 10),
		  buttonLocations[buttonIndex].y + (active ? 5 : 0));
      // play & back buttons
    } else {
      tPainter.drawMenuString(((buttonIndex == 4) ? "PLAY" : "BACK"),
			       !active, buttonLocations[buttonIndex], g);
    }
  }
  

  /**
   * Initialises/resets level chooser state
   */
  private void reset() {
    if (levelSetIndex != CUSTOM) {
      maxAllowedLevel =	ProgressHandler.getReachedLevel((levelSetIndex == 0) ?
							Progress.SINGLE :
							Progress.COOP);
      /* make sure index loaded from progress file is not out of range
       * this could happen if one of the default levelsets is shortened
       */
      if (maxAllowedLevel >= levelSet[levelSetIndex].getSize()) {
	maxAllowedLevel = levelSet[levelSetIndex].getSize() - 1;
      }
      levelIndex = maxAllowedLevel;
      twoPlayer = (levelSetIndex == COOP);
    } else {
      maxAllowedLevel = levelSet[CUSTOM].getSize() - 1;
      levelIndex = 0;
      twoPlayer = customSetAsTwoPlayer;
    }
      // do not set to the same as current level index so update forced
      oldLevelIndex = -2;
  }


  public Rectangle updateUI(Point mouseLocation) {
    lastUpdateWasKeyPress = false;

    if (mouseLocation == null) {
      return null;
    }
    oldButton = activeButton;
    activeButton = NONE;

    // determine current position
    for (int i = 0; i < buttonLocations.length; i++) {
      if (buttonLocations[i].contains(mouseLocation)) {
	activeButton = i;
	break;
      }
    }

    // pressed button - only ever highlight original mouse down button
    if (pressedButton != NONE && activeButton != pressedButton) {
      activeButton = NONE;
    }

    // calculate clip area
    if (oldButton == activeButton) {
      return null;
    } else {
      // [] -> button
      if (oldButton == NONE) {
	return buttonLocations[activeButton];
	// button -> []
      } else if (activeButton == NONE) {
	return buttonLocations[oldButton];
	// button -> button
      } else {
	return buttonLocations[activeButton].union(buttonLocations[oldButton]);
      }
    }
  }


  public int clickResult(Point mouseLocation) {
    int retval = NONE;

    if (pressedButton != NONE && pressedButton == activeButton) {
      switch(pressedButton) {
      case PREVIOUS:
	if (changeActiveLevel(-1)) {
	  retval = LEVEL_CHANGED;
	}
	break;
      case FIRST:
	if (jumpToStart()) {
	  retval = LEVEL_CHANGED;
	}
	break;
      case NEXT:
	if (changeActiveLevel(1)) {
	  retval = LEVEL_CHANGED;
	}
	break;
      case LAST:
	if (jumpToEnd()) {
	  retval = LEVEL_CHANGED;
	}
	break;
      case PLAY:
	retval = PLAY;
	break;
      case CLOSE:
	retval = CLOSE;
      }
    }
    pressedButton = NONE;
    return retval;
  }


  /**
   * Returns currently highlighted level
   */
  public int getChosenLevel() {
    return levelIndex;
  }


  public void mousePressed(int button, Point mouseLocation) {
    if (lastUpdateWasKeyPress) {
      updateUI(mouseLocation);
    }
    if (button == MouseEvent.BUTTON1) {
      pressedButton = activeButton;
    }
  }


  public int keyResult(int keyCode) {
    lastUpdateWasKeyPress = true;

    switch (keyCode) {
    case KeyEvent.VK_UP:
      if (changeActiveLevel(10)) {
	return LEVEL_CHANGED;
      } else {
	return NONE;
      }
    case KeyEvent.VK_DOWN:
      if (changeActiveLevel(-10)) {
	return LEVEL_CHANGED;
      } else {
	return NONE;
      }
    case KeyEvent.VK_LEFT:
      if (changeActiveLevel(-1)) {
	return LEVEL_CHANGED;
      } else {
	return NONE;
      }
    case KeyEvent.VK_RIGHT:
      if (changeActiveLevel(1)) {
	return LEVEL_CHANGED;
      } else {
	return NONE;
      }
    case KeyEvent.VK_ENTER:
      return PLAY;
    case KeyEvent.VK_ESCAPE:
    case KeyEvent.VK_F9:
      return CLOSE;
    default:
      return NONE;
    }
  }

  
  /**
   * Change active level with respect to currently active level
   */
  private boolean changeActiveLevel(int changeDelta) {
    if ((levelIndex == 0 && changeDelta < 0) ||
	levelIndex == maxAllowedLevel && changeDelta > 0) {
      return false;
    } else {
      setLevelIndex(levelIndex + changeDelta);
      return true;
    }
  }


  /**
   * Highlights first level in set
   */
  private boolean jumpToStart() {
    if (levelIndex > 0) {
      setLevelIndex(0);
      return true;
    } else {
      return false;
    }
  }


  /**
   * Highlights last level in set
   */
  private boolean jumpToEnd() {
    if (levelIndex < maxAllowedLevel) {
      setLevelIndex(maxAllowedLevel);
      return true;
    } else {
      return false;
    }
  }


  /**
   * Sets highlighted level
   */
  private void setLevelIndex(int level) {
    oldLevelIndex = levelIndex;
    levelIndex = level;
    if (levelIndex < 0) {
      levelIndex = 0;
    } else if (levelIndex > maxAllowedLevel) {
      levelIndex = maxAllowedLevel;
    }
  }


  /**
   * Works out locations and dimensions of all levelchooser components
   */
  private void calculateGeometry(Graphics2D g) {
    levelPreviewLocations = new Rectangle[3];
    buttonLocations = new Rectangle[6];

    // level number
    Rectangle tBounds = Geom.calcTextBounds("Level 99", txtFont, g);
    levelStringOffset = tBounds.getLocation();
    levelString =
      new Rectangle(frames.getDrawArea().x +
		    ((frames.getDrawArea().width - tBounds.width) / 2),
		    frames.getCtrlOffset().y + (frames.getBorderSpacing() / 2),
		    tBounds.width, tBounds.height);

    // main preview
    levelPreviewLocations[1] =
      new Rectangle(frames.getDrawArea().x +
		    ((frames.getDrawArea().width - levelSmall.width) / 2),
		    levelString.y + levelString.height +
		    (frames.getBorderSpacing() / 2),
		    levelSmall.width, levelSmall.height);
    previewBorder =
      new Square(levelPreviewLocations[1].x - previewBorderOffset,
		 levelPreviewLocations[1].y - previewBorderOffset +
		 previewBorderOffsetSub,
		 levelPreviewLocations[1].width +
		 (2 * previewBorderOffset) - previewBorderOffsetSub);
    

    // previous preview
    levelPreviewLocations[0] =
      new Rectangle(levelPreviewLocations[1].x -
		    (previewSpacing + levelMicro.width),
		    ((frames.getDrawArea().height - levelMicro.height) / 2) +
		    previewHorizontalOffset,
		    levelMicro.width, levelMicro.height);
    
    // next preview
    levelPreviewLocations[2] =
      new Rectangle(levelPreviewLocations[1].x +
		    levelPreviewLocations[1].width + previewSpacing,
		    levelPreviewLocations[0].y,
		    levelMicro.width, levelMicro.height);
    

    int buttonSize = frames.getTileSize() + 10; // +10 for 3D effect
    int buttonRowWidth = (buttonSize * 4) + (buttonSpacing * 3);


    // <<, <, >, and >> buttons
    buttonLocations[0] =
      new Square(levelPreviewLocations[1].x +
		 ((levelPreviewLocations[1].width - buttonRowWidth) / 2),
		 levelPreviewLocations[1].y + levelSmall.height +
		 buttonSpacing,
		 buttonSize);

    for (int i = 1; i < buttonLocations.length - 1; i++) {
      buttonLocations[i] =
	new Square(buttonLocations[i - 1].x + buttonLocations[i - 1].width +
		   buttonSpacing,
		   buttonLocations[i - 1].y,
		   buttonSize);
    }

    buttonLocations[4] =
      new Rectangle(frames.getDrawArea().x +
		    ((frames.getDrawArea().width -
		      Geom.getTileButtonWidth("Play", true, frames)) / 2),
		    buttonLocations[0].y + buttonLocations[0].height +
		    buttonSpacing,
		    Geom.getTileButtonWidth("Play", true, frames),
		    (frames.getTileSize() + 10));

    buttonLocations[5] =
      new Rectangle(buttonLocations[4].x,
		    buttonLocations[4].y + buttonLocations[4].height +
		    buttonSpacing,
		    buttonLocations[4].width,
		    buttonLocations[4].height);
  }

}