/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.Point;

/**
 * Handles gameplay mechanics of individual levels. Note: This class only
 * understands two modes: single and two-player. (The <code>GamePlayer</code>
 * class deals with game mode specifics.)
 * @author V.Termanis
 */
public class LevelPlayer {

  /**
   * <code>moveOne</code> return code indicating succesful move
   */
  public static final int MOVED = 0;
  /**
   * <code>moveOne</code> return code indicating the level has been completed
   */
  public static final int FINISHED = 1;
  /**
   * <code>moveOne</code> return code indicating the move was invalid
   */
  public static final int INVALID_MOVE = 2;
  /**
   * <code>moveOne</code> return code indicating no more moves are available
   */
  public static final int LOST = 3;
  /**
   * <code>moveOne</code> return code indicating the just moved player is stuck
   */
  public static final int STUCK = 4;

  // [y][x]
  private GameTile[][] board;
  private int tilesLeft, clearedByP1, clearedByP2;
  private Point p1Position, p2Position, start, start2;
  // notStarted - meaning at least one valid move has been made
  private boolean notStarted, playerTwoActive, twoPlayerMode, p1Dead, p2Dead;
  // just for P1 normal tutorial level 11
  protected boolean teleportDeath;


  /**
   * Creates a new <code>LevelPlayer</code> instance with the given level.
   * @param level the level to load
   * @throws NullPointerException if the parameter is <code>null</code>
   */
  public LevelPlayer(Level level)
    throws NullPointerException {
    setLevel(level, false);
  }


  /**
   * Creates a new <code>LevelPlayer</code> instance with the given level and
   * in either single or two player mode.
   * @param level the level to load
   * @param twoPlayer indicate whether this should be a two player level or not
   * @throws NullPointerException if the parameter is <code>null</code>
   * @throws IllegalArgumentException if two player mode has been requested yet
   * the supplied level is a single player level
   */  
  public LevelPlayer(Level level, boolean twoPlayer)
    throws NullPointerException, IllegalArgumentException {
    setLevel(level, twoPlayer);
  }

  
  /**
   * Loads the provided level.
   * @param level the level to load
   * @throws NullPointerException if the paramter is <code>null</code>
   */
  public void setLevel(Level level)
    throws NullPointerException {
    setLevel(level, false);
  }


  /**
   * Loads the provided level in either single or two-player mode.
   * @param level the level to load
   * @param twoPlayer indicate whether this should be a two player level or not
   * @throws NullPointerException if the paramter is <code>null</code>
   * @throws IllegalArgumentException if two player mode has been requested yet
   * the supplied level is a single player level
   */
  public void setLevel(Level level, boolean twoPlayer)
    throws NullPointerException, IllegalArgumentException {
    setLevel(level, twoPlayer, false);
  }


  /**
   * Loads the provided level in either single or two-player mode.
   * @param level the level to load
   * @param twoPlayer indicate whether this should be a two player level or not
   * @param swappedStarts exchanges first and second player starting positions.
   * Note: This has no effect if the <code>twoPlayer</code> parameter is not
   * set to <code>true</code>
   * @throws NullPointerException if the paramter is <code>null</code>
   * @throws IllegalArgumentException if two player mode has been requested yet
   * the supplied level is a single player level
   */
  public void setLevel(Level level, boolean twoPlayer, boolean swappedStarts)
    throws NullPointerException, IllegalArgumentException {
    if (level == null) {
      throw new NullPointerException("level is null");
    }
    if (twoPlayer == true && level.getStart2() == null) {
      throw new IllegalArgumentException("Cannot play single player level in" +
					 " two player mode.");
    }
    twoPlayerMode = twoPlayer;
    start = level.getStart();
    if (twoPlayerMode) {
      if (swappedStarts) {
	start = level.getStart2();
	start2 = level.getStart();
      } else {
	start2 = level.getStart2();
      }
    } else {
      start2 = null;
    }
    board = new GameTile[level.getTiles().length][];
    for (int y = 0; y < board.length; y++) {
      board[y] = new GameTile[level.getTiles()[0].length];
      for (int x = 0; x < board[y].length; x++) {
	if (level.tile(x, y) != null) {
	  board[y][x] = new GameTile(level.tile(x, y));
	}
      }
    }
    reset();
  }


  /**
   * Moves active player by one in the given direction.
   * @param direction e.g <code>UP</code> or <code>GamePlayer.AUTOSTART</code>
   * which is used to start a level (and perform initial automatic moves if
   * there are any, one by one)
   * @return see <A HREF="#field_summary">field summary</code>
   */
  public int moveOne(int direction) {
    if (direction == GamePlayer.AUTOSTART && notStarted) {
      notStarted = false;
    }
    
    if (activePlayerDead()) {
      return INVALID_MOVE;
    }

    if (!validMove(direction) &&
	tileType(getActivePosition()) != Tile.TELEPORT) {
      if (gravityPull() == -2) {
	markActivePlayerAsDead();
	return otherPlayerDead() ? LOST : STUCK;
      } else if (gravityPull() < 0) {
	return INVALID_MOVE;
      }
    }

    switch(tileType(getActivePosition())) {
    case Tile.NORMAL:
    case Tile.ARROW:
    case Tile.ICE:
    case Tile.GRAVITY:
      int pullDirection;
      // gravity takes priority over ice
      if ((pullDirection = gravityPull()) >= 0) {
	setActiveLocation(adjacentTile(pullDirection));
	gameTile(getActivePosition()).decHeight();
      } else {
	setActiveLocation(adjacentTile(direction));
	gameTile(getActivePosition()).decHeight();
      }
      break;
    case Tile.TELEPORT:
      if (tilesLeft > 0) {
	Point tpDestination =
	  TptTile.getDestination(tile(getActivePosition()));
	/* move to destination regardless of whether it is valid
	 * (so can indicate where player died)
	 */
	setActiveLocation(tpDestination);
	if (gameTile(tpDestination).getHeight() > 0) {
	  gameTile(getActivePosition()).decHeight();
	} else {
	  markActivePlayerAsDead();
	  return otherPlayerDead() ? LOST : STUCK;
	}
      }
    }
    decTiles();
    if (tilesLeft == 0) {
      return FINISHED;
    } else {
      // additional life loss scenarios
      switch (tileType(getActivePosition())) {
	// teleport to nothing
      case Tile.TELEPORT:
	Point destination = TptTile.getDestination(tile(getActivePosition()));
	if (gameTile(destination).getHeight() == 0) {
	  // show where offending destination is to player
	  setActiveLocation(destination);
	  gameTile(getActivePosition()).disable();
	  markActivePlayerAsDead();
	  /* only required for tutorial level
	   * (see CUSTOM_DEATH_MESSAGE_LEVEL in GamePlayer)
	   */
	  teleportDeath = true;
	  return otherPlayerDead() ? LOST : STUCK;
	}
	break;
	// slide to nothing
      case Tile.ICE:
	if (!validMove(direction) && gravityPull() < 0) {
	  markActivePlayerAsDead();
	  // other player might have been killed by this player's move
	  if (!otherPlayerDead() && !canStillMove(!playerTwoActive)) {
	    markOtherPlayerAsDead();
	  }
	  return (otherPlayerDead() ? LOST : STUCK);
	}
      }
      //death by gravity
      if (gravityTileCount() == 2 && !oppositeGravity()) {
	markActivePlayerAsDead();
	return otherPlayerDead() ? LOST : STUCK;
      } else {
	boolean newDeath = false;
	// one or both players out of moves?
	if (!canStillMove(playerTwoActive)) {
	  newDeath = true;
	  markActivePlayerAsDead();
	}
	if (!otherPlayerDead() && !canStillMove(!playerTwoActive)) {
	  newDeath = true;
	  markOtherPlayerAsDead();
	}
	// return value decision based on whether players are alive
	if (activePlayerDead() && otherPlayerDead()) {
	  return LOST;
	} else if (newDeath) {
	  return STUCK;
	} else {
	  return MOVED;
	}
      }
    }
  }


  /**
   * Returns whether the active player has automatic moves left given the
   * last direction they moved in
   * @param direction e.g. <code>UP</code>
   * @return <code>true</code> if there are more automatic moves left
   */
  public boolean automaticMovesLeft(int direction) {
    boolean retVal =
     (!activePlayerDead() && tilesLeft > 0 && gravityPull() != -2 &&
	    ((tileType(getActivePosition()) == Tile.ICE &&
	      validMove(direction)) ||
	     (tileType(getActivePosition()) == Tile.TELEPORT) ||
	     (gravityPull() >= 0 &&
	      tileType(getActivePosition()) != Tile.TELEPORT)));
    return retVal;
  }

  
  /**
   * Determines whether force exists for dislocating current position
   * Note: Takes direction limiting (arrow) tiles into account
   */
  private int gravityPull() {
    switch(gravityTileCount()) {
      // NO PULL
    case 0:
    case 4:
      return -1;
      /* POSSIBLY death by gravity
       * (if next to each other and not opposite)
       */
    case 2:
      return (oppositeGravity() ? -1 : -2);
      // DIRECTIONAL PULL
    case 1:
      // work out which of the four surrounding tiles is active
      for (int i = 0; i < 4; i++) {
	if (isGravity(i)) {
	  // not possible to have an invalid move at this point?
	  return (validMove(i) ? i : -1);
	}
      }
      break;
    case 3:
      // work out which of the four surrounding tiles is NOT active
      for (int i = 0; i < 4; i++) {
	if (!isGravity(i)) {
	  return (validMove(oppositeDirection(i)) ?
		  oppositeDirection(i) : -1);
	}
      }
    }
    /* should never get this far (as gravityTileCount return value should be
     * between 0 and 4 inclusive.
     */
    return -1;
  }


  /**
   * Are UP+DOWN and/or LEFT+RIGHT gravity?
   */
  private boolean oppositeGravity() {
    return ((isGravity(GamePlayer.UP) && isGravity(GamePlayer.DOWN)) ||
	    (isGravity(GamePlayer.LEFT) && isGravity(GamePlayer.RIGHT)));
  }


  /** total number of gravity tiles around currently active tile
   *  Note: Does not consider diagonal neighbours
   */
  private int gravityTileCount() {
    int total = 0;

    for (int i = 0; i < 4; i++) {
      if (isGravity(i)) {
	total++;
      }
    }
    return total;
  }



  /**
   * Returns whether an adjacent tile is a gravity tile.
   * Note: will return false if of height zero
   */
  private boolean isGravity(int direction) {
    Point dirPoint = adjacentTile(direction);

    return (dirPoint != null && validCoordinates(dirPoint) &&
	    gameTile(dirPoint) != null &&
	    gameTile(dirPoint).getHeight() > 0 &&
	    tileType(dirPoint) == Tile.GRAVITY);
  }


  /**
   * Resets currently loaded level back to starting position.
   */
  public void reset() {
    tilesLeft = 0;
    for (int y = 0; y < board.length; y++) {
      for (int x = 0; x < board[0].length; x++) {
	if (gameTile(x, y) != null) {
	  gameTile(x, y).reset();
	  tilesLeft += gameTile(x, y).getHeight();
	}
      }
    }
    p1Position = start;
    p2Position = start2;
    notStarted = true;
    playerTwoActive = p1Dead = p2Dead = teleportDeath = false;
    clearedByP1 = clearedByP2 = 0;
  }


  /**
   * Calculates whether the player can still move or not
   * @param playerTwo indicate whether to calculate for first or second player
   * @return <code>true</code> if there are still moves available
   */
  protected boolean canStillMove(boolean playerTwo) {
    boolean swapPlayers = (playerTwo != playerTwoActive);

    if (swapPlayers) {
      playerTwoActive = !playerTwoActive;
    }

    boolean result = !activePlayerDead() &&
      (tileType(getActivePosition()) == Tile.TELEPORT ||
       validMove(GamePlayer.UP) || validMove(GamePlayer.DOWN)
       || validMove(GamePlayer.LEFT) || validMove(GamePlayer.RIGHT));

    if (swapPlayers) {
      playerTwoActive = !playerTwoActive;
    }

    return result;
  }
  

  /**
   * Indicates whether moving in a particular direction (from the currently
   * active tile) is a valid move
   */
  private boolean validMove(int direction) {
    Point newLocation = adjacentTile(direction);

    if (newLocation != null && validCoordinates(newLocation) &&
	gameTile(newLocation) != null) {
      if (gameTile(newLocation).getHeight() > 0 &&
	  ((tileType(getActivePosition()) == Tile.ARROW &&
	    (((ArwTile) tile(getActivePosition())).allowsDirection(direction)))
	   || tileType(getActivePosition()) != Tile.ARROW)) {
	return true;
      }
    }
    return false;
  }


  /**
   * Returns coordinates of adjacent tile given a direction to move in
   */
  private Point adjacentTile(int direction) {
    Point newLocation = new Point(getActivePosition());
    switch(direction) {
    case GamePlayer.UP:
      newLocation.translate(0, -1);
      break;
    case GamePlayer.DOWN:
      newLocation.translate(0, 1);
      break;
    case GamePlayer.LEFT:
      newLocation.translate(-1, 0);
      break;
    case GamePlayer.RIGHT:
      newLocation.translate(1, 0);
      break;
    default:
      newLocation = null;
    }
    return newLocation;
  }

  
  /**
   * E.g. UP -> DOWN or LEFT -> RIGHT
   */
  private int oppositeDirection(int direction)
    throws IllegalArgumentException {
    switch(direction) {
    case GamePlayer.UP:
      return GamePlayer.DOWN;
    case GamePlayer.DOWN:
      return GamePlayer.UP;
    case GamePlayer.LEFT:
      return GamePlayer.RIGHT;
    case GamePlayer.RIGHT:
      return GamePlayer.LEFT;
    }
    // should never get this far
    throw new IllegalArgumentException("Invalid direction");
  }


  /**
   * Determines whether a set of coordinates are actually on the game board
   */
  private boolean validCoordinates(Point p) {
    if (p.x < 0 || p.x >= board[0].length || p.y < 0 || p.y >= board.length) {
      return false;
    } else {
      return true;
    }
  }


  /**
   * Sets active player location (on game board)
   */
  private void setActiveLocation(Point p) {
    if (p != null) {
      if (playerTwoActive) {
	p2Position = p;
      } else {
	p1Position = p;
      }
      // should never happen
    } else {
      throw new NullPointerException("p is null");
    }
  }


  /**
   * Set the active player.
   * @param secondPlayer indicate whether Player 2 should be active or not
   */
  public void setActivePlayerTo(boolean secondPlayer) {
    playerTwoActive = secondPlayer;
  }

  
  /**
   * Indicates whether the second or first player is currently active.
   * @return <code>true</code> if Player 2 is active
   */
  public boolean secondPlayerActive() {
    return playerTwoActive;
  }


  /**
   * Reduces counter of tiles left for active player.
   * Note: It would of course be possible to simply count all the tiles left
   * on the board after each move but that is highy inefficient.
   */
  private void decTiles() {
    tilesLeft--;
    if (playerTwoActive) {
      clearedByP2++;
    } else {
      clearedByP1++;
    }
  }


  /**
   * Returns the type of a tile with the given coordinates.
   * See the <code>Tile</code> for a list of possible codes. Will return <code>
   * -1</code> if there is no tile at the provided location.
   */
  private int tileType(Point p) {
    if (gameTile(p) != null) {
      return gameTile(p).getType();
    } else {
      return -1;
    }
  }


  /**
   * Returns the underlying tile (without <code>GameTile</code> wrapper) for
   * the the given coordinates (or <code>null</code> if there is no tile at
   * this location)
   */
  private Tile tile(Point p) {
    if (gameTile(p) != null) {
      return gameTile(p).getTile();
    } else {
      return null;
    }
  }


  /**
   * Returns a specific tile with the given coordinates (or <code>null</code>
   * if there is no tile at this location)
   */
  private GameTile gameTile(Point p) {
    return gameTile(p.x, p.y);
  }


  /**
   * Returns a specific tile with given coordinates (or <code>null</code> if
   * there is no tile at this location)
   */
  private GameTile gameTile(int x, int y) {
    return board[y][x];
  }

  
  /**
   * Used by <code>GamePlayer</code> to directly access the game board (for
   * rendering)
   * @return the game board
   */
  protected GameTile[][] getGameTiles() {
    return board;
  }


  /**
   * Flags currently active player as dead
   * @return <code>LOST</code> if both players are dead or <code>STUCK</code>
   * the other player can still move
   */
  public int setActivePlayerDead() {
    markActivePlayerAsDead();
    if (otherPlayerDead()) {
      return LOST;
    } else {
      return STUCK;
    }
  }


  /**
   * Indicates whether the currently active player is dead or not
   * @return <code>true</code> if the active player is dead
   */
  public boolean activePlayerDead() {
    return playerTwoActive ? p2Dead : p1Dead;
  }


  /**
   * Indicates whether the currently inactive player is dead or not
   */
  private boolean otherPlayerDead() {
    return twoPlayerMode ? (playerTwoActive ? p1Dead : p2Dead) : true;
  }


  /**
   * Flags currently inactive player as dead
   */
  private void markOtherPlayerAsDead() {
    if (twoPlayerMode) {
      playerTwoActive ^= true;
      markActivePlayerAsDead();
      playerTwoActive ^= true;
    }
  }


  /**
   * Flags the currently active player as dead
   */
  private void markActivePlayerAsDead() {
    gameTile(getActivePosition()).disable();
    if (playerTwoActive) {
      p2Dead = true;
    } else {
      p1Dead = true;
    }
  }

  
  /**
   * Returns position of currently active player
   * @return x and y coordinates of player's position
   */
  public Point getActivePosition() {
    return playerTwoActive ? p2Position : p1Position;
  }


  /**
   * Returns the grid position of player 1
   * @return x and y coordinates of player's position
   */
  public Point getP1Position() {
    return p1Dead ? null : p1Position;
  }


  /**
   * Returns the grid position of player 2
   * @return x and y coordinates of player's position
   */
  public Point getP2Position() {
    return p2Dead ? null : p2Position;
  }


  /**
   * Returns how many tiles are left in the level
   * @return the number of tiles left
   */
  public int tilesLeft() {
    return tilesLeft;
  }


  /**
   * Returns number of tiles cleared by player 1
   * @return number of tiles cleared by player 1
   */
  public int tilesClearedByP1() {
    return clearedByP1;
  }


  /**
   * Returns number of tiles cleared by player 2
   * @return number of tiles cleared by player 2
   */
  public int tilesClearedByP2() {
    return clearedByP2;
  }


  /**
   * Returns whether a level has been started. Or in other words whether at
   * least one move, valid or invalid, has been made.
   * @return <code>true</code> if the level has been started
   */
  public boolean hasNotStarted() {
    return notStarted;
  }

}