/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.util.ArrayList;
import java.net.URL;
import java.io.*;

/**
 * Deals with handling and storage of a collection of levels.
 * @author V.Termanis
 */
public class LevelSet implements Serializable {

  /* had to be introduced after including getInvalidLevelsIndices() method
   * (in v0.974) to remain backwards compatible (with v0.97 onwards)
   */
  static final long serialVersionUID = -9071795372727729115L;

  // size to which ArrayList for storing levels is initialised to
  private static final int INITIAL_LEVEL_ARRAY_SIZE = 3;
  
  private String name;
  private ArrayList <Level> levels;
  private int levelWidth, levelHeight, tileHeightLimit;
  /* userMade indicates whether this level was save in development mode thus
   * making it read-only in user (default) mode
   */
  private boolean twoPlayer, userMade;
 
  /**
   * Constructs a new single-player <code>LevelSet</code> instance with the
   * given attributes.
   * @param name name of the levelset
   * @param levelWidth width in tiles of levels stored in this set
   * @param levelHeight height in tiles of levels stored in this set
   * @param tileHeightLimit maximum height of any one pile of tiles
   * @throws IllegalArgumentException if any of the following are satisfied:
   * <ul>
   * <li><code>levelWidth</code> < 2</li>
   * <li><code>levelHeight</code> < 2</li>
   * <li><code>tileHeightLimit</code> < 1</li></ul>
   * @throws NullPointerException if <code>name</code> is <code>null</code>
   */
  public LevelSet(String name, int levelWidth, int levelHeight,
		  int tileHeightLimit)
    throws IllegalArgumentException, NullPointerException {
    initLevelSet(name, levelWidth, levelHeight, tileHeightLimit, false);
  }
  

  /**
   * Constructs a <code>LevelSet</code> instance with the given attributes.
   * @param name name of the levelset
   * @param levelWidth width in tiles of levels stored in this set
   * @param levelHeight height in tiles of levels stored in this set
   * @param tileHeightLimit maximum height of any one pile of tiles
   * @param twoPlayer defines whether this will be a single or two-player
   * levelset
   * @throws IllegalArgumentException if any of the following are satisfied:
   * <ul>
   * <li><code>levelWidth</code> < 2</li>
   * <li><code>levelHeight</code> < 2</li>
   * <li><code>tileHeightLimit</code> < 1</li></ul>
   * @throws NullPointerException if <code>name</code> is <code>null</code>
   */
  public LevelSet(String name, int levelWidth, int levelHeight,
		  int tileHeightLimit, boolean twoPlayer)
    throws IllegalArgumentException, NullPointerException {
    initLevelSet(name, levelWidth, levelHeight, tileHeightLimit, twoPlayer);
  }

  
  /**
   * Used by all constructors to validate and set the level attributes.
   */
  private void initLevelSet(String name, int levelWidth, int levelHeight,
			    int tileHeightLimit, boolean twoPlayer)
    throws IllegalArgumentException, NullPointerException {
    if (name == null) {
      throw new NullPointerException("name is null");
    } else if (levelWidth < 2 || levelHeight < 2 || tileHeightLimit < 1) {
      throw new IllegalArgumentException("one of the parameters is invalid");
    }
    this.name = name;
    this.levelWidth = levelWidth;
    this.levelHeight = levelHeight;
    this.tileHeightLimit = tileHeightLimit;
    this.twoPlayer = twoPlayer;
    userMade = !Main.develModeOn();
  }


  /**
   * Returns number of levels stored in this <code>LevelSet</code>.
   * @return number of levels
   */
  public int getSize() {
    return (levels != null) ? levels.size() : 0;
  }


  /**
   * Returns name of this <code>LevelSet</code>.
   * @return name of this set
   */
  public String getName() {
    return name;
  }


  /**
   * Changes the name of this <code>LevelSet</code>.
   * @param name new name for this set
   * @throws NullPointerException if <code>name</code> is <code>null</code>
   */
  public void setName(String name) throws NullPointerException {
    if (name == null) {
      throw new NullPointerException("name is null");
    }
    this.name = name;
  }


  /**
   * Returns whether this is a two-player <code>LevelSet</code>.
   * @return <code>true</code> if this is a two-player set
   */
  public boolean isTwoPlayer() {
    return twoPlayer;
  }


  /**
   * Creates and adds a new level to this <code>LevelSet</code>.
   * @return the newly added <code>Level</code>
   */
  public Level addLevel() {
    if (levels == null) {
      levels = new ArrayList <Level> (INITIAL_LEVEL_ARRAY_SIZE);
    }
    levels.add(new Level(levelWidth, levelHeight, tileHeightLimit, twoPlayer));
    return levels.get(levels.size() - 1);
  }


  /**
   * Adds a new level manually. Note: The level to be added will not be checked
   * for the constraints (width, height, etc.) specified when the set was
   * created.
   * @param l the <code>Level</code> to add
   * @throws NullPointerException if <code>l</code> is <code>null</code>
   */
  public void addLevel(Level l) throws NullPointerException {
    if (l == null) {
      throw new NullPointerException("l is null");
    }
    if (levels == null) {
      levels = new ArrayList <Level> (INITIAL_LEVEL_ARRAY_SIZE);
    }
    levels.add(l);
  }
  

  /**
   * Removes a level from this <code>LevelSet</code>.
   * @param index position of the level to be removed in this set
   * @throws IndexOutOfBoundsException if <code>index</code> is invalid
   */
  public void removeLevel(int index) throws IndexOutOfBoundsException {
    levels.remove(index);
  }


  /**
   * Swaps specified level with next one.
   * @param index position of the level to be swapped with its right neighbour
   * @return <code>false</code> if the level could not be swapped (i.e. if
   * trying to swap last level as there is no next level), <code>true</code>
   * otherwise
   * @throws IndexOutOfBoundsException if <code>index</code> is invalid
   */
  public boolean swapLevel(int index) throws IndexOutOfBoundsException {
    if (index > levels.size() - 2) {
      return false;
    } else {
      Level tmp = levels.get(index);
      levels.set(index, levels.get(index + 1));
      levels.set(index + 1, tmp);
      return true;
    }
  }


  /**
   * Provides direct access to a level in this set.
   * @param index position of th level to retrieve
   * @return the chosen <code>Level</code>
   * @throws IndexOutOfBoundsException if <code>index</code> is invalid
   */
  public Level getLevel(int index) throws IndexOutOfBoundsException {
    return levels.get(index);
  }


  /**
   * Removes all invalid and empty levels from this <code>LevelSet</code>.
   * Note: Validity of a level is defined in the following method:<code>
   * Level.isValid()</code>
   */
  public void purgeInvalidLevels() {
    for (int i = 0; i < levels.size(); i++) {
      if (!levels.get(i).isValid()) {
	levels.remove(i);
	i--;
      }
    }
  }


  /**
   * Removes all empty levels from this <code>LevelSet</code>
   */
  public void purgeEmptyLevels() {
    for (int i = 0; i < levels.size(); i++) {
      if (levels.get(i).isEmpty()) {
	levels.remove(i);
	i--;
      }
    }
  }


  /**
   * Returns indices of invalid levels in this <code>LevelSet</code>.
   * Note: Empty levels are not counted as invalid.
   * @return array of invalid level indices or <code>null</code> if all levels
   * are valid
   */
  public int[] getInvalidLevelIndices() {
    /* this might not be that efficient but it's probably less ugly than
     * resizing an array manually (alternative would be to just return
     * the ArrayList itself
     */
    ArrayList <Integer> invalidLevelList = new ArrayList <Integer> (3);
    for (int i = 0; i < levels.size(); i++) {
      if (!levels.get(i).isEmpty() && !levels.get(i).isValid()) {
	invalidLevelList.add(new Integer(i));
      }
    }
    if (invalidLevelList.size() == 0) {
      return null;
    } else {
      int[] invalidLevels = new int[invalidLevelList.size()];
      for (int i = 0; i < invalidLevelList.size(); i++) {
	invalidLevels[i] = invalidLevelList.get(i).intValue();
      }
      return invalidLevels;
    }
  }


  /**
   * Returns whether the level can be saved or not.
   * Note: Levels which have been saved in development mode cannot be altered
   * in default (user) mode.
   * @return <code>true</code> if the level can be saved.
   */
  public boolean readOnly() {
    return !(Main.develModeOn() ? true : userMade);
  }


  /**
   * Writes the specified <code>LevelSet</code> to a file.
   * Note: If the level is read-only (see <code>readOnly</code> then the
   * level will definitely not be saved.
   * @param levelSet the set to save
   * @param filename name (and path) of the file to save the set in
   * @param overwrite specifies whether to force the file to be overwritten if
   * it already exists
   * @return <code>false</code> if<ul>
   * <li>there are no levels to save</li>
   * <li>the file already exists and the <code>overwrite</code> parameter is
   * not set</li>
   * <li>the existing file cannot be written to</li></ul>
   * @throws IOException if the write to disk attempt was not successful
   */
  public static boolean saveToDisk(LevelSet levelSet, String filename,
				   boolean overwrite) throws IOException {
    if (!levelSet.readOnly()) {
      File outFile = new File(filename);
      
      if (levelSet.levels.size() == 0 ||
	  (outFile.exists() && (!overwrite || !outFile.canWrite()))) {
	return false;
      }
      FileOutputStream fStream = new FileOutputStream(outFile);
      ObjectOutputStream oStream = new ObjectOutputStream(fStream);
      oStream.writeObject(levelSet);
      oStream.close();
      fStream.close();
      return true;
    } else {
      return false;
    }
  }

  
  /**
   * Loads a <code>LevelSet</code> from the given location.
   * Note: This method can be used to load a set from:<ul>
   * <li>local or remote disk</li>
   * <li>website</li>
   * <li>JAR</li></ul>
   * @param location full link to location of file from which to load the set
   * @return the loaded <code>LevelSet</code>
   * @throws ClassNotFoundException if the file does not contain a <code>
   * LevelSet</code> (or wrong <code>LevelSet</code> format revision)
   * @throws IOException if the read from file attempt was not successful
   * @throws NullPointerException if <code>location</code> is <code>null</code>
   */
  public static LevelSet load(URL location)
    throws ClassNotFoundException, IOException, NullPointerException {
    if (location == null) {
      throw new NullPointerException("location is null");
    }
    InputStream fStream = location.openStream();
    LevelSet s = load(fStream);
    fStream.close();
    return s;
  }


  /**
   * Loads a <code>LevelSet</code> from a file.
   * @param filename name (and path) of the file to load the set from
   * @return the loaded <code>LevelSet</code>
   * @throws ClassNotFoundException if the file does not contain a <code>
   * LevelSet</code> (or wrong <code>LevelSet</code> format revision)
   * @throws IOException if the read from file attempt was not successful
   * @throws NullPointerException if <code>filename</code> is <code>null</code>
   */
  public static LevelSet load(String filename)
    throws ClassNotFoundException, NullPointerException, IOException {
    if (filename == null) {
      throw new NullPointerException("filename is null");
    }
    FileInputStream fStream = new FileInputStream(filename);
    LevelSet s = load(fStream);
    fStream.close();
    return s;
   }


  /**
   * Common routine for loading form either file or URL.
   */
  private static LevelSet load(InputStream stream)
    throws ClassNotFoundException, IOException {
    ObjectInputStream oStream = new ObjectInputStream(stream);
    LevelSet s = (LevelSet) oStream.readObject();
    oStream.close();
    // if level is being edited in developer mode then set the required flag
    if (Main.develModeOn()) {
      s.userMade = false;
    }
    return s;
  }

}