/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import javax.imageio.ImageIO;
import com.jgoodies.looks.plastic.PlasticXPLookAndFeel;
import com.jgoodies.looks.plastic.theme.LightGray;


/**
 * Class from which <code>JTP</code> is launched.
 * @author V.Termanis
 */
public class Main {

  /**
   * Version string
   */
  protected static final String version = "1.02";
  /**
   * Copyright message
   */
  protected static final String cpMessage =
    "JTilePile Copyright (C) 2008 by V.Termanis\n" +
    "TilePile game concept (C) 2008 by J.Mitchell\n" +
    "\"Komika\" TrueType fonts (C) 1999-2001 WolfBainX & Apostrophic Labs"+
    "\n\"JGoodies Looks\" Look&Feel  (C) 2001-2008 Karsten Lentzsch\n" +
    "All rights reserved.";
  /**
   * Colour for loading progress overlay on splash screen
   */
  protected static final Color progressColor = new Color(0, 0, 255, 85);
  /**
   * How much of the loading screen (height wise) should be filled with the
   * overlay colour
   */
  protected static final int progressHeight = 255;

  // if not set to 0 game will expire after specified number of milliseconds
  private static final int DEMO_TIMEOUT = 0; //120000;

  // Relative path to changelog location
  private static final String changeLog = "JTP/ChangeLog.txt";

  // Relative path to program icon
  private static final String iconPath = "JTP/img/JTP_icon.png";

  // indicates whether the game is being run in applet mod
  private static boolean appletMode = false;

  /**
   * Launch method.
   * Possible parameters (all will be ignored if more than one specified):<br>
   * <ul><li><code>-q</code>/<code>--quality</code> - Use antialiasing and high
   * quality rendering everywhere</li>
   * <li><code>-f</code>/<code>--fast</code> - Use no antialiasing  and low
   * quality rendering for maximum performance</li></ul>
   * To use development mode start with <code>JTP.Development</code> flag
   * set to <code>true</code>.
   */
  public static void main(String[] args) {

    SplashControl splash = new SplashControl(29, progressColor,
					     progressHeight);

    System.out.println("JTilePile v" + version + "\n\n" + cpMessage + "\n");

    parseArgs(args);

    if (develModeOn()) {
      System.out.println("*** Development mode enabled ***\n");
    }
    splash.updateProgress();

    // JGoodies Look&Feel
    setCustomLookAndFeel();
    splash.updateProgress();
    
    // game window
    final NonClosingJFrame frame = new NonClosingJFrame("Tile Pile");
    splash.updateProgress();

    // program icon
    try {
      Image icon = ImageIO.read(ClassLoader.getSystemClassLoader().
				getResource(iconPath));
      frame.setIconImage(icon);
    } catch (Exception e) {
      System.out.println("Warning: Failed to load program icon");
    }
    splash.updateProgress();

    Screen p = new Screen(1);
    splash.updateProgress();
    // register for focus gain/loss events
    frame.registerScreen(p);
    frame.add(p);
    frame.pack();

    if (centered(frame)) {
      splash.updateProgress();    
      frame.setResizable(false);
      try {
	p.start(splash);
      } catch (Exception e) {
	System.out.println();
	e.printStackTrace();
	JOptionPane.showMessageDialog(frame, "A fatal error occured:\n\n" +
				      e.getMessage() + "\n" +
				      stackTraceToString(e),
				      "Tile Pile v" + version + " Error",
				      JOptionPane.ERROR_MESSAGE);
	System.exit(1);
      }
      /* If the parent JFrame does not pick up key events then keyboard
       * input can only be used after the Screen object (canvas) has been
       * focused (via mouse click), so add listener to parent component.
       */
      frame.addKeyListener(p.getKeyHandler());
      frame.setVisible(true);

      /* DEMO version functionality - game closes automatically after specified
       * amount of time (milliseconds)
       */
      if (!develModeOn()) {
	limitGameTime(DEMO_TIMEOUT, frame);
      }

      // do not continue if window does not fit on screen
    } else {
      JOptionPane.showMessageDialog(frame,
				    "Please increase your screen resolution. "
				    + "(Minimum required width: " +
				    (p.getSize().width + 10) + ", height: " +
				    (p.getSize().height + 30) + ")",
				    "Tile Pile",
				    JOptionPane.INFORMATION_MESSAGE);
      System.exit(0);
    }
  }


  /**
   * Tries to apply custom look and feel.
   */
  protected static void setCustomLookAndFeel() {
    PlasticXPLookAndFeel.setPlasticTheme(new LightGray());
    try {
      UIManager.setLookAndFeel(new PlasticXPLookAndFeel());
      // should never happen as theme is included
    } catch (Exception e) {
      System.out.println("Warning: Failed to load custom L&F");
    }
  }


  /**
   * Formats traceback stack by inserting line breaks.
   */
  protected static String stackTraceToString(Exception e) {
    String result = "";
    for (int i = 0; i < e.getStackTrace().length; i++) {
      result += e.getStackTrace()[i].toString() + "\n";
    }
    return result;
  }


  /**
   * Primitive command line argument parser
   * (If there were any more parameters the java-getopt package would be the
   * better solution)
   */
  private static void parseArgs(String[] args) {
    boolean renderMode = false;

    for (int i = 0; i < args.length; i++) {
      // FAST MODE
      if (args[i].equals("-f") || args[i].equals("--fast")) {
	if (!renderMode) {
	  renderMode = true;
	  if (!setProperty("JTP.RenderQuality",
			   Integer.toString(ScreenFrames.RENDERING_SPEED))) {
	    System.out.println("Warning: Failed to enable fast mode.");
	  } else {
	    System.out.println("Fast mode enabled");
	  }
	} else {
	  System.out.println("Ignoring \"" + args[i] +
			     "\" (fast/quality mode already set)");
	}
	// QUALITY MODE
      } else if (args[i].equals("-q") || args[i].equals("--quality")) {
	if (!renderMode) {
	  renderMode = true;
	  if (!setProperty("JTP.RenderQuality",
			   Integer.toString(ScreenFrames.RENDERING_QUALITY))) {
	    System.out.println("Warning: Failed to enable quality mode.");
	  } else {
	    System.out.println("Quality mode enabled");
	  }
	} else {
	  System.out.println("Ignoring \"" + args[i] +
			     "\" (fast/quality mode already set)");
	}
      }
    }
    System.out.println();
  }


  /**
   * Sets a system property gracefully.
   * @param name name of the property to set
   * @param value value of the property to set
   * @return <code>true</code> if the property was successfully set, <code>
   * false</code> otherwise.
   */  
  protected static boolean setProperty(String name, String value) {
    try {
      System.setProperty(name, value);
      return true;
    } catch (SecurityException e) {
      return false;
    }
  }


  /**
   * Tries to center the frame is possible.
   */
  protected static boolean centered(JFrame frame) {
    GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
    Point center = ge.getCenterPoint();
    Rectangle bounds = ge.getMaximumWindowBounds();

    // is there enough space for the game window?
    if (bounds.width < frame.getWidth() ||
	bounds.height < frame.getHeight()) {
      return false;
    }
    frame.setLocation(center.x - (frame.getWidth() / 2),
		      center.y - (frame.getHeight() / 2));
    frame.validate();
    return true;
  }


  /**
   * Indicates whether development mode has been turned on.
   * Note: To turn on development mode set <code>JTP.Development=true</code>.
   * @return <code>true</code> if development mode is on.
   */
  public static boolean develModeOn() {
    String result = "false";
    try {
      result = System.getProperty("JTP.Development", "false");
    } catch (Exception e) {}
    return result.equalsIgnoreCase("true");
  }


  /**
   * Used by <code>MainApplet</code> class to indicate the game is running
   * in applet mode. (So e.g. Exit button is not shown in the main menu and
   * the progress file is saved to a meaningful location given the game is
   * running through a browser plugin
   */
  public static void setAppletMode() {
    appletMode = true;
  }

  
  /**
   * Queries whether the game is running as an applet. (This has previously
   * been set using <code>setAppletMode</code>.)
   * @return <code>true</code> if the game is running in applet mode,
   * <code>false</code> otherwise.
   */
  public static boolean inAppletMode() {
    return appletMode;
  }


  /**
   * "Demo" functionality - ends game after given number of milliseconds.
   */
  private static void limitGameTime(int timeoutMillis, final Component c) {
    if (timeoutMillis > 0) {
      JOptionPane.showMessageDialog(c,
				    "Demo version: Game will close in " +
				    (timeoutMillis / 60000) + " minutes",
				    "Tile Pile",
				    JOptionPane.INFORMATION_MESSAGE);

      (new Timer(timeoutMillis,
		 new ActionListener() {
		   public void actionPerformed(ActionEvent e) {
		     JOptionPane.
		       showMessageDialog(c,
					 "End of demo. Thank you for trying " +
					 "this game.",
					 "Tile Pile",
					 JOptionPane.INFORMATION_MESSAGE);
		     System.exit(0);
		   }})).start();
    }
  }

}
