/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

/**
 * Defines function and rendering of the menu component.
 * @author V.Termanis
 */
public class Menu {

  // possible mouseover / keyboard states

  // different menu entry selected - only applies to keyboard control
  public static final int MENU_MOVED = -3;
  // sub menu changed
  public static final int MENU_CHANGED = -2;
  public static final int NONE = -1;
  public static final int MAIN_ONEP = 0;
  public static final int MAIN_TWOP = 1;
  public static final int MAIN_EDITOR = 2;
  public static final int MAIN_HIGHSCORES = 3;
  public static final int MAIN_HELP = 4;
  public static final int MAIN_EXIT = 5;
  public static final int ONEP_NORMAL = 6;
  public static final int ONEP_CHALLENGE = 7;
  public static final int ONEP_LOAD = 8;
  public static final int ONEP_BACK = 9;
  public static final int TWOP_COOP = 10;
  public static final int TWOP_VS = 11;
  public static final int TWOP_BACK = 12;
  public static final int TWOP_COOP_NORMAL = 13;
  public static final int TWOP_COOP_CHALLENGE = 14;
  public static final int TWOP_COOP_LOAD = 15;
  public static final int TWOP_COOP_BACK = 16;
  public static final int TWOP_VS_SIM = 17;
  public static final int TWOP_VS_ALT = 18;
  public static final int TWOP_VS_LOAD = 19;
  public static final int TWOP_VS_BACK = 20;
  public static final int EDITOR_NEW = 21;
  public static final int EDITOR_LOAD = 22;
  public static final int EDITOR_BACK = 23;

  // possible menus
  /**
   * Main menu
   */
  public static final int MENU_MAIN = 0;
  /**
   * One player submenu
   */
  public static final int MENU_ONEP = 1;
  /**
   * Two player submenu
   */
  public static final int MENU_TWOP = 2;
  /**
   * Two player cooperative submenu
   */
  public static final int MENU_TWOP_COOP = 3;
  /**
   * Two player versus submenu
   */
  public static final int MENU_TWOP_VS = 4;
  /**
   * Editor submenu
   */
  public static final int MENU_EDITOR = 5;

  // menu strings
  private static final String[][] menuStrings =
  {
    {"ONE PLAYER", "TWO PLAYER", "EDITOR", "HIGHSCORES", "HELP", "EXIT"},
    {"NORMAL", "CHALLENGE", "LOAD", "BACK"},
    {"COOPERATIVE", "VERSUS", "BACK"},
    {"NORMAL", "CHALLENGE", "LOAD", "BACK"},
    {"SIMULTANEOUS", "ALTERNATE", "LOAD", "BACK"},
    {"NEW", "LOAD", "BACK"}
  };


  /*
   * Absolute locations (and size) of all buttons used for quick look-up
   * (e.g. to determine whether the mouse is hovering over a particular button)
   */
  private Rectangle[][] buttonLocations;
  // class which handles rendering of borders 
  private ScreenFrames frames;
  // used to draw the menu buttons
  private TilePainter tPainter;
  // active - mouse over, press - depressed button
  private int activeMenu, activeButton, pressedButton, oldButton;
  private boolean lastUpdateWasKeyPress;

  /**
   * Constructs a new <code>Menu</code> instance with the given parameters.
   * @param frames for drawing borders
   * @param tilePainter for drawing individual menu entries
   * @param vOffset vertical padding (from top of screen) for menu
   * @param buttonSpacing space in between buttons (in pixels)
   */
  public Menu(ScreenFrames frames, TilePainter tilePainter,
	      int vOffset, int buttonSpacing) {
    activeButton = pressedButton = oldButton = NONE;
    activeMenu = MENU_MAIN;
    this.frames = frames;
    this.tPainter = tilePainter;
    setButtonLocations(vOffset, buttonSpacing);
  }


  /**
   * Updates state of menu based on current mouse position.
   * @param mouseLocation current mouse position
   * @return clipping rectangle for updating screen ir <code>null</code>
   * if nothing has to be updated.
   */
  public Rectangle updateUI(Point mouseLocation) {
    if (mouseLocation == null) {
      return null;
    }

    int newButton = getActiveButton(mouseLocation);
    
    // no need to update anything if mouse not over any button
    if (lastUpdateWasKeyPress) {
      if (newButton == NONE) {
	return null;
      } else {
	lastUpdateWasKeyPress = false;
      }
    }

    oldButton = activeButton;
    activeButton = newButton;

    // pressed key - only ever highlight original mouse down button
    if (pressedButton != NONE && activeButton != pressedButton) {
      activeButton = NONE;
    }

    // calculate clip area
    if (oldButton == activeButton) {
      return null;
    } else {
      // only need to redraw one button
      if (oldButton == NONE) {
	return buttonLocations[activeMenu][activeButton];
      } else if (activeButton == NONE) {
	return buttonLocations[activeMenu][oldButton];
      } else {
	// clipping area for old and new button
	return buttonLocations[activeMenu][activeButton].
	  union(buttonLocations[activeMenu][oldButton]);
      }
    }
  }


  /**
   * Use to make sure pressing a menu button via mouse/keyboard is handled
   * exactly the same after the mouse leaves the screen.
   */
  protected void setLastUpdateAsKey() {
    lastUpdateWasKeyPress = true;
  }


  /**
   * Does given mouse position correspond to location of a button
   */
  private int getActiveButton(Point mouseLocation) {
    int newButton = NONE;
    for (int i = 0; i < buttonLocations[activeMenu].length; i++) {
      if (buttonLocations[activeMenu][i].contains(mouseLocation)) {
	newButton = i;
	break;
      }
    }
    return newButton;
  }
  

  /**
   * Determine whether a menu button has been clicked
   * (to be called after mousePressed()
   */
  public int clickResult(Point mouseLocation) {
    int absoluteButton = NONE;
    
    /* special case: clicking "nowhere" after keyboard has highlighted
     * a particular button should not result in any kind of selection
     * instead screen update is forced, i.e. clicking "nowhere" results
     * in the button being deselected
     */
    if (lastUpdateWasKeyPress) {
      lastUpdateWasKeyPress = false;
      activeButton = getActiveButton(mouseLocation);
      if (activeButton == NONE) {
	pressedButton = activeButton;
	return MENU_CHANGED;
      }
    }

    // only a click if still over the same button as before
    if (pressedButton != NONE && pressedButton == activeButton) {
      absoluteButton = activeButton + menuButtonOffset(activeMenu);
    }
    pressedButton = NONE;
    if (isMenuChanger(absoluteButton)) {
      switch(absoluteButton) {
      case MAIN_ONEP:
	setMenu(MENU_ONEP, mouseLocation);
	break;
      case MAIN_TWOP:
      case TWOP_COOP_BACK:
      case TWOP_VS_BACK:
	setMenu(MENU_TWOP, mouseLocation);
	break;
      case ONEP_BACK:
      case TWOP_BACK:
      case EDITOR_BACK:
	setMenu(MENU_MAIN, mouseLocation);
	break;
      case TWOP_COOP:
	setMenu(MENU_TWOP_COOP, mouseLocation);
	break;
      case TWOP_VS:
	setMenu(MENU_TWOP_VS, mouseLocation);
	break;
	/* special case - need to check (in Screen class) whether this has
	 * been unlocked or not
	 */
      case MAIN_EDITOR:
	return MAIN_EDITOR;
      }
      absoluteButton = MENU_CHANGED;
    }
    return absoluteButton;
  }


  // inform button has been pressed (not clicked)
  public void mousePressed(int button, Point mouseLocation) {
    if (lastUpdateWasKeyPress) {
      updateUI(mouseLocation);
    }
    if (button == MouseEvent.BUTTON1) {
      pressedButton = activeButton;
    }
  }


  public int keyResult(int keyCode) {
    lastUpdateWasKeyPress = true;
    /* keypress when no button active defaults to selecting first
     * menu entry (but does not do anything apart from that
     */
    if (activeButton == NONE) {
      if (keyCode == KeyEvent.VK_UP || keyCode == KeyEvent.VK_DOWN) {
	activeButton = 0;
	return MENU_MOVED;
      } else if (keyCode != KeyEvent.VK_F9 &&
		 !(keyCode == KeyEvent.VK_ESCAPE && activeMenu != MENU_MAIN)) {
	return NONE;	
      }
    }

    switch (keyCode) {
    case KeyEvent.VK_UP:
      oldButton = activeButton;
      if (activeButton > 0) {
	activeButton--;
      } else {
	activeButton = buttonLocations[activeMenu].length - 1;
      }
      return MENU_MOVED;
    case KeyEvent.VK_DOWN:
      oldButton = activeButton;
      if (activeButton < buttonLocations[activeMenu].length - 1) {
	activeButton++;
      } else {
	activeButton = 0;
      }
      return MENU_MOVED;
    case KeyEvent.VK_ESCAPE:
      if (activeMenu == MENU_MAIN) {
	return NONE;
      } else {
	switch(activeMenu) {
	case MENU_ONEP:
	  activeButton = 0;
	  activeMenu = MENU_MAIN;
	  break;
	case MENU_TWOP:
	  activeButton = 1;
	  activeMenu = MENU_MAIN;
	  break;
	case MENU_TWOP_COOP:
	  activeButton = 0;
	  activeMenu = MENU_TWOP;
	  break;
	case MENU_TWOP_VS:
	  activeButton = 1;
	  activeMenu = MENU_TWOP;
	  break;
	case MENU_EDITOR:
	  activeButton = 2;
	  activeMenu = MENU_MAIN;
	}
	return MENU_CHANGED;
      }
    case KeyEvent.VK_ENTER:
      int absoluteButton = activeButton + menuButtonOffset(activeMenu);
      if (isMenuChanger(absoluteButton)) {
	switch(absoluteButton) {
	case MAIN_ONEP:
	  setMenu(MENU_ONEP, 0);
	  break;
	case MAIN_TWOP:
	case TWOP_COOP_BACK:
	  setMenu(MENU_TWOP, 0);
	  break;
	case TWOP_VS_BACK:
	  setMenu(MENU_TWOP, 1);
	  break;
	case ONEP_BACK:
	  setMenu(MENU_MAIN, 0);
	  break;
	case TWOP_BACK:
	  setMenu(MENU_MAIN, 1);
	  break;
	case EDITOR_BACK:
	  setMenu(MENU_MAIN, 2);
	  break;
	case TWOP_COOP:
	  setMenu(MENU_TWOP_COOP, 0);
	  break;
	case TWOP_VS:
	  setMenu(MENU_TWOP_VS, 0);
	  break;
	  /* special case - need to check (in Screen class) whether this has
	   * been unlocked or not
	   */
	case MAIN_EDITOR:
	  return MAIN_EDITOR;
	}
	return MENU_CHANGED;
      } else {
	return absoluteButton;
      }
    case KeyEvent.VK_F9:
      if (!Main.inAppletMode()) {
	return Menu.MAIN_EXIT;
      }
    default:
      return NONE;
    }
  }


  // clip area after keypress
  public Rectangle keyUpdateUI() {
    return (oldButton != NONE ? buttonLocations[activeMenu][activeButton].
	    union(buttonLocations[activeMenu][oldButton]) :
	    buttonLocations[activeMenu][activeButton]);
  }
  

  // only draw changes since last mouse move
  public void drawUpdate(Graphics2D g) {
    //    drawAll(g);
    if (oldButton != activeButton) {
      drawMenuButton(activeButton, true, g);
      drawMenuButton(oldButton, false, g);
    }
  }

  
  // draw the whole menu
  public void drawAll(Graphics2D g) {
    frames.drawAll(true, g);
    for (int i = 0; i < buttonLocations[activeMenu].length; i++) {
      drawMenuButton(i, (i == activeButton), g);
    }
  }


  /**
   * Indicates whether a particular menu entry will result in a menu change
   * (e.g a submenu has been selected)
   */
  private boolean isMenuChanger(int menuIndex) {
    return (menuIndex == MAIN_ONEP || menuIndex == MAIN_TWOP ||
	    menuIndex == ONEP_BACK || menuIndex == TWOP_BACK ||
	    menuIndex == TWOP_COOP || menuIndex == TWOP_COOP_BACK ||
	    menuIndex == TWOP_VS || menuIndex == TWOP_VS_BACK ||
	    menuIndex == MAIN_EDITOR || menuIndex == EDITOR_BACK);
  }


  /**
   * Set active menu and highlight entry currently under mouse pointer.
   * @param menuIndex submenu to switch to
   * @param mousePosition current absolute mouse position
   */
  public void setMenu(int menuIndex, Point mousePosition)
    throws IllegalArgumentException {
    if (menuIndex < 0 || menuIndex > buttonLocations.length) {
      throw new IllegalArgumentException("one of the parameters is invalid");
    }
    activeMenu = menuIndex;
    oldButton = activeButton = NONE;
    updateUI(mousePosition);
  }


  /**
   * Set active menu and entry.
   * @param menuIndex submenu to switch to
   * @param menuEntryIndex which option to highlight (choose <code>NONE</code>
   * if highlighting not desired at start)
   */
  public void setMenu(int menuIndex, int menuEntryIndex)
    throws IllegalArgumentException {
    if (menuIndex < 0 || menuIndex > buttonLocations.length ||
	menuEntryIndex < 0 ||
	menuEntryIndex > (buttonLocations[menuIndex].length - 1)) {
      throw new IllegalArgumentException("one of the parameters is invalid");
    }
    activeMenu = menuIndex;
    activeButton = menuEntryIndex;
  }


  /**
   * Sets which menu entry is highlighted.
   * @param i of the menu entry to highlight
   * @throws IllegalArgumentException if <code>i</code> is outside the
   * range of valid menu entries for the currently active menu.
   */
  public void setEntryIndex(int i) throws IllegalArgumentException {
    if (i < 0 || i >= buttonLocations[activeMenu].length) {
      throw new IllegalArgumentException("one of the parameters is invalid");
    }
    activeButton = i;
  }


  /**
   * Draws a single menu button
   */
  private void drawMenuButton(int index, boolean active, Graphics2D g) {
    
    if (index != NONE) {
      tPainter.drawMenuString(menuStrings[activeMenu][index], !active,
			      buttonLocations[activeMenu][index], g);
    }
  }


  /**
   * Populates button location list for quick mouse-over lookup later
   */
  private void setButtonLocations(int vOffset, int buttonSpacing) {
    int
      // + 10 for "3D" effect
      buttonHeight = frames.getTileSize() + 10,
      // horizontal alignment of all buttons (centered)
      xCentre = frames.getDrawArea().x + (frames.getDrawArea().width / 2);
    
    buttonLocations = new Rectangle[menuStrings.length][];

    for (int menuIndex = 0; menuIndex < menuStrings.length; menuIndex++) {
      buttonLocations[menuIndex] =
	new Rectangle[menuStrings[menuIndex].length];
      int bOffset = menuButtonOffset(menuIndex);
      for (int i = 0; i < buttonLocations[menuIndex].length; i++) {
	// first menu entry in sub menu relative to offset etc ...
	if (i == 0) {
	  buttonLocations[menuIndex][i] =
	    new Rectangle(xCentre - (buttonWidth(menuIndex, i) / 2),
			  frames.getOffset().y + frames.getBorderSpacing() +
			  vOffset, buttonWidth(menuIndex, i), buttonHeight);
	  // ... all others can use previous entry as reference
	} else {
	  // next button location is relative to previous button
	  int newY = buttonLocations[menuIndex][i-1].y + buttonHeight +
	    buttonSpacing;
	  buttonLocations[menuIndex][i] =
	    new Rectangle(xCentre - (buttonWidth(menuIndex, i) / 2), newY,
			  buttonWidth(menuIndex, i), buttonHeight);
	}
      }
    }
    // APPLET MODE - Remove EXIT button from list
    if (Main.inAppletMode()) {
      Rectangle[] main = new Rectangle[buttonLocations[MENU_MAIN].length - 1];
      for (int i = 0; (i < buttonLocations[MENU_MAIN].length - 1); i++) {
	main[i] = buttonLocations[MENU_MAIN][i];
      }
      buttonLocations[MENU_MAIN] = main;
    }
  }


  /**
   * Width of a particular menu button in pixels
   */
  private int buttonWidth(int menuIndex, int buttonIndex) {
    return Geom.getTileButtonWidth(menuStrings[menuIndex][buttonIndex], true,
				   frames);
  }


  /**
   * Calculates position of a particular menu entry for the currently active
   * menu
   */
  private int menuButtonOffset(int menuIndex) {
    int mOffset = 0;
    for (int i = 0; i < menuIndex; i++) {
      mOffset += menuStrings[i].length;
    }
    return mOffset;
  }

}