/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

/**
 * Class for keeping track of which keys a player is holding down.
 * Note: All methods in this class are <b>synchronized</b>.
 * @author V.Termanis
 */
public class PlayerKeys {

  /**
   * Constand indicating no key is being held down.
   */
  public static final int NONE = -1;

  private int oneOffKey, pressedKey;
  private boolean repeatsOn;

  /**
   * Create a new <code>PlayerKeys</code> instance.
   */
  public PlayerKeys() {
    clear();
  }


  /**
   * Returns active key.
   * @return currently active key or <code>NONE</code> if no key is being
   * held down.
   */
  public int activeKey() {
      return pressedKey;
  }


  /**
   * Sets a one-off key.
   * @param keyIndex code of one-off key
   * @throws IllegalArgumentException if <code>keyIndex != NONE</code>.
   */
  public void setOneOffKey(int keyIndex) {
    synchronized(this) {
      if (keyIndex < NONE) {
	throw new IllegalArgumentException();
      }
      oneOffKey = keyIndex;
    }
  }


  /**
   * Returns whether a one-off key is active.
   * @return <code>true</code> is a one-off key is set.
   */
  public boolean oneOffKeySet() {
    synchronized(this) {
      return (oneOffKey != NONE);
    }
  }


  /**
   * Returns the currently set one-off key and clears it.
   * @return the one-off key (could be <code>NONE</code>).
   */
  public int getAndClearOneOffKey() {
    synchronized(this) {
      int tmp = oneOffKey;
      oneOffKey = NONE;
      return tmp;
    }
  }


  /**
   * Set whether any of the keys are in repeat mode or not.
   * @param on flags repeats as on or off.
   */
  public void setRepeats(boolean on) {
    repeatsOn = on;
  }


  /**
   * Indicates whether the keys are in repeat mode or not.
   * @return <code>true</code> if repeats are on.
   */
  public boolean isRepeating() {
    return repeatsOn;
  }


  /**
   * Empties currently held down key cache.
   * Note: This also includes the repeats flag and one-off key.
   */
  public void clear() {
    synchronized(this) {
      oneOffKey = pressedKey = NONE;
      repeatsOn = false;
    }
  }


  /**
   * Registers a new key as being held down.
   * @param keyIndex code of newly pressed key
   * @throws IllegalArgumentException if <code>keyIndex < 0</code>.
   * @return <code>true</code> if this is a different key from the current one.
   */
  public boolean keyDown(int keyIndex) throws IllegalArgumentException {
    synchronized(this) {
      if (keyIndex == pressedKey) {
	return false;
      }

      if (keyIndex < 0) {
	throw new IllegalArgumentException();
      }
      
      pressedKey = keyIndex;
      repeatsOn = false;
      return true;
    }
  }


  /**
   * Unregisters key which from 'held down' list. Note: Also turns off repeat
   * mode and clears tigger key if necessary.
   * @param keyIndex code of key to be released
   * @throws IllegalArgumentException if <code>keyIndex < 0</code>.
   * @return <code>false</code> if the key was not previously registered using
   * <code>keyDown</code> method.
   */
  public boolean keyUp(int keyIndex) throws IllegalArgumentException {
    synchronized(this) {
      if (keyIndex != pressedKey) {
	return false;
      }

      if (keyIndex < 0) {
	throw new IllegalArgumentException();
      }
      clear();
      return true;
    }
  }

}