/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

/**
 * Keeps track of highscores, levels reached and areas unlocked.
 * @author V.Termanis
 */
public class Progress implements java.io.Serializable {

  /**
   * Number of entries in each highscore table.
   */
  public static final int HIGH_SCORE_COUNT = 9;
  // number of highscore tables
  private static final int HIGH_SCORE_TABLES = 2;

  /**
   * Single player.
   */
  public static final boolean SINGLE = false;
  /**
   * Cooperative.
   */
  public static final boolean COOP = true;

  /* introduce "minor annoyance" for those who try to change progress.dat
   * manually
   */
  private static final int MASK = 715827887;

  private int[] levelReached;
  private int[][] highScores;
  private String[][] highScoreNames;


  /**
   * Constructs a new <code>Progress</code> instance.
   */
  public Progress() {
    levelReached = new int[2];
    levelReached[0] = levelReached[1] = MASK;
    highScores = new int[HIGH_SCORE_TABLES][];
    highScoreNames = new String[HIGH_SCORE_TABLES][];

    for (int i = 0; i < HIGH_SCORE_TABLES; i++) {
      highScores[i] = new int[HIGH_SCORE_COUNT];
      highScoreNames[i] = new String[HIGH_SCORE_COUNT];
      for(int j = 0; j < HIGH_SCORE_COUNT; j++) {
	highScores[i][j] = MASK;
      }
    }
  }


  /**
   * Returns progress made in the levelset.
   * @param coop specifies which levelset - can be either <code>SINGLE</code>
   * or <code>COOP</code>.
   * @return level reached
   */
  public int getReachedLevel(boolean coop) {
    return (levelReached[coop ? 1 : 0] ^ MASK);
  }


  /**
   * Updates progress made in the levelset. Note: Reached level will only be
   * updated if it is greater than currently stored value.
   * @param levelNumber level reached
   * @param coop specifies which levelset - can be either <code>SINGLE</code>
   * or <code>COOP</code>.
   * @throws IllegalArgumentException if <code>levelNumber < 0</code>.
   */
  public void setReachedLevel(int levelNumber, boolean coop)
    throws IllegalArgumentException {
    if (levelNumber < 0) {
      throw new IllegalArgumentException("levelnumber is less than zero");
    }
    if (levelNumber > getReachedLevel(coop)) {
      levelReached[coop ? 1 : 0] = (levelNumber ^ MASK);
    }
  }


  /**
   * Checks whether the given score can go in the highscore table.
   * @param coop specifies which levelset - can be either <code>SINGLE</code>
   * or <code>COOP</code>.
   * @param score the score.
   * @return position in the highscore table the score will take or
   * <code>-1</code> if the score is not high enough.
   */
  public int isHighScoreWhere(boolean coop, int score) {
    if (score > (highScores[coop ? 1 : 0][0] ^ MASK)) {
      int index = 0;
      for (; (index < (HIGH_SCORE_COUNT - 1) &&
	      score > (highScores[coop ? 1 : 0][index + 1] ^ MASK)); index++);
      return index;
    } else {
      return -1;
    }
  }


  /**
   * Checks whether the given score can go in the highscore table.
   */
  private boolean isHighScore(boolean coop, int score) {
    return (isHighScoreWhere(coop, score) >= 0);
  }


  /**
   * Adds score to the highscore table.
   * @param coop specifies which levelset - can be either <code>SINGLE</code>
   * or <code>COOP</code>.
   * @param score the score.
   * @return <code>true</code> if the score was added or <code>false</code>
   * if <code>score</code> is not a highscore.
   * @throws NullPointerException if <code>name</code> is <code>null</code>.
   * @throws IllegalArgumentException if <code>score < 1</code>.
   */
  public boolean addHighScore(boolean coop, int score, String name)
    throws NullPointerException, IllegalArgumentException {
    if (name == null) {
      throw new NullPointerException("name is null");
    }
    if (score < 1) {
      throw new IllegalArgumentException("score is less than one");
    } else if (!isHighScore(coop, score)) {
      return false;
    }
    /* find right position for highscore and shift others accordingly
     * note: older scores are rated higher than new ones if they are equal
     */
    int table = coop ? 1 : 0;
    int pos = 0;
    while (pos < (HIGH_SCORE_COUNT - 1)) {
      if (score > (highScores[table][pos + 1] ^ MASK)) {
	highScores[table][pos] = highScores[table][pos + 1];
	highScoreNames[table][pos] = highScoreNames[table][pos + 1];
	pos++;
      } else {
	break;
      }
    }
    highScores[table][pos] = (score ^ MASK);
    highScoreNames[table][pos] = name;
    return true;
  }


  /**
   * Retrieves score from the highscore table.
   * @param coop specifies which levelset - can be either <code>SINGLE</code>
   * or <code>COOP</code>.
   * @param index in highscore list of wanted score.
   * @return the score
   * @throws ArrayIndexOutOfBoundsException if <code>index < 0</code> or
   * <code>index >= HIGH_SCORE_COUNT</code>.
   */
  public int getScore(boolean coop, int index)
    throws ArrayIndexOutOfBoundsException {
    return (highScores[coop ? 1 : 0][index] ^ MASK);
  }


  /**
   * Retrieves name associated with a score from the highscore table.
   * @param coop specifies which levelset - can be either <code>SINGLE</code>
   * or <code>COOP</code>.
   * @param index in highscore list of wanted score.
   * @return the score
   * @throws ArrayIndexOutOfBoundsException if <code>index < 0</code> or
   * <code>index >= HIGH_SCORE_COUNT</code>.
   */
  public String getName(boolean coop, int index)
    throws ArrayIndexOutOfBoundsException {
    return highScoreNames[coop ? 1 : 0][index];
  }

}