/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.io.*;

/**
 * Provides methods for storing highscore and progress data to disk and
 * handling of area unlocking.
 * @author V.Termanis
 */
public class ProgressHandler {

  // progress saving result
  /**
   * Indicates progress data could not be saved.
   */
  public static final int FAILED_TO_SAVE = -1;
  /**
   * Indicates progress data was saved.
   */
  public static final int SAVED = 0;
  /**
   * Indivates progress data was saved and an area unlocked.
   */
  public static final int UNLOCKED = 1;
  
  // progress milestones
  /**
   * Progress milestone: Single player challenge mode
   */
  public static final int ONEP_CHALLENGE = 0;
  /**
   * Progress milestone: Two player cooperative mode
   */
  public static final int TWOP_COOP = 1;
  /**
   * Progress milestone: Two player challenge mode
   */
  public static final int TWOP_CHALLENGE = 2;
  /**
   * Progress milestone: Two player versus mode
   */
  public static final int TWOP_VS = 3;
  /**
   * Progress milestone: Two player versus alternate mode
   */
  public static final int TWOP_VS_ALT = 4;
  /**
   * Progress milestone: Loading of custom levelsets
   */
  public static final int LOAD_CUSTOM = 5;
  /**
   * Progress milestone: Editor
   */
  public static final int EDITOR = 6;
  
  // name/path of progress storage file
  private static String progressFileName = "Progress";

  // unlockable area levels and their string descriptions
  private static final int[] unlockPoints = {30, 40, 50, 60, 70, 80, 90};
  private static final String[] unlockMessages =
  {"Challenge Mode", "Two Player Cooperative Mode",
   "Two Player Challenge Mode", "Two Player Versus Simultaneous Mode",
   "Two Player Versus Alternate Mode", "Custom Levelset loading", "Editor"};


  /**
   * Sets location of progress file. Note: No check is performed on whether
   * this is a valid location, i.e. whether the file can be read from / written
   * to.
   * @param path the relative or absolute path (and name) of the progress
   * file
   * @throws NullPointerException if <code>path</code> is <code>null</code>
   */
  public static void setProgressFilePath(String path)
    throws NullPointerException{
    if (path == null) {
      throw new NullPointerException("path is null");
    }
    progressFileName = path;
  }


  /**
   * Returns location of progress file.
   * @return the relative or absolute path (and name) of the progress file
   */
  public static String getProgressFilePath() {
    return progressFileName;
  }


  /**
   * Returns string representation of unlocked area.
   * @param levelIndex level currently unlocked.
   * @return string representation of unlocked area of <code>null</code> if
   * completing the given level did not unlock a new area.
   */
  public static String getUnlockMessage(int levelIndex) {
    int index = getUnlockType(levelIndex);
    if (index != -1) {
      return (unlockMessages[index] + " unlocked!");
    } else {
      return null;
    }
  }

  
  private static int getUnlockType(int levelIndex) {
    int result = -1;
    for (int i = 0; i < unlockPoints.length; i++) {
      if (levelIndex == unlockPoints[i]) {
	result = i;
	break;
      }
    }
    return result;
  }

  
  /**
   * Returns whether a particular level will unlock an area
   * @param levelIndex the level to check against
   * @return the area code if this area has not been unlocked or
   * <code>-1</code> otherwise.
   */
  public static int levelWillUnlockMode(int levelIndex) {
    int area = -1;
    if (isMilestoneLevel(levelIndex)) {
      if (areaUnlocked((area = getUnlockType(levelIndex)))) {
	area = -1;
      }
    }
    return area;
  }
  

  /**
   * Returns level which unlocks a certain area.
   * @param type one of <code>ONEP_CHALLENGE</code>, <code>TWOP_COOP</code>,
   * <code>TWOP_CHALLENGE</code>, <code>TWOP_VS</code>,
   * <code>TWOP_VS_ALT</code> or <code>EDITOR</code>.
   * @return the level which unlocks the area.
   * @throws IllegalArgumentException if <code>type</code> is not a valid area.
   */
  public static int getAreaUnlockPoint(int type) {
    if (type < 0 || type > 6) {
      throw new IllegalArgumentException("type is invalid");
    }
    return unlockPoints[type];
  }


  /**
   * Creates message informing player which level they have to beat to unlock
   * a particular area.
   * @param type one of <code>ONEP_CHALLENGE</code>, <code>TWOP_COOP</code>,
   * <code>TWOP_CHALLENGE</code>, <code>TWOP_VS</code>,
   * <code>TWOP_VS_ALT</code> or <code>EDITOR</code>.
   * @return the constructed message
   * @throws IllegalArgumentException if <code>type</code> is not a valid area.
   */
  public static String getRequestMessage(int type, boolean inGame) {
    if (type < 0 || type > 6) {
      throw new IllegalArgumentException("type is invalid");
    }
    if (inGame) {
      return ("Complete this level to unlock " + unlockMessages[type]);
    } else{
      return ("Complete One Player Normal Mode level "
	      + unlockPoints[type] +
	      " to unlock " + unlockMessages[type]);
    }
  }
  
  
  /**
   * Checks whether a particular area has been unlocked.
   * @param type one of <code>ONEP_CHALLENGE</code>, <code>TWOP_COOP</code>,
   * <code>TWOP_CHALLENGE</code>, <code>TWOP_VS</code>,
   * <code>TWOP_VS_ALT</code> or <code>EDITOR</code>.
   * @return the constructed message
   * @throws IllegalArgumentException if <code>type</code> is not a valid area.
   */
  public static boolean areaUnlocked(int type)
    throws IllegalArgumentException {
    return (getReachedLevel(false) >= getAreaUnlockPoint(type));
  }


  /**
   * Checks whether any progress has been made in a levelset.
   * @param coop either <code>Progress.SINGLE</code> or
   * <code>Progress.COOP</code>.
   * @return <code>true</code> if at least one level has been completed.
   */
  public static boolean progressMade(boolean coop) {
    return (getReachedLevel(coop) > 0);
  }


  /**
   * Returns which level has been reached.
   * <b>Note:</b> In development mode this will always return
   * <code>Integer.MAX_VALUE</code>.
   * @param coop either <code>Progress.SINGLE</code> or
   * <code>Progress.COOP</code>.
   * @return <ul>
   * <li>the reached level</li>
   * <li><code>0</code> if no progress has been made, or
   * <li><code>-1</code> if progress data could not be loaded from disk.</li>
   * </ul>
   */
  public static int getReachedLevel(boolean coop) {
    if (!Main.develModeOn()) {
      Progress progress = null;
      try {
	progress = loadProgress();
	if (progress != null) {
	  return progress.getReachedLevel(coop);
	} else {
	  return 0;
	}
      } catch (Exception e) {
	return -1;
      }
    } else {
      return Integer.MAX_VALUE;
    }
  }


  /**
   * Updates progress made in levelset. Note: Reached level will only be
   * updated if it is greater than currently stored value.
   * @param levelNumber the level reached.
   * @param coop either <code>Progress.SINGLE</code> or
   * <code>Progress.COOP</code>.
   * @return <ul>
   * <li><code>SAVED</code> if the updated progress was saved to disk</li>
   * <li><code>UNLOCKED</code> if progress was saved and a new are has been
   * unlocked, or</li>
   * <li><code>FAILED_TO_SAVE</code> if there was a problem in writing the
   * progress information to disk.</li></ul>
   * @throws IllegalArgumentException if <code>levelNumber < 0</code>.
   */
  public static int setReachedLevel(int levelNumber, boolean coop)
    throws IllegalArgumentException {
    Progress progress = loadProgress();
    int oldReachedLevel = progress.getReachedLevel(coop);
    progress.setReachedLevel(levelNumber, coop);
    try {
      saveProgress(progress);
      // only report as unlocked if level hasn't been reached previously
      if (!coop && levelNumber > oldReachedLevel &&
	  isMilestoneLevel(levelNumber)) {
	return UNLOCKED;
      } else {
	return SAVED;
      }
    } catch (IOException e) {
      return FAILED_TO_SAVE;
    }
  }


  /**
   * Loads progress data from disk.
   * @return loaded progress or a new instance of <code>Progress</code> if
   * loading from disk failed.
   */
  public static Progress loadProgress() {
    Progress progress = null;
    try {
      File progressFile = new File(progressFileName);
      if (progressFile.exists()) {
	if (progressFile.canRead()) {
	  FileInputStream fStream = new FileInputStream(progressFile);
	  ObjectInputStream oStream = new ObjectInputStream(fStream);
	  progress = (Progress) oStream.readObject();
	  oStream.close();
	  fStream.close();
	} else {
	  throw new IOException("Could not read progress file " +
				progressFileName);
	}
      }
      // if for whatever reason could not read progress, create froms scratch
    } catch (Exception e) {}
    if (progress == null) {
      progress = new Progress();
    }
    return progress;
  }


  /**
   * Saves progress data to disk.
   * @param progress the progress information to save.
   * @throws IOException if there was a problem in writing the progress
   * information to disk
   * @throws NullPointerException if <code>progress</code> is <code>null</code>
   */
  public static void saveProgress(Progress progress)
    throws IOException, NullPointerException {
    if (progress == null) {
      throw new NullPointerException("progress is null");
    }
    FileOutputStream fStream = new FileOutputStream(progressFileName);
    ObjectOutputStream oStream = new ObjectOutputStream(fStream);
    oStream.writeObject(progress);
    oStream.close();
    fStream.close();
  }


  /**
   * Checks whether a given level unlocks an area.
   */
  private static boolean isMilestoneLevel(int levelIndex) {
    for (int i = 0; i < unlockPoints.length; i++) {
      if (levelIndex == unlockPoints[i]) {
	return true;
      }
    }
    return false;
  }

}