/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.IOException;
import java.io.InputStream;
import java.awt.geom.Rectangle2D;
import java.net.URL;

import javax.sound.sampled.*;


/**
 * Controls all other game components and performs double-buffered rendering.
 * @author V.Termanis
 */
public class Screen extends Canvas implements MouseInputListener {

  /*=== GLOBAL Constants ===*/

  /* minimum delay between screen updates (in milliseconds) when using mouse
   * (1000/30 ~ 33fps max)
   */
  private static final long maxUpdateInterval = 30;
  // discard old keyboard events (milliseconds)
  private static final int maxKeyboardHistory = 100;
  // number of rows and columns
  private static final int gridSize = 15;
  private static final String
    levelSetDirectory = "levelsets/",
    fileExtension = "dat",
    customFontName = "fnt/komika_text_bold.ttf",
    tileFontName = "fnt/komika_hand.ttf";

  private static final String[] defaultSetNames = {"JTP", "2p_coop", "2p_vs",
						   "2p_vs_alt"};
  /* stored here as both GamePlayer and LevelChooser components use these
   * safe to use static version as they are loaded on startup and not re-read
   * from disk ever again
   */
  private static LevelSet[] defaultSets;


  /*= Frame Constants =*/
  private static final Point fOffset = new Point(20, 20);
  private static final Color
    fGridFrameColor = new Color(245, 240, 119),
    fCtrlFrameColor = new Color(245, 186, 92);
  private static final int
    fBorderSpacing = 20,
    fCurvature = 25;
  private static final float fThickness = 3f;

  /*= Tile Constants =*/
  private static final Font
  // correct size for use with "Komika Text"
    tNumFont = new Font("Monospaced", Font.BOLD, 22),
  // correct size for use with "Komika Hand"
    tCharFont = new Font("SansSerif", Font.BOLD, 24);
  private static final Color
    tCharColor = new Color(132, 62, 176),
    tNumColor = Color.black;
  private static final int
  // size of tiles as displayed on screen
    tileSize = 40,
  // pixel height per tile level
    heightPerLevel = 5;

  /*= Splash Constants =*/
  private static final Font sFont = new Font("SansSerif", Font.BOLD, 28);
  private static final Color
    sBgColor =  new Color(0, 0, 0, 200),//Color.black,
    sTxtColor = fGridFrameColor,
    sBorderColor = fCtrlFrameColor;
  private static final int
    sLineSpacing = 10,
    sBorderSpacing = 15,
    sCurvature = 10;
  private static final float sBorderThickness = 3f;

  /*= Editor Constants =*/
  private static final Color
    eBgColor = Color.black,
    eGridColor = new Color(64,64,64),
    eGridHighlight = new Color(245, 120, 0),
    eTxtColor = tCharColor,
    eButtonHighlightColor = fCtrlFrameColor;
  private static final Font eTxtFont = new Font("SansSerif", Font.PLAIN, 20);
  private static final int
    eButtonSpacing = 20,
  // maximum tile pile height
    eTileHeightLimit = 5;
  private static final float eButtonHighlightThickness = 3f;
  // governs proportions of ctrl frame "sub-frames"
  private static final Point eCtrlFrameSplit = new Point(145, 252);
  // different cursor when mouse over grid
  private static final Cursor eGridCursor = new Cursor(Cursor.HAND_CURSOR);

  /*= Game Constants =*/
  private static final Color
    gBgColor = Color.black,
    gTxtColor = fGridFrameColor;
  private static final Font gTxtFont = new Font("SansSerif", Font.BOLD, 24);
  private static final int
    gStringSpacing = 21,
    gNumSpacing = 105,
    gCtrlTopFrameSize = 100;
  // status separation for normal, challenge, simultaneous and alternate modes
  private static final Point[] gCtrlFrameSplit = {new Point(57, 80),
						  new Point(136, 80),
						  new Point(186, 80),
						  new Point(186, 140)};
  private static final String gBeepSoundName = "snd/beep.wav";

  /*= Levelchooser Constants =*/
  private static final Font lTxtFont = new Font("SansSerif", Font.BOLD, 26);
  private static final Color lTxtColor = gTxtColor;

  /*= Challenge Constants =*/
  private static final int
    cNumberOfLevels = 20,
    cLives = 10,
    cMinTileCount = 10,
    cMinTileCountIncrement = 10,
    cMaxTileHeight = 5,    
    cTileOffset = 0,
    cGenRetries = 10;

  /*= Highscore Constants =*/
  private static final int
    hMaxHighscoreLetters = 15,
    hButtonSpacing = 25;
  private static final Color hButtonHighlightColor = fCtrlFrameColor;
  private static final float hButtonHighlightThickness =
    eButtonHighlightThickness;
  
  /*= Menu Constants =*/
  private static final Color mBgColor = Color.black;
  private static final int
    mVerticalOffset = 30,
    mButtonSpacing = 45;

  /*=== LOCAL Variables ===*/  

  /*= Game component indices =*/
  private static final int MENU = 0;
  private static final int GAME = 1;
  private static final int EDITOR = 2;
  private static final int HIGHSCORES = 3;
  private static final int LEVELCHOOSER = 4;

  // the main components
  private Menu menu;
  private GamePlayer gamePlayer;
  private Editor editor;
  private Highscores highscores;
  private LevelChooser levelChooser;
  private ScreenFrames frames;

  // handles game key input, redirects all other input back to this class
  private KeyHandler keyHandler;
  // for loading screen & warning/error messages
  private JOptionPane optionPane;
  private JFileChooser fileChooser;
  // cursor change for editor & game
  private Cursor currentCursor, nextCursor, blankCursor;
  /* indicates which part of the game is currently running (see above),
   * last direction keys chosen (so they're not repeated)
   * current mouse-down button (so don't interpret wrong mouse release event)
   */
  private int activePart, mouseDown;
  // individual & whole grid tile drawer
  private TilePainter tPainter;
  // clipping for upcoming drawing update
  private Rectangle clipArea;
  /* ensure painting done before next update, initialisation indicator,
   * nothing is executed unless true, for editor control
   */ 
  private boolean paintingDone, run, buttonClicked, keyPressed, areaChange,
    cursorHidden, screenExited, gameMoveComplete, keyboardRepeatsAllowed,
    recordScore, keyDown;
  // Popup message display component
  private SplashMessage message;
  // in-game image storage
  private Images imgData;
  // double-buffering
  private BufferStrategy buffering;
  private Graphics2D g2;
  // limiting of screen update rate
  private long lastUpdateTime, currentTime;
  /* indicates whether a message is currently being displayed or not
   * Will be one of : 0 - nothing being displayed
   *                  1 - ordinary message
   *                  2 - start of level message
   */
  private int messageStatus;
  private Timer messageTimer;
  private ActionListener messageCloser =
    new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
	messageTimer.stop();
	// take into account mouse movement during message display
	handleMouseMovement(scaled(getMousePosition()), false, false);
	updateScreen();
      }
    };
  // indicates whether to up/down - scale the game
  private double scaleFactor;


  /**
   * Creates a new <code>Screen</code> instance.
   * @param scaleFactor how much to scale output by (a value of <code>1</code>
   * results in no scaling)
   * @throws NumberFormatException if <code>scaleFactor <= 0</code>
   */
  public Screen(double scaleFactor) throws NumberFormatException {
    this.scaleFactor = scaleFactor;
    mouseDown = -1;
    run = buttonClicked = keyPressed = areaChange = cursorHidden =
      screenExited = keyDown = false;
    paintingDone = gameMoveComplete = keyboardRepeatsAllowed = true;
    messageTimer = new Timer(1000, messageCloser);
    initFrames();
  }


  /**
   * Returns size of whole (scaled) game screen.
   * @return size in pixels
   */
  public Dimension getSize() {
    if (frames != null && frames.getWholeDrawArea() != null) {
      return new Dimension((int) (frames.getWholeDrawArea().getSize().width *
				  scaleFactor),
			   (int) (frames.getWholeDrawArea().getSize().height *
				  scaleFactor));
    } else {
      return super.getSize();
    }
  }


  /**
   * Draws screen update, automatically choosing right component to render.
   * @param g graphics context for rendering
   */
  public void update(Graphics g) {
    
    /* if update forced by i.e. overlaying another window then repaint,
     * discaring message
     */
    if (messageStatus != 0) {
      paintingDone = true;
      messageTimer.stop();
    }

    if (run && paintingDone) {
      paintingDone = false;

      // POPUP MESSAGES
      if (message.requested() || messageStatus == 1) {
	handleMessages();
	if (messageStatus != 0) {
	  return;
	}
      } else {
	messageStatus = 0;
      }

      // HIGHSCORE
      if (recordScore) {
	if (gamePlayer.scoringEnabled()) {
	  logHighscore();
	}
	recordScore = false;
      }

      try {
	g2 = getBufferedGraphics();
	// === GAME ===
	if (activePart == GAME) {
	  gamePlayer.draw(g2);
	  areaChange = false;
	  // === EDITOR ===
	} else if (activePart == EDITOR) {
	  if (areaChange) {
	    editor.drawAll(g2);
	    areaChange = false;
	  } else if (buttonClicked) {
	    editor.drawClickUpdate(g2);
	    buttonClicked = false;
	  } else if (keyPressed) {
	    editor.drawKeyUpdate(g2);
	    keyPressed = false;
	  } else if (clipArea != null) {
	    g2.setClip(clipArea);
	    editor.drawUpdate(g2);
	    clipArea = null;
	  } else {
	    editor.drawAll(g2);
	  }
	  // === HIGHSCORES ===
	} else if (activePart == HIGHSCORES) {
	  if (clipArea != null) {
	    g2.setClip(clipArea);
	    highscores.drawUpdate(g2);
	    clipArea = null;
	  } else {
	    highscores.drawAll(g2);
	    areaChange = false;
	  }
	  // === LEVEL CHOOSER ===
	} else if (activePart == LEVELCHOOSER) {
	  levelChooser.drawAll(g2);
	  areaChange = false;
	  // === MENU ===
	} else {
	  if (areaChange || clipArea == null) {
	    menu.drawAll(g2);
	    areaChange = false;
	  } else {
	    g2.setClip(clipArea);
	    menu.drawUpdate(g2);
	    clipArea = null;
	  }
	}
      } finally {
	g2.dispose();
      }
      buffering.show();
      paintingDone = true;
    }
    // === GAME - check for automatic moves on start of level (e.g. Gravity)
    if (activePart == GAME && gamePlayer.levelHasNotStarted()) {
      keyHandler.autoStart();
    }
  }


  /**
   * Draws screen updated (requirement by <code>Canvas</code> class)
   */
  public void paint(Graphics g) {
    update(null);
  }

  
  /**
   * Forces screen update.
   */
  protected void updateScreen() {
    update(null);
  }


  /**
   * If window loses focus do not preserve mouse/keyboard status
   */
  protected void handleDeactivation() {
    keyDown = false;
    mouseDown = -1;
    keyHandler.reset();
  }


  /**
   * Provides access to global image resources
   * @return the image resources
   */
  protected Images getImageData() {
    return imgData;
  }

  
  /**
   * Returns absolute point after scaling on canvas.
   */
  private Point scaled(Point p) {
    return new Point((int) (p.x / scaleFactor), (int) (p.y / scaleFactor));
  }


  /**
   * Returns whether the game is scaled up/down.
   * @return <code>true</code> if the game is being scaled.
   */ 
  public boolean scaled() {
    return scaleFactor != 1;
  }


  /**
   * Handles mouse movement events.
   * @param e the incoming mouse event
   */
  public void mouseMoved(MouseEvent e) {
    if (run && !areaChange && (paintingDone || screenExited) && !keyDown &&
	!keyHandler.keyDown()) {
      handleMouseMovement(scaled(e.getPoint()), true, true);
    }
  }


  /**
   * Handles mouse movement events.
   */
  private void handleMouseMovement(Point mouseLocation, boolean drawUpdate,
				   boolean unhideMouse) {
    currentTime = System.currentTimeMillis();
    if (currentTime - lastUpdateTime >= maxUpdateInterval ||
	screenExited) {
      /* Special rule to make that activating a message in the menu via
       * mouse or keyboard is handled the same if the mouse thereafter leaves
       * the screen.
       */
      if ((activePart == MENU) && screenExited) {
	menu.setLastUpdateAsKey();
      }
      lastUpdateTime = currentTime;
      screenExited = false;
      // === GAME ===
      if (activePart == GAME) {
	if (cursorHidden && unhideMouse) {
	  cursorHidden = false;
	  setCursor(currentCursor);
	}
	if (gamePlayer.updateUI(mouseLocation) && drawUpdate) {
	  updateScreen();
	}
	// === EDITOR ===
      } else if (activePart == EDITOR) {
	if ((clipArea = editor.updateUI(mouseLocation)) != null ||
	    editor.redrawWanted()) {
	  if (editor.cursorChange()) {
	    swapCursor();
	  }
	  if (drawUpdate) {
	    updateScreen();
	  }
	}
	// === MENU ===
      } else if (activePart == MENU) {
	if ((clipArea = menu.updateUI(mouseLocation)) != null) {
	  if (drawUpdate) {
	    updateScreen();
	  }
	}
	// === LEVEL CHOOSER ===
      } else if (activePart == LEVELCHOOSER) {
	if ((clipArea = levelChooser.updateUI(mouseLocation)) != null) {
	  if (drawUpdate) {
	    updateScreen();
	  }
	}	  
	// === HIGHSCORES ===
      } else {
	if ((clipArea = highscores.updateUI(mouseLocation)) != null) {
	  if (drawUpdate) {
	    updateScreen();
	  }
	}
      }
    }
    // do not preserve clip if not going to draw now anyway
    if (!drawUpdate) {
      clipArea = null;
    }
  }


  /**
   * Handles mouse click events (once button has been released)
   * @param e the incoming mouse event.
   */
  public void mouseReleased(MouseEvent e) {
    /* ignore mouse presses which were started whilst app. was busy
     * (e.g. handling keyboard input instead)
     */
    if (mouseDown == -1 || e.getButton() != mouseDown) {
      return;
    }

    // === EDITOR ===
    if (activePart == EDITOR) {
      int result = editor.clickResult(e.getButton(), this);
      if (result == Editor.CLOSE || result == Editor.SAVECLOSE) {
	setActivePart(MENU, true);
      } else if (result == Editor.LEVEL_PLAY) {
	activateTestMode();
      } else if (result != Editor.NONE) {
	buttonClicked = true;
	updateScreen();
      }
      // === LEVEL CHOOSER ===
    } else if (activePart == LEVELCHOOSER) {
      int result = levelChooser.clickResult(scaled(getMousePosition()));
      if (result == LevelChooser.LEVEL_CHANGED) {
	updateScreen();
      } else if (result == LevelChooser.PLAY) {
	keyboardRepeatsAllowed =
	  (gamePlayer.getGameMode() == GamePlayer.SINGLE);
	/* guaranteed to have correct levelset (because this was loaded
	 * last in performMenuChoice), so no need to load again
	 */
	gamePlayer.setLevel(levelChooser.getChosenLevel());
	setActivePart(GAME, true);
      } else if (result == LevelChooser.CLOSE) {
	setActivePart(MENU, true);
      }
      // === HIGHSCORES ===
    } else if (activePart == HIGHSCORES) {
      int result = highscores.clickResult(scaled(getMousePosition()));
      if (result == Highscores.CLOSE) {
	setActivePart(MENU, true);
      } else if (result != Highscores.NONE) {
	updateScreen();
      }
      // === MENU ===
    } else if (activePart == MENU) {
      performMenuChoice(menu.clickResult(scaled(getMousePosition())));
      // === GAME ===
    } else {
      if (gamePlayer.clickResult(e.getButton()) == GamePlayer.QUIT) {
	setActivePart(gamePlayer.testModeActive() ? EDITOR : MENU, true);
      }
    }
    mouseDown = -1;
  }

  
  /**
   * performs actions from menu system depending on chosen menu entry
   * c = result from either Menu.clickResult or Menu.keyResult
   */
  private void performMenuChoice(int c) {
    String fName;
    switch(c) {
    case Menu.MENU_CHANGED:
      updateScreen();
      break;
    case Menu.MAIN_EXIT:
      System.exit(0);
    case Menu.ONEP_NORMAL:
    case Menu.TWOP_COOP_NORMAL:
    case Menu.TWOP_VS_SIM:
    case Menu.TWOP_VS_ALT:
      if (c == Menu.ONEP_NORMAL ||
	  ProgressHandler.areaUnlocked(ProgressHandler.TWOP_VS_ALT) ||
	  (c != Menu.TWOP_VS_ALT &&
	   ProgressHandler.areaUnlocked(ProgressHandler.TWOP_VS)) ||
	  (c == Menu.TWOP_COOP_NORMAL &&
	   ProgressHandler.areaUnlocked(ProgressHandler.TWOP_COOP))) {
	
	gamePlayer.setLevelSet(defaultSets[mapMenuEntryToLevelSet(c)],
			       mapMenuEntryToLevelSet(c), false,
			       getBufferedGraphics());
	if (c == Menu.ONEP_NORMAL) {
	  keyboardRepeatsAllowed = true;
	  // Andrej's company advert (only in JTP level set)
	  gamePlayer.activateMediaMerge();
	} else {
	  keyboardRepeatsAllowed = false;
	}
	if (c == Menu.ONEP_NORMAL || c == Menu.TWOP_COOP_NORMAL) {
	  gamePlayer.enableUnlocking();
	  if (ProgressHandler.progressMade(c == Menu.TWOP_COOP_NORMAL)) {
	    /* do not show popup message for first level as gameplayer
	     * will not be launched after all
	     */
	    message.cancel();
	    // show LEVEL CHOOSER
	    levelChooser.setLevelSet((c == Menu.TWOP_COOP_NORMAL) ?
				     LevelChooser.COOP :
				     LevelChooser.SINGLE);
	    setActivePart(LEVELCHOOSER, true);
	    break;
	  }
	}
	setActivePart(GAME, true);
      } else {
	int area = -1;
	switch (c) {
	case Menu.TWOP_COOP_NORMAL:
	  area = ProgressHandler.TWOP_COOP;
	  break;
	case Menu.TWOP_VS_SIM:
	  area = ProgressHandler.TWOP_VS;
	  break;
	case Menu.TWOP_VS_ALT:
	  area = ProgressHandler.TWOP_VS_ALT;
	}
	message.request(ProgressHandler.getRequestMessage(area, false));
	updateScreen();
      }
      break;
    case Menu.ONEP_LOAD:
    case Menu.TWOP_COOP_LOAD:
    case Menu.TWOP_VS_LOAD:
      // is the player allowed to load levels?
      if (!ProgressHandler.areaUnlocked(ProgressHandler.LOAD_CUSTOM)) {
	message.request(ProgressHandler.
			getRequestMessage(ProgressHandler.LOAD_CUSTOM, false));
	updateScreen();
	return;
      }
      fName = IO.getUserFile(fileChooser, IO.GAME_LOAD, fileExtension, this);
      int mode = GamePlayer.SINGLE;
      if (fName != null) {
	if (c == Menu.TWOP_VS_LOAD) {
	  mode = GamePlayer.VS_CHOOSE;
	} else if (c == Menu.TWOP_COOP_LOAD) {
	  mode = GamePlayer.COOP;
	}
	if (IO.gameLoadLevelSet(gamePlayer, fName, optionPane, this, mode,
				getBufferedGraphics())) {
	  // show levelchooser if more than one level (single & coop only)
	  if ((mode == GamePlayer.SINGLE || mode == GamePlayer.COOP) &&
	      gamePlayer.getLevelSet().getSize() > 1) {
	    /* do not show popup message for first level as gameplayer
	     * will not be launched after all
	     */
	    message.cancel();
	    levelChooser.setCustomLevelSet(gamePlayer.getLevelSet(),
					   (mode == GamePlayer.COOP));
	    levelChooser.setLevelSet(LevelChooser.CUSTOM);
	    /* no need to update screen as closure of file open dialog forces
	     * screen update (same for setActivePart in below block
	     */
	    setActivePart(LEVELCHOOSER, false);
	  } else {
	    setActivePart(GAME, false);
	    keyboardRepeatsAllowed = (c == Menu.ONEP_LOAD);
	  }
	}
      }
      break;
    case Menu.MAIN_EDITOR:
      if (ProgressHandler.areaUnlocked(ProgressHandler.EDITOR)) {
	if (keyDown || keyHandler.keyDown()) {
	  menu.setMenu(Menu.MENU_EDITOR, 0);
	} else {
	  menu.setMenu(Menu.MENU_EDITOR, scaled(getMousePosition()));
	}
      } else {
	message.request(ProgressHandler.
			getRequestMessage(ProgressHandler.EDITOR, false));
	
      }
      updateScreen();
      break;
    case Menu.EDITOR_NEW:
    case Menu.EDITOR_LOAD:
      fName =
	IO.getUserFile(fileChooser,
		       c == Menu.EDITOR_NEW ? IO.EDITOR_NEW : IO.EDITOR_LOAD,
		       fileExtension, this);
      if (fName != null &&
	  IO.editorLoadLevelSet(editor, fName, optionPane, this,
				getBufferedGraphics())) {
	setActivePart(EDITOR, false);
      }
      break;
    case Menu.ONEP_CHALLENGE:
    case Menu.TWOP_COOP_CHALLENGE:
      if (ProgressHandler.areaUnlocked(ProgressHandler.TWOP_CHALLENGE) ||
	  (c == Menu.ONEP_CHALLENGE &&
	   ProgressHandler.areaUnlocked(ProgressHandler.ONEP_CHALLENGE))) {
	int gameMode =
	  (c == Menu.ONEP_CHALLENGE) ? gamePlayer.SINGLE : gamePlayer.COOP;
	gamePlayer.
	  setLevelSet(Challenge.
		      genSet(cNumberOfLevels, cMinTileCount,
			     cMinTileCountIncrement, gridSize, cTileOffset,
			     cMaxTileHeight, cGenRetries,
			     (c == Menu.TWOP_COOP_CHALLENGE)),
		      gameMode, true, getBufferedGraphics());
	gamePlayer.setLives(cLives, getBufferedGraphics());
	setActivePart(GAME, true);
	keyboardRepeatsAllowed = (c == Menu.ONEP_CHALLENGE);
      } else {
	message.request(ProgressHandler.
			getRequestMessage((c == Menu.ONEP_CHALLENGE) ?
					  ProgressHandler.ONEP_CHALLENGE :
					  ProgressHandler.TWOP_CHALLENGE,
					  false));
	updateScreen();
      }
      break;
    case Menu.MAIN_HIGHSCORES:
      if (ProgressHandler.areaUnlocked(ProgressHandler.ONEP_CHALLENGE)) {
      setActivePart(HIGHSCORES, true);
      } else {
	message.request("Highscores are for Challenge modes which you " +
			"haven't unlocked yet");
	updateScreen();
      }
      break;
    case Menu.MAIN_HELP:
      message.requestHelp();
      updateScreen();
    }
  }


  /**
   * Handles mouse-down events.
   * @param e the incoming mouse event.
   */
  public void mousePressed(MouseEvent e) {
    // popup-messages discarded by mouse down event
    if (messageStatus != 0 && !message.hasForcedLength()) {
      messageCloser.actionPerformed(null);
      return;
    }

    /* only register if keyboard not being used and another mouse button is not
     * already held down
     */
    if (paintingDone && !keyDown && !keyHandler.keyDown() &&
	(mouseDown == -1)) {
      mouseDown = e.getButton();
      if (activePart == MENU) {
	handleMouseMovement(scaled(getMousePosition()), true, false);
	menu.mousePressed(e.getButton(), scaled(e.getPoint()));
      } else if (activePart == EDITOR && editor.mousePressed(e.getButton())) {
	updateScreen();
      } else if (activePart == LEVELCHOOSER) {
	levelChooser.mousePressed(e.getButton(), scaled(e.getPoint()));
      } else if (activePart == HIGHSCORES) {
	highscores.mousePressed(e.getButton(), scaled(e.getPoint()));
      }
    }
  }


  /**
   * Handles mouse movement while key is depressed.
   * Note: Forwards for processing to <code>mouseMoved</code> method.
   * @param e the incoming mouse event.
   */
  public void mouseDragged(MouseEvent e) {
    mouseMoved(e);
  }


  /**
   * Prevents button or grid tile to be active when mouse moved off screen
   * @param e the incoming mouse event.
   */
  public void mouseExited(MouseEvent e) {
    /*
    if (activePart != GAME &&
	messageStatus != 0 && !message.hasForcedLength()) {
      messageCloser.actionPerformed(null);
    } else {
    */
      screenExited = true;
      mouseMoved(e);
      //    }
  }


  /**
   * Handler for key releases.
   * Note: This serves only to inform the class that now keys are held down.
   * @param e the incoming key release event.
   */
  public void keyReleased(KeyEvent e) {
    keyDown = false;
  }


  /**
   * Handler for highscore name input.
   * @param e the incoming key press event.
   */
  public void keyTyped(KeyEvent e) {
    if (activePart == HIGHSCORES &&
	!(System.currentTimeMillis() - e.getWhen() > maxKeyboardHistory) &&
	highscores.keyTypedResult(e.getKeyChar()) == Highscores.NAME_CHANGED) {
      updateScreen();
    }
  }


  /**
   * Handler for all keyboard events (except in-game).
   * This method basically bypasses <code>KeyHandler</code>.
   * @param e the incoming key press event.
   */
  public void keyPressedNormal(KeyEvent e) {
    // popup-messages discarded by keypress
    if (messageStatus != 0 && !message.hasForcedLength() &&
	e.getKeyCode() == KeyEvent.VK_SPACE) {
      messageCloser.actionPerformed(null);
      return;
    }

    if (paintingDone && (mouseDown == -1)) {
      keyDown = true;
      // discard old keyboard events
      if (System.currentTimeMillis() - e.getWhen() > maxKeyboardHistory) {
	return;
      }

      if (activePart == EDITOR) {
	int result = editor.keyResult(e.getKeyCode(), this);
	if (result == Editor.CLOSE) {
	  setActivePart(MENU, false);
	  if (editor.cursorChange(true)) {
	    swapCursor();
	  }
	  menu.setEntryIndex(2);
	  updateScreen();
	} else if (result == Editor.LEVEL_PLAY) {
	  activateTestMode();
	} else if (result != Editor.NONE) {
	  keyPressed = true;
	  updateScreen();
	}
	// === MENU ===
      } else if (activePart == MENU) {
	int result = menu.keyResult(e.getKeyCode());
	if (result == Menu.MENU_MOVED) {
	  clipArea = menu.keyUpdateUI();
	  updateScreen();
	} else {
	  performMenuChoice(result);
	}
	// === LEVEL CHOOSER ===
      } else if (activePart == LEVELCHOOSER) {
	int result = levelChooser.keyResult(e.getKeyCode());
	if (result == LevelChooser.LEVEL_CHANGED) {
	  updateScreen();
	} else if (result == LevelChooser.CLOSE) {
	  setActivePart(MENU, true);
	} else if (result == LevelChooser.PLAY) {
	  keyboardRepeatsAllowed =
	    (gamePlayer.getGameMode() == GamePlayer.SINGLE);
	  /* guaranteed to have JTP as current levelset (because this was
	   * loaded last in performMenuChoice) so no need to load again
	   */
	  gamePlayer.setLevel(levelChooser.getChosenLevel());
	  setActivePart(GAME, true);
	}
	// === HIGHSCORE TABLE ===
      } else {
	switch(highscores.keyResult(e.getKeyCode(), this)) {
	case Highscores.SINGLE:
	case Highscores.COOP:
	case Highscores.NAME_CHANGED:
	case Highscores.SCORE_SET:
	  updateScreen();
	  break;
	case Highscores.CLOSE:
	  setActivePart(MENU, false);
	  /* make sure highscore entry highlighted in main menu when returning
	   * becase highscore table "feels" like part of menu system
	   */
	  menu.setEntryIndex(3);
	  updateScreen();
	  break;
	}
      }
    }
  }


  /**
   * Handler for all keyboard events whilst in-game.
   * @param keyCode1 player 1 key or control key (e.g. F9)
   * @param keyCode2 player 2 key
   * @param eventTime so that really old keyboard events can be discarded
   */
  public void keyPressedGame(int keyCode1, int keyCode2, long eventTime) {
    // popup-messages discarded by keypress
    if (messageStatus != 0 && !message.hasForcedLength() &&
	keyCode1 == KeyEvent.VK_SPACE) {
      messageCloser.actionPerformed(null);
      return;
    }

    if (paintingDone && (mouseDown == -1) && gameMoveComplete) {
      gameMoveComplete = false;
      keyDown = true;
      //discard old events
      if (System.currentTimeMillis() - eventTime > maxKeyboardHistory) {
	keyDown = false;
	gameMoveComplete = true;
	return;
      }

      int result = gamePlayer.performMoves(keyCode1, keyCode2);

      switch(result) {
      case GamePlayer.RESET:
      case GamePlayer.MOVED:
      case GamePlayer.LIFE_LOST:
      case GamePlayer.FINISHED_LEVEL:
      case GamePlayer.FINISHED_SET:
	recordScore = (result == GamePlayer.FINISHED_SET);
	if (!cursorHidden) {
	  cursorHidden = true;
	  setCursor(blankCursor);
	}
	// stop & toggle timer
	if (keyCode1 != GamePlayer.AUTOSTART && result == GamePlayer.MOVED) {
	  gamePlayer.stopAndToggleTimer();
	}
	updateScreen();
	// restart timer
	if (keyCode1 != GamePlayer.AUTOSTART && result == GamePlayer.MOVED) {
	  gamePlayer.unpauseTimer();
	}
	break;
      case GamePlayer.GAMEOVER:
      case GamePlayer.QUIT:
	recordScore = (result == GamePlayer.GAMEOVER);
	if (result == GamePlayer.QUIT) {
	  setActivePart(gamePlayer.testModeActive() ? EDITOR : MENU, true);
	} else {
	  updateScreen();
	}
	break;
      case GamePlayer.SKIPPED_LEVEL:
	updateScreen();
      }

      keyDown = false;
      gameMoveComplete = true;
    }
  }
    

  /**
   * Initialises various game components.
   * Note: This method <b>must</b> be called before trying to show this
   * component on screen.
   * @param splash instance of <code>SplashControl</code> in use.
   * @throws IOException if any of the (non-optional) game resources could not
   * be loaded.
   */
  public void start(SplashControl splash) throws IOException {
    initComponents(splash);
    run = true;
    updateScreen();
  }


  /**
   * Used by <code>KeyHandler</code> class to decide whether another move
   * should be executed.
   */
  protected boolean automaticMovesLeft() {
    return gamePlayer.automaticMovesLeft();
  }


  /**
   * Renders all popup message (in menu, before level and after level)
   */
  private void handleMessages() {
    switch (messageStatus) {
    case 0:
      try {
	g2 = getBufferedGraphics();
	// re-draw tiles to reflect current state
	if (activePart == GAME || recordScore) {
	  gamePlayer.draw(g2);
	}
	message.display(frames.getDrawArea(), g2);
      } finally {
	g2.dispose();
      }
      buffering.show();
      messageStatus = 1;
      // preserves splash message for intended duration if necessary
      if (message.getDisplayLength() > 0) {
	messageTimer.setInitialDelay((int) message.getDisplayLength());
	messageTimer.restart();
      }
      // this will allow cancelling of custom popup messages via keyboard
      keyHandler.enableSpaceKey();
      break;
    case 1:
      messageStatus = 0;
      if (activePart == GAME) {
	/* only go back to menu here so completed / failed level
	 * could be displayed (beforehand)
	 */
	if (gamePlayer.getState() == GamePlayer.FINISHED_SET ||
	    gamePlayer.getState() == GamePlayer.GAMEOVER) {
	  setActivePart(MENU, false);
	}
	gamePlayer.updateState(true);
	// need to also show next level message?
	if (message.requested()) {
	  try {
	    g2 = getBufferedGraphics();
	    gamePlayer.draw(g2);
	    message.display(frames.getDrawArea(), g2);
	  } finally {
	    g2.dispose();
	  }
	  buffering.show();
	  messageStatus = 2;
	  if (message.getDisplayLength() > 0) {
	    messageTimer.setInitialDelay((int) message.getDisplayLength());
	    messageTimer.restart();
	  }
	}
      }
    }
  }

  
  /**
   * Switches to level test mode when using editor
   */
  private void activateTestMode() {
    if (editor.cursorChange(true)) {
      swapCursor();
    }
    keyboardRepeatsAllowed = !editor.getTestSet().isTwoPlayer();
    gamePlayer.setLevelSet(editor.getTestSet(),
			   editor.getTestSet().isTwoPlayer() ?
			   GamePlayer.VS : GamePlayer.SINGLE,
			   false, getBufferedGraphics());
    gamePlayer.activateTestMode();
    setActivePart(GAME, true);
  }


  /**
   * Deals with a new highscore and subsequently displays highscore table
   */
  private void logHighscore() {
    keyHandler.setMode(KeyHandler.NORMAL, false);
    highscores.
      dealWithScore((gamePlayer.gameMode() == GamePlayer.COOP),
		    gamePlayer.getScore());
    setActivePart(HIGHSCORES, false);
  }


  /**
   * Changes which part of the game is active (e.g. editor or menu)
   */
  private void setActivePart(int part, boolean updateScreen) {
    areaChange = true;
    keyDown = false;
    mouseDown = -1;
    switch (part) {
    case MENU:
      keyHandler.setMode(KeyHandler.NORMAL, false);
      frames.setCtrlFrameType(ScreenFrames.CTRL_NONE);
      frames.setBgColor(mBgColor);
      if (activePart == GAME) {
	// make sure MediaMerge tile not used anymore
	gamePlayer.revokeMediaMergeTile();
	// return to main menu
	menu.setMenu(Menu.MENU_MAIN, scaled(getMousePosition()));
	// make sure cursor not hidden
	setCursor(currentCursor);
	cursorHidden = false;
	keyboardRepeatsAllowed = true;
      } else if (activePart == EDITOR || activePart == HIGHSCORES ||
		 activePart == LEVELCHOOSER) {
	menu.setMenu(Menu.MENU_MAIN, scaled(getMousePosition()));
      } else {
	menu.updateUI(scaled(getMousePosition()));
      }
      break;
    case EDITOR:
      // make sure cursor not hidden when exiting test mode and switch key mode
      if (activePart == GAME) {
	keyHandler.setMode(KeyHandler.NORMAL, true);
	setCursor(currentCursor);
	cursorHidden = false;
      }
      frames.setCtrlFrameType(ScreenFrames.CTRL_EDITOR);
      frames.setBgColor(eBgColor);
      editor.updateUI(scaled(getMousePosition()));
      if (editor.cursorChange()) {
	swapCursor();
      }
      break;
    case GAME:
      int mode = gamePlayer.getGameMode();
      keyHandler.setMode((mode == GamePlayer.SINGLE) ? 
			 KeyHandler.GAME_1P : KeyHandler.GAME_2P,
			 mode == GamePlayer.SINGLE || mode == GamePlayer.COOP);
      frames.setCtrlFrameType(gamePlayer.currentFrameMode());
      frames.setBgColor(gBgColor);
      gamePlayer.updateUI(scaled(getMousePosition()));
      break;
    case LEVELCHOOSER:
      // update unlocked level etc.
      levelChooser.updateUI(scaled(getMousePosition()));
      break;
    case HIGHSCORES:
      highscores.updateUI(scaled(getMousePosition()));
    }
    activePart = part;
    if (updateScreen) {
      updateScreen();
    }
  }

  
  /**
   * Does what is says on the tin.
   */
  private int mapMenuEntryToLevelSet(int menuEntry) {
    switch(menuEntry) {
    case Menu.ONEP_NORMAL:
      return GamePlayer.SINGLE;
    case Menu.TWOP_COOP_NORMAL:
      return GamePlayer.COOP;
    case Menu.TWOP_VS_SIM:
      return GamePlayer.VS;
    case Menu.TWOP_VS_ALT:
      return GamePlayer.VS_TURN_BASED;
    default:
      return -1;
    }
  }
  
  /**
   * Swaps cursor (with blank in-game or with "hand" when using editor)
   */
  private void swapCursor() {
    Cursor tmp = currentCursor;
    currentCursor = nextCursor;
    nextCursor = tmp;
    setCursor(currentCursor);
  }


  /**
   * Used by <code>GameTimer</code> to return focus to menu when time runs out.
   */
  protected void returnToMenu() {
    recordScore = true;
    setActivePart(MENU, true);
  }


  /**
   * calculates dimensions of key areas (frames, grid and controls)
   */
  private void initFrames() {
    Square gridFrame = new Square(fOffset.x, fOffset.y, (gridSize * tileSize) +
				  (fBorderSpacing * 2));

    Rectangle ctrlFrame =
      new Rectangle(gridFrame.x + gridFrame.width + fBorderSpacing,
		    gridFrame.y, (tileSize * 3) + (eButtonSpacing * 2) +
		    (fBorderSpacing * 2), gridFrame.height);
    Rectangle drawArea = (new Rectangle()).union(ctrlFrame);
    // include bottom & right spacings outside frames
    drawArea.setSize(drawArea.getSize().width + fBorderSpacing,
		     drawArea.getSize().height + fBorderSpacing);
    frames = new ScreenFrames(drawArea, gridFrame, ctrlFrame, eCtrlFrameSplit,
			      gCtrlTopFrameSize, gCtrlFrameSplit,
			      mBgColor, fGridFrameColor, fCtrlFrameColor,
			      fBorderSpacing, fCurvature, tileSize,
			      heightPerLevel, fThickness);
  }  


  /**
   * Intialisation which can only be performed one once component visible
   * hence this cannot be done in the constructor.
   */
  private void initComponents(SplashControl splash) throws IOException {
    setCursor(new Cursor(Cursor.WAIT_CURSOR));

    // double buffering
    createBufferStrategy(2);
    buffering = getBufferStrategy();
    g2 = getBufferedGraphics();
    splash.updateProgress();

    // "fake" simulatenous keyboard input (still via awt/swing keyboard events)
    keyHandler = new KeyHandler(this);
    splash.updateProgress();

    // default levelsets
    defaultSets = new LevelSet[defaultSetNames.length];
    String fName = null;
    try {
      for (int i = 0; i < defaultSets.length; i++) {
	fName = levelSetDirectory + defaultSetNames[i] + "." + fileExtension;
	defaultSets[i] = LevelSet.load(getClass().getResource(fName));
	splash.updateProgress();
      }
    } catch (Exception e) {
      throw new IOException("Failed to load " + fName);
    }

    // image resources
    imgData = new Images();
    splash.updateProgress();
    while (imgData.moreToLoad()) {
      imgData.loadNextSet();
      splash.updateProgress();
    }
    currentCursor = new Cursor(Cursor.DEFAULT_CURSOR);
    nextCursor = eGridCursor;
    splash.updateProgress();
    blankCursor = getToolkit().
      createCustomCursor(new BufferedImage(1, 1, BufferedImage.TYPE_INT_ARGB),
			 new Point(0, 0), "null");
    splash.updateProgress();

    // customised font (game)
    Font customFont =
      loadFont(customFontName, Font.TRUETYPE_FONT, Font.BOLD, 24);
    splash.updateProgress();


    // game components
    message = new SplashMessage(deriveFont(customFont, sFont), sBgColor,
				sTxtColor, sBorderColor, sLineSpacing,
				sBorderSpacing, sCurvature, sBorderThickness,
				frames, imgData, g2);
    splash.updateProgress();

    // customised font (menus, highscores)
    Font tileFont =
      loadFont(tileFontName, Font.TRUETYPE_FONT, Font.BOLD, 24);
    splash.updateProgress();

    tPainter = new TilePainter(imgData, frames, this,
			       deriveFont(customFont, tNumFont),
			       deriveFont(tileFont, tCharFont),
			       tNumColor, tCharColor, eGridColor,
			       eGridHighlight, g2);
    splash.updateProgress();
    menu = new Menu(frames, tPainter, mVerticalOffset, mButtonSpacing);
    activePart = MENU;
    splash.updateProgress();
    highscores = new Highscores(imgData, frames, tPainter,
				hButtonHighlightColor,
				hButtonHighlightThickness, hButtonSpacing,
				hMaxHighscoreLetters);
    splash.updateProgress();
    gamePlayer = new GamePlayer(imgData, frames, tPainter, gTxtColor,
				deriveFont(customFont, gTxtFont),
				gStringSpacing, gNumSpacing, null, null,
				gBeepSoundName, message, this);
    splash.updateProgress();
    editor = new Editor(imgData, frames, tPainter, gridSize,
			eTileHeightLimit, eButtonSpacing,
			eButtonHighlightThickness, eButtonHighlightColor,
			deriveFont(customFont, eTxtFont),
			eGridColor, eGridHighlight, eTxtColor, 
			null, fileExtension, null);
    splash.updateProgress();
    levelChooser = new LevelChooser(tPainter, imgData, frames, defaultSets,
				    deriveFont(customFont, lTxtFont),
				    lTxtColor, g2);
    splash.updateProgress();

    // miscellaneous
    optionPane = new JOptionPane();
    splash.updateProgress();
    fileChooser = new JFileChooser(".");
    fileChooser.setFileFilter(new LevelSetFilter(fileExtension));
    fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    // now that everything is loaded, allow user input
    addMouseMotionListener(this);
    addMouseListener(this);
    addKeyListener(keyHandler);
    setCursor(currentCursor);
  }

  
  /**
   * Allows parent class (JFrame) to forward key events to handler of this
   * class.
   */
  protected KeyHandler getKeyHandler() {
    return keyHandler;
  }


  private Font loadFont(String fontName, int type, int derivedStyle,
			int derivedSize) {
    Font customFont;
    try {
      InputStream fontStream =
	getClass().getResource(fontName).openStream();
      customFont = Font.createFont(type, fontStream);
      customFont = customFont.deriveFont(derivedStyle, derivedSize);
      fontStream.close();
    } catch (Exception e) {
      customFont = new Font("SansSerif", Font.BOLD, 24);
      System.out.println("Warning: Skipping " + customFontName +
			 " (Default font will be used)");
    }
    return customFont;
  }

		
  /**
   * Generates modified copy of font <code>f</code> with size/style of font
   * <code>style</code>.
   */
  private Font deriveFont(Font f, Font style) {
    return f.deriveFont(style.getStyle(), style.getSize());
  }
  

  /**
   * Gets rendering context for next frame and applies rendering properties.
   */
  private Graphics2D getBufferedGraphics() {
    Graphics2D g2 = (Graphics2D) buffering.getDrawGraphics();
    g2.scale(scaleFactor, scaleFactor);
    switch(frames.getRenderQuality()) {
    case ScreenFrames.RENDERING_SPEED:
      g2.setRenderingHint(RenderingHints.KEY_RENDERING,
			  RenderingHints.VALUE_RENDER_SPEED);
      g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
			  RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
      g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			  RenderingHints.VALUE_ANTIALIAS_OFF);
      g2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
			  RenderingHints.VALUE_ALPHA_INTERPOLATION_SPEED);
      g2.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING,
			  RenderingHints.VALUE_COLOR_RENDER_SPEED);
      g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL,
			  RenderingHints.VALUE_STROKE_NORMALIZE);
      break;
    case ScreenFrames.RENDERING_BALANCED:
      g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
			  RenderingHints.VALUE_TEXT_ANTIALIAS_ON);      
      break;
    case ScreenFrames.RENDERING_QUALITY:
      g2.setRenderingHint(RenderingHints.KEY_RENDERING,
			  RenderingHints.VALUE_RENDER_QUALITY);
      g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
			  RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
      // for scaled versions do not use antialiasing
      if (scaleFactor == 1) {
	g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			    RenderingHints.VALUE_ANTIALIAS_ON);      
      } else {
	g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
			    RenderingHints.VALUE_INTERPOLATION_BILINEAR);
	g2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
			    RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
      }
    }
    return g2;
  }

  
  /**
   * Overloaded from <code>Component</code> class. This method will not return
   * <code>null</code> but instead a <code>Point</code> instance with
   * coordinates [-1,-1]. (In line with what e.g. <code>mouseExited</code>
   * returns in the case where the mouse has left the screen.)
   * @return absolute coordinates of the mouse on this <code>Canvas</code> or
   * [-1,-1] if it is not over this component.
   * @throws HeadlessException if GraphicsEnvironment.isHeadless() returns true
   */
  public Point getMousePosition() throws HeadlessException {
    Point mPos = super.getMousePosition();
    if (mPos != null) {
      return mPos;
    } else {
      return new Point(-1, -1);
    }
  }


  /**
   * Not used (defined by <code>MouseInputListener</code>)
   */
  public void mouseClicked(MouseEvent e) {}


  /**
   * Not used (defined by <code>MouseInputListener</code>)
   */
  public void mouseEntered(MouseEvent e) {}

}