/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;


/**
 * Responsible for drawing borders and frames, also provides dimensions to
 * other game components.
 * @author V.Termanis
 */
public class ScreenFrames {

  /* speed uses no antialiasing, balanced some, and quality always
   * quality mode really only is for taking screenshots
   */
  public static final int RENDERING_SPEED = 0;
  public static final int RENDERING_BALANCED = 1;
  public static final int RENDERING_QUALITY = 2;

  // draw no, one or three frames (editor only)
  public static final int CTRL_NONE = -1;
  public static final int CTRL_P_NORMAL = 0;
  public static final int CTRL_P_CHALLENGE = 1;
  public static final int CTRL_P_SIMULTANEOUS = 2;
  public static final int CTRL_P_ALTERNATE = 3;
  public static final int CTRL_DEFAULT = 4;
  public static final int CTRL_EDITOR = 5;

  private Rectangle drawArea, gridArea, gridFrame, ctrlFrame, windowFrame;
  private Rectangle[][] gameCtrlFrames;
  private Rectangle[] editorCtrlFrames;
  private BasicStroke stroke;
  private Color bgColor, gridFrameColor, ctrlFrameColor;
  private int borderSpacing, curvature, tileSize, heightPerLevel;
  private int renderQuality, ctrlFrameType;


  /**
   * Creates a new <code>ScreenFrames</code> instance with the specified
   * parameters.
   * @param drawArea dimensions of the whole game screen
   * @param gridFrame dimensions of the tile grid
   * @param ctrlFrame dimensions of the status/button display (RHS)
   * @param editorFrameSplit defines dimensions of the three editor control
   * frames (i.e. height of first two, third takes the rest)
   * @param gameCtrlTopFrameSize defines dimensions of the top status frame
   * in-game
   * @param gameCtrlFrameSplit dimensions for the other in-game status frames
   * (normal, cahellgen, simultaneous, alternate)
   * @param bgColor default background colour
   * @param gridFrameColor colour of frame around tile grid
   * @param ctrlFrameColor colour of frame around status/control buttons
   * @param borderSpacing padding between frames and screen borders
   * @param curvature how curvy should the frame corners be
   * @param tileSize legacy option - should be set to 40
   * @param heightPerLevel legacy option - should be set to 5
   * @param frameThickness how thick should frames be
   */
  public ScreenFrames(Rectangle drawArea, Rectangle gridFrame,
		      Rectangle ctrlFrame, Point editorFrameSplit,
		      int gameCtrlTopFrameSize, Point[] gameCtrlFrameSplit,
		      Color bgColor, Color gridFrameColor,
		      Color ctrlFrameColor, int borderSpacing,
		      int curvature, int tileSize, int heightPerLevel,
		      float frameThickness)
    throws NullPointerException, IllegalArgumentException {
    if (drawArea == null || gridFrame == null || ctrlFrame == null ||
	gridFrameColor == null || ctrlFrameColor == null) {
      throw new NullPointerException("at least one parameter is null");
    }

    if (borderSpacing < 0 || curvature < 0 || tileSize < 1 ||
	frameThickness < 0) {
      throw new IllegalArgumentException("at least one parameter is invalid");
    }
    this.stroke = new BasicStroke(frameThickness);
    this.drawArea = drawArea;
    this.gridFrame = gridFrame;
    this.ctrlFrame = ctrlFrame;
    this.bgColor = bgColor;
    this.gridFrameColor = gridFrameColor;
    this.ctrlFrameColor = ctrlFrameColor;
    this.borderSpacing = borderSpacing;
    this.curvature = curvature;
    this.tileSize = tileSize;
    this.heightPerLevel = heightPerLevel;

    renderQuality = RENDERING_BALANCED;
    try {
      renderQuality =
	Integer.parseInt(System.getProperty("JTP.RenderQuality", "1"));
    } catch (SecurityException se) {
      System.out.println("Warning: Failed to read JTP.RenderQuality");
    } catch (NumberFormatException nfe) {
      System.out.println("Warning: Invalid JTP.RenderQuality value.");
    }
    if (renderQuality < 0 || renderQuality > 2) {
      System.out.println("Warning: Invalid JTP.RenderQuality value.");      
      renderQuality = RENDERING_BALANCED;
    }
    
    windowFrame = new Rectangle(gridFrame.x, gridFrame.y,
				ctrlFrame.x - gridFrame.x + ctrlFrame.width,
				gridFrame.height);
    gridArea = new Rectangle(gridFrame);
    gridArea.grow(-borderSpacing, -borderSpacing);

    // custom layout for individual game types
    gameCtrlFrames = new Rectangle[4][];
    int bottomBoxSize = 36 + (borderSpacing * 2);

    for (int i = 0; i < gameCtrlFrames.length; i++) {
      gameCtrlFrames[i] = new Rectangle[4];
      // top box always the same
      gameCtrlFrames[i][0] =
	new Rectangle(ctrlFrame.x, ctrlFrame.y, ctrlFrame.width,
		      gameCtrlTopFrameSize);
      for (int j = 1; j < gameCtrlFrames[i].length - 1; j++) {
	gameCtrlFrames[i][j] =
	  new Rectangle(ctrlFrame.x,
			gameCtrlFrames[i][j-1].y +
			gameCtrlFrames[i][j-1].height + borderSpacing,
			ctrlFrame.width,
			getCoordinate(gameCtrlFrameSplit[i], j));
      }
      // bottom box always the same (close button only)
      gameCtrlFrames[i][3] =
	new Rectangle(ctrlFrame.x,
		      ctrlFrame.y + ctrlFrame.height - bottomBoxSize,
		      ctrlFrame.width, bottomBoxSize);
    }

    // custom control frame layout for editor
    editorCtrlFrames = new Rectangle[3];
    editorCtrlFrames[0] =
      new Rectangle(ctrlFrame.x, ctrlFrame.y, ctrlFrame.width,
		    editorFrameSplit.x);
    editorCtrlFrames[1] =
      new Rectangle(ctrlFrame.x,
		    editorCtrlFrames[0].y + editorCtrlFrames[0].height +
		    (borderSpacing / 2),
		    ctrlFrame.width, editorFrameSplit.y);
    editorCtrlFrames[2] =
      new Rectangle(ctrlFrame.x,
		    editorCtrlFrames[1].y + editorCtrlFrames[1].height +
		    (borderSpacing / 2),
		    ctrlFrame.width,
		    ctrlFrame.height -
		    (borderSpacing +
		     editorCtrlFrames[0].height + editorCtrlFrames[1].height));
    ctrlFrameType = CTRL_NONE;
  }


  /**
   * Only used by constructor to get either x or y value
   */
  private int getCoordinate(Point p, int index) {
    if (index == 1) {
      return p.x;
    } else {
      return p.y;
    }
  }

  
  /**
   * Used by other game components to determine how they should render.
   * @return the rendering quality (see Field summary)
   */
  public int getRenderQuality() {
    return renderQuality;
  }
  

  /**
   * Sets the type of control frame to use 
   * @param type (e.g. <code>CTRL_P_NORMAL</code>)
   */
  public void setCtrlFrameType(int type) throws IllegalArgumentException {
    if (type < -1 || type > 5) {
      throw new IllegalArgumentException("Frame type invalid");
    }
    ctrlFrameType = type;
  }


  /**
   * Returns the type of control frame in use
   * @return frame type
   */
  public int getCtrlFrameType() {
    return ctrlFrameType;
  }

 
  /**
   * Draws all frames (as currently configured)
   * @param background should the background be cleared also
   * @param g graphics context for rendering
   */
  public void drawAll(boolean background, Graphics2D g) {
    drawAll(background, true, g);
  }


  /**
   * Draws all frames (as currently configured)
   * @param background should the background be cleared also
   * @param thirdBox render third control frame?
   * @param g graphics context for rendering
   */
  public void drawAll(boolean background, boolean thirdBox, Graphics2D g) {
    if (background) {
      Geom.fillRect(drawArea, bgColor, g);
    }
    setRenderingHints(g, true);
    switch(ctrlFrameType) {
    case CTRL_P_NORMAL:
    case CTRL_P_CHALLENGE:
    case CTRL_P_SIMULTANEOUS:
    case CTRL_P_ALTERNATE:
      Geom.drawRoundRect(gridFrame, curvature, gridFrameColor, stroke, g);
      for (int i = 0; i < gameCtrlFrames[ctrlFrameType].length; i++) {
	if (thirdBox || (!thirdBox && i != 2)) {
	  Geom.drawRoundRect(gameCtrlFrames[ctrlFrameType][i], curvature,
			     ctrlFrameColor, stroke, g);
	}
      }
      break;
    case CTRL_EDITOR:
      Geom.drawRoundRect(gridFrame, curvature, gridFrameColor, stroke, g);
      for (int i = 0; i < editorCtrlFrames.length; i++) {
	Geom.drawRoundRect(editorCtrlFrames[i], curvature, ctrlFrameColor,
			   stroke, g);
      }
      break;
    case CTRL_NONE:
      Geom.drawRoundRect(windowFrame, curvature, gridFrameColor, stroke, g);
      break;
    case CTRL_DEFAULT:
      Geom.drawRoundRect(gridFrame, curvature, gridFrameColor, stroke, g);
      Geom.drawRoundRect(ctrlFrame, curvature, ctrlFrameColor, stroke, g);
      break;
    }
    setRenderingHints(g, false);
  }


  /**
   * Turns rendering options for the current graphics context on or off.
   * Note: This only works for the <code>RENDERING_BALANCED</code> mode.
   * @param g graphics context for rendering
   * @param on should the hints be turned on or off?
   */
  public void setRenderingHints(Graphics2D g, boolean on) {
    if (renderQuality == RENDERING_BALANCED) {
      if (on) {
	g.setRenderingHint(RenderingHints.KEY_RENDERING,
			   RenderingHints.VALUE_RENDER_QUALITY);
      } else {
	g.setRenderingHint(RenderingHints.KEY_RENDERING,
			   RenderingHints.VALUE_RENDER_DEFAULT);
      }
    }
    if (renderQuality != RENDERING_SPEED) {
      if (on) {
	g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			   RenderingHints.VALUE_ANTIALIAS_ON);
      } else {
	g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			   RenderingHints.VALUE_ANTIALIAS_OFF);
      }
    }
  }


  /**
   * Only draws the frame around tile grid
   * @param background should the background be cleared also?
   * @param g graphics context for rendering
   */
  public void drawGridFrame(boolean background, Graphics2D g) {
    if (background) {
      Geom.fillRect(drawArea, bgColor, g);
    }
    Stroke old = g.getStroke();
    Geom.drawRoundRect(gridFrame, curvature, gridFrameColor, stroke, g);
  }


  /**
   * Returns size of whole screen.
   * @return dimensions
   */
  public Rectangle getWholeDrawArea() {
    return drawArea;
  }


  /**
   * Returns size of draw area.
   * @return either dimensions of full window or, if the frame mode is not
   * <code>CTRL_NONE</code>, dimensions of tile grid.
   */
  public Rectangle getDrawArea() {
    if (ctrlFrameType != CTRL_NONE) {
      return gridFrame;
    } else {
      return windowFrame;
    }
  }

  // global offset because gridframe is drawn starting from top-left
  /**
   * Returns offset from top-left window corner where to render tile grid
   * (not frame)
   * @return location and dimensions
   */
  public Rectangle getOffset() {
    return getGridOffset();
  }

  
  /**
   * Returns location of tile grid frame
   * @return location and dimensions
   */
  public Rectangle getGridOffset() {
    return gridFrame;
  }


  /**
   * Returns default location of control frame 
   * @return location and dimensions
   */
  public Rectangle getCtrlOffset() {
    return ctrlFrame;
  }


  /**
   * Returns location of a sub-control frame (in-game only)
   * @param index which sub-frame
   * @return <code>null</code> if current ctrl setting is not one of the game
   * ones or the dimensions of the sub-frame
   */
  public Rectangle getGameCtrlOffset(int index)
    throws IndexOutOfBoundsException {
    if (index < 0 || index > 3) {
      throw new IndexOutOfBoundsException("index less than zero or too large");
    }
    if (ctrlFrameType < 0 || ctrlFrameType > 3) {
      return null;
    } else {
      return gameCtrlFrames[ctrlFrameType][index];
    }
  }

  
  /**
   * Returns location of control frames as used in editor
   * @return location and dimensions
   */
  public Rectangle[] getEditorCtrlOffset() {
    return editorCtrlFrames;
  }


  /**
   * Returns border spacing used (between frames and screen border)
   * @return spacing in pixels
   */
  public int getBorderSpacing() {
    return borderSpacing;
  }


  /**
   * Sets the background colour to use
   * @param c the new background colour to use
   */
  public void setBgColor(Color c) throws NullPointerException {
    if (c == null) {
      throw new NullPointerException("c is null");
    }
    bgColor = c;
  }


  /**
   * Returns currently used background colour
   * @return colour
   */
  public Color getBgColor() {
    return bgColor;
  }


  /**
   * Returns colour of the control frame
   * @return colour
   */
  public Color getCtrlFrameColor() {
    return ctrlFrameColor;
  }


  /**
   * Returns width/height of tiles
   * @return size in pixels
   */
  public int getTileSize() {
    return tileSize;
  }

  
  /**
   * Returns height per tile
   * @return height in pixels
   */
  public int getHeightPerLevel() {
    return heightPerLevel;
  }


  /**
   * Returns dimensions of the tile grid.
   * @return dimensions
   */
  public Rectangle getGridArea() {
    return gridArea;
  }


  /**
   * Returns thickness rounded up to the nearest integer (pixel)
   * @return frame thickness in pixels
   */
  public int getFrameThickness() {
    return (int) Math.ceil(stroke.getLineWidth());
  }

}