/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.awt.font.TextLayout;
import java.awt.image.BufferedImage;
import java.util.ArrayList;

/**
 * Displays various "pop-up" messages. (e.g. level end message)
 * Note: Only the most recently requested message will be displayed.
 * @author V.Termanis
 */
public class SplashMessage {

  /**
   * Short display length.
   */
  public static final long SHORT = 1000;
  /**
   * Default display length.
   */
  public static final long DEFAULT = 1500;
  /**
   * Long display length.
   */
  public static final long LONG = 2000;

  // smaller sized message telling player how to discard custom popups
  private static final Font pressSpaceFont = new Font("SansSerif", Font.BOLD,
						      16);
  private static final String pressSpaceString = "Press space / click mouse";


  private static final String[] helpMessage =
  {"            Use the cursor keys to move",
   "          Press Escape to restart a level",
   "Play One Player Normal Mode for a tutorial"};
  
  private String message;
  // in milliseconds
  private long displayLength;
  private Color bgColor, txtColor, borderColor;
  private int lineSpacing, borderSpacing, curvature, p1Score, p2Score;
  private Rectangle pressSpaceBounds;
  private BasicStroke borderStroke;
  private Font splashFont;
  private ScreenFrames frames;
  private Images imgData;
  // indicate what kind of message is to be displayed
  private boolean startMessage, p2, forcedLength, help;


  /**
   * Constructs a new <code>SplashMessage</code> instance with the given
   * attributes. One instance can be used to display any number of splash
   * messages.
   * @param f the font to use for text
   * @param bg background colour
   * @param txt text colour
   * @param bd border colour
   * @param lineSpacing in pixels between lines
   * @param borderSpacing spacing between text and border (both vertical and
   * horizontal)
   * @param curvature roundness of edges
   * @param borderThickness in pixels
   * @param frames provides access to screen information (e.g. to render in
   * the middle of the screen)
   * @param imgData images required by some times up splash message
   * @param g context used to determine text size (one-off)
   * @throws IllegalArgumentException if any of <code>lineSpacing</code>,
   * <code>borderSpacing</code>, <code>borderThickness</code> or
   * <code>curvature</code> are <code>< 0</code>.
   * @throws NullPointerException if any of <code>f</code>, <code>bg</code> or
   * <code>txt</code> are <code>null</code>
   */
  public SplashMessage(Font f, Color bg, Color txt, Color bd, int lineSpacing,
		       int borderSpacing, int curvature, float borderThickness,
		       ScreenFrames frames, Images imgData, Graphics2D g)
    throws IllegalArgumentException, NullPointerException {
    if (f == null || bg == null || txt == null || bd == null) {
      throw new NullPointerException("at least one parameter is null");
    }
    if (lineSpacing < 0 || borderSpacing < 0 || borderThickness < 0 ||
	curvature < 0) {
      throw new IllegalArgumentException("at least one parameter is invalid");
    }
    splashFont = f;
    bgColor = bg;
    txtColor = txt;
    borderColor = bd;
    this.lineSpacing = lineSpacing;
    this.borderSpacing = borderSpacing;    
    this.curvature = curvature;
    this.frames = frames;
    this.imgData = imgData;
    borderStroke = new BasicStroke(borderThickness);
    message = null;
    displayLength = 0;
    p1Score = p2Score = -1;
    pressSpaceBounds = Geom.calcTextBounds(pressSpaceString,
					   pressSpaceFont, g);
    startMessage = p2 = forcedLength = help = false;
  }


  /**
   * Draws the requested message to the screen.
   * @param drawArea used to calculate center of screen
   * @param g graphics context for drawing
   */
  public void display(Rectangle drawArea, Graphics2D g) {
    if (message != null || help) {
      displayString(drawArea, g);
    } else if (startMessage) {
      displayStartMessage(drawArea, g);
    } else {
      displayScore(drawArea, g);
    }
  }


  /** 
   * Displays in centre of screen - does nothing if no message specified.
   */
  private void displayString(Rectangle drawArea, Graphics2D g)
    throws NullPointerException {
    if (drawArea == null || g == null) {
      throw new NullPointerException("drawArea or g is null");
    }

    boolean showPressSpace = false;
    Rectangle textBounds;
    String[] lines;

    if (help) {
      lines = helpMessage;
      help = false;
    } else {
      lines = lineBreaker(message, g);
    }

    // size of widest/tallest line ...
    Dimension textSize = Geom.calcMaxLineDimensions(lines, splashFont, g);
    int maxLineHeight = textSize.height;
    // ... and of whole message
    textSize.height =
      (textSize.height * lines.length) + (lineSpacing * (lines.length - 1));

    // calculate dimensions and location of whole message
    int width = textSize.width + (2 * borderSpacing);
    int height = textSize.height + (2 * borderSpacing);
    // make sure there is enough space for "Press SPACE" if required
    if (displayLength == 0) {
      width = Math.max(width, pressSpaceBounds.width + (2* borderSpacing));
      height += pressSpaceBounds.height + lineSpacing;
    }
    // include "offsets" (x & y) in calculation
    int x = (drawArea.width - width) / 2 + drawArea.x;
    int y = (drawArea.height - height) / 2 + drawArea.y;

    drawSplashFrame(x, y, width, height, g);
    g.setColor(txtColor);
    g.setFont(splashFont);
    /* if there is only one line draw it centred (in case it is shorter than
     * the "Press space / click mouse" message as this is also centred)
     */
    if (lines.length == 1) {
      textBounds = Geom.calcTextBounds(lines[0], splashFont, g);
      Geom.drawString(lines[0],
		      x + ((width - textBounds.width) / 2),
		      y + borderSpacing - textBounds.y,
		      txtColor, g);
      // multi-line messages are left aligned
    } else {
      for (int i = 0; i < lines.length; i++) {
	textBounds = Geom.calcTextBounds(lines[i], splashFont, g);
	Geom.drawString(lines[i], x + borderSpacing - textBounds.x,
			y + borderSpacing - textBounds.y +
			((lineSpacing + maxLineHeight) * i),
			txtColor, g);
      }
    }
    // custom level messages indicate they can be discarded by space key
    if (displayLength == 0) {
      g.setFont(pressSpaceFont);
      Geom.drawString(pressSpaceString,
		      x + ((width - pressSpaceBounds.width) / 2),
		      y + height - borderSpacing - pressSpaceBounds.height -
		      pressSpaceBounds.y,
		      txtColor, g);
    }
    message = null;
  }


  /**
   * Displays score in center of screen.
   */
  private void displayScore(Rectangle drawArea, Graphics2D g) {
    Rectangle textBounds =
      Geom.calcTextBounds(" " + p1Score + " : " + p2Score + "  ",
			  splashFont,
			  g);
    int imgSize = imgData.getImage(Images.T_START).getWidth();
    
    int width = textBounds.width  + (2 * (borderSpacing + imgSize));
    int height = imgSize + (2 * borderSpacing);
    int x = (drawArea.width - width) / 2 + drawArea.x;
    int y = (drawArea.height - height) / 2 + drawArea.y;

    drawSplashFrame(x, y, width, height, g);
    g.setColor(txtColor);
    g.setFont(splashFont);

    g.drawImage(imgData.getImage(Images.T_START2), null,
		x + borderSpacing, y + borderSpacing);
    Geom.drawString(" " + p2Score + " : " + p1Score,
		    x + borderSpacing - textBounds.x + imgSize,
		    y + ((height - textBounds.height) / 2) - textBounds.y,
		    txtColor, g);
    g.drawImage(imgData.getImage(Images.T_START), null,
		x + width - borderSpacing - imgSize,
		y + borderSpacing);

    p1Score = p2Score = -1;
  }
  

  /**
   * Displays two player start message in center of screen.
   */
  private void displayStartMessage(Rectangle drawArea, Graphics2D g) {
    String string = " to go first. Get ready...";
    
    Rectangle textBounds =
      Geom.calcTextBounds(string, splashFont, g);
    int imgSize = imgData.getImage(Images.T_START).getWidth();
    int width = textBounds.width + (2 * borderSpacing) + imgSize;
    int height = imgSize + (2 * borderSpacing);
    int x = (drawArea.width - width) / 2 + drawArea.x;
    int y = (drawArea.height - height) / 2 + drawArea.y;

    drawSplashFrame(x, y, width, height, g);
    g.setColor(txtColor);
    g.setFont(splashFont);

    g.drawImage(imgData.getImage(p2 ? Images.T_START2 : Images.T_START), null,
		x + borderSpacing, y + borderSpacing);
    Geom.drawString(string, x + borderSpacing - textBounds.x + imgSize,
		    y + ((height - textBounds.height) / 2) - textBounds.y,
		    txtColor, g);
    startMessage = p2 = false;
  }


  /**
   * Draws splashmessage background and border
   */
  private void drawSplashFrame(int x, int y, int width, int height,
			       Graphics2D g) {
    g.setColor(bgColor);
    g.fillRoundRect(x, y, width, height, curvature, curvature);
    g.setColor(borderColor);
    g.setStroke(borderStroke);
    frames.setRenderingHints(g, true);
    g.drawRoundRect(x, y, width, height, curvature, curvature);
    frames.setRenderingHints(g, false);
  }


  /**
   * Request help message (horribly hardcoded)
   */
  protected void requestHelp() {
    help = true;
  }


  /**
   * Requests player start message to be displayed.
   * @param p2 indicates for which player the message is
   * @param time length in milliseconds to display message for
   * @throws IllegalArgumentException if <code>time < 1</code>
   */
  public void request(boolean p2, long time)
    throws IllegalArgumentException {
    cancel();
    if (time < 1) {
      throw new IllegalArgumentException("time is less than one");
    }
    startMessage = true;
    this.p2 = p2;
    displayLength = time;
    forcedLength = true;
  }


  /**
   * Requests player scores to be displayed.
   * @param p1Score player one score
   * @param p2Score player two score
   * @param time length in milliseconds to display message for
   * @throws IllegalArgumentException if <code>time < 1</code>
   */
  public void request(int p1Score, int p2Score, long time) 
    throws IllegalArgumentException {
    cancel();
    if (time < 1) {
      throw new IllegalArgumentException("time is less than one");
    }
    this.p1Score = p1Score;
    this.p2Score = p2Score;
    displayLength = time;
    forcedLength = true;
  }


  /**
   * Requests a text message to be displayed.
   * @param message message to display. If this is <code>null</code> then
   * any previous message request is cancelled.
   * @param time length in milliseconds to display message for
   * @throws IllegalArgumentException if <code>time < 1</code>
   * <code>null</code>.
   */
  public void request(String message, long time)
    throws IllegalArgumentException {
    if (time < 1) {
      throw new IllegalArgumentException("time is less than one");
    }
    cancel();
    this.message = message;
    if (message != null) {
      displayLength = time;
      forcedLength = true;
    }
  }


  /**
   * Checks whether the given message is valid.
   * @param message the string to check
   * @return <code>false</code> if any of the following is true:
   * <ul><li><code>message</code> is <code>null</code></li>
   * <li>length of <code>message</code> is zero</li>
   * <li><code>message</code> only contains whitespace characters:
   * <code>[\t\n\x0B\f\r]</code></li></ul>
   */
  public static boolean isValidMessage(String message) {
    return (message != null && message.length() > 0 &&
	    message.replaceAll("\\s","").length() > 0);
  }


  /**
   * Requests a text message to be displayed.
   * The message display length will depend on the length of the message.
   * Note: If the message is too large to fit on the screen the display
   * length will be calculated just for the part of the message which is
   * actually shown.
   * @param message message to display. If this is <code>null</code> then
   * any previous message request is cancelled.
   */
  public void request(String message) {
    cancel();
    this.message = message;
    displayLength = 0;
    forcedLength = false;
  }


  /**
   * Sets display length for any currently requested message.
   * @param time length in milliseconds to display message for
   * @throws IllegalArgumentException if <code>time < 1</code>
   */
  public void setDisplayLength(long time) throws IllegalArgumentException {
    if (time < 1) {
      throw new IllegalArgumentException("time is less than one");
    }
    displayLength = time;
    forcedLength = true;
  }


  /**
   * Returns whether a particular display length has been set or not.
   * @return <code>true</code> if message is supposed to be displayed for a
   * certain length or <code>false</code>, if this class it to determine the
   * display length (and it is o.k. to cancel display of said message at any
   * point).
   */
  public boolean hasForcedLength() {
    return forcedLength;
  }
  

  /**
   * Returns lenght for which the requested message will be displayed.
   * @return time to display in milliseconds, or zero if the display length is
   * to be calculated by this class.
   */
  public long getDisplayLength() {
    return displayLength;
  }
  

  /**
   * Cancels any currently requested messages.
   */
  public void cancel() {
    message = null;
    p1Score = p2Score = -1;
    startMessage = false;
  }


  /**
   * Checks whether a message have been requested.
   * @return <code>true</code> if a message has been requested
   */
  public boolean requested() {
    return (message != null || p1Score != -1 || startMessage == true ||
	    help == true);
  }

  
  /**
   * Breaks a string up into lines which fit onto the screen properly.
   */
  private String[] lineBreaker(String s, Graphics2D g) {
    ArrayList <String> lines = new ArrayList <String>();
    String line = new String();
    String[] words = s.split("\\s");
    int maxLineHeight = 0;
    
    for (int i = 0; i < words.length; i++) {
      // start of line
      if (line.length() == 0) {
	// ignore zero length word at start (i.e. whitespace)
	if (words[i].length() == 0) {
	  continue;
	}
	int width = Geom.calcTextBounds(words[i], splashFont, g).width;
	// is the word too long to fit on one line?
	if (width > maxLineWidth()) {
	  int subStringLength = calcSubStringLength(words[i], g);
	  line += words[i].substring(0, subStringLength) + "-";
	  words[i] = words[i].substring(subStringLength);
	  // will need to look at the rest of the word so do not advance
	  i--;
	} else {
	  line += words[i];
	}
	// not end of line
      } else {
	Rectangle bounds = Geom.calcTextBounds(line + " " + words[i],
					       splashFont, g);
	if (bounds.getWidth() < maxLineWidth()) {
	  line += " " + words[i];
	  // end of line
	} else {
	  /* determine whether line height of current line is the maximum
	   * height so far. Then use this maximum height to determine whether
	   * the current line will make the whole message too long or not
	   */
	  maxLineHeight =
	    Math.max(maxLineHeight,
		     Geom.calcTextBounds(line, splashFont, g).height);
	  if ((maxLineHeight * (lines.size() + 1)) +
	      (lineSpacing * lines.size()) >
	    maxTextHeight(true)) {
	    // add "..." to previous line to indicate it was chopped
	    String lastLine = lines.remove(lines.size() - 1);
	    lines.add(moddedLastLine(lastLine));
	    // so do not add last line again (below)
	    line = new String();
	    break;
	  }
	  lines.add(line);
	  // will need to add word on next line so repeat this iteration
	  i--;
	  line = new String();
	}
      }
    }
    // incomplete last line which still needs to be added
    if (line.length() > 0) {
      lines.add(line);
    }
    
    String[] lineArray = new String[lines.size()];
    return lines.toArray(lineArray);
  }


  /**
   * Returns legnth of substring which will fit on one line
   * (together with '-', if a word has to be split up.
   */
  private int calcSubStringLength(String s, Graphics2D g) {
    int sWidth = Geom.calcTextBounds(s, splashFont, g).width;
    // take initial guess
    int length = (int)
      Math.floor((((double) maxLineWidth()) / ((double) sWidth)) *
		 ((double) s.length())) - 3;
    // increase until maximum
    while (Geom.calcTextBounds(s.substring(0, length), splashFont, g).width <=
	   maxLineWidth()) {
      length++;
    }
    return (length - 1);
  }


  /**
   * Adds "..." after last word or chops a few letters off the only word on
   * the line.
   */
  private static String moddedLastLine(String s) {
    String[] split = s.split(" ");

    if (split.length > 1) {
      return
	(s.substring(0, (s.length() - split[split.length - 1].length())) +
	 "...");
    } else {
      return (s.substring(0, (s.length() - 2)) + "...");
    }
  }


  /**
   * Calculates the maximum width for a line of text in the given context.
   */
  private int maxLineWidth() {
    return frames.getDrawArea().width -
      ((frames.getBorderSpacing() + borderSpacing) * 2);
  }  


  /**
   * Calculates the maximum height for whole message of text in the given
   * context.
   */
  private int maxTextHeight(boolean includePressSpaceMessage) {
    if (includePressSpaceMessage) {
      return frames.getDrawArea().height -
	(pressSpaceBounds.height + lineSpacing) -
	((frames.getBorderSpacing() + borderSpacing) * 2);
    } else {
      return frames.getDrawArea().height -
	((frames.getBorderSpacing() + borderSpacing) * 2);
    }
  }

}