/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

/**
 * Serves as a parent for all startable tiles.
 * Most methods described here are overloaded.
 * @author V.Termanis
 */
public class StartableTile extends Tile {

  /**
   * Indicates whether this is a start or not
   */
  protected boolean levelStart;

  
  /**
   * Used to check for tile attribute validity upon creation.
   * @throws TileException if one or more attributes are invalid for a
   * particular tile type.
   */
  protected void tileCheck() throws TileException {
    if (height < 0) {
      throw new TileException("StartableTile.tileCheck(): Invalid height.");
    }
  }


 /**
   * Used to apply all attributes to this tile from another one.
   * @param tile the <code>Tile</code> instance from which to get attributes.
   */
  protected void transferAttributesFrom(Tile tile) {
    height = tile.height;
    levelStart = tile.isStartable() ? ((StartableTile) tile).isStart() : false;
    if (!levelStart && height == 0) {
      height++;
    }
  }

  
  /**
   * Used to apply all attributes to this tile from another one.
   * @param tile the <code>Tile</code> instance from which to get attributes.
   */
  protected void transferAttributesFrom(StartableTile tile) {
    height = tile.height;
    levelStart = tile.isStart();
  }

  
  /**
   * Sepcifies whether this is a start tile.
   * @return <code>true</code> if this is a start.
   */
  public boolean isStart() {
    return levelStart;
  }
  

  /**
   * Specifies whether this can be a start tile.
   * Not all types of tile can be a start (e.g. <code>ICE</code>).
   * @return <code>true</code> if this can be a start tile.
   */  
  public boolean isStartable() {
    return true;
  }


  /**
   * Set whether this tile is a start or not.
   * @param isStart used to toggle start status of this tile.
   * @throws TileException if the height of this tile (zero) does not permit it
   * to be changed from a start to a non-start tile.
   */
  public void setStart(boolean isStart) throws TileException {
    if (!isStart && height == 0) {
      throw new TileException("StartableTile.setStart(): " +
			      "Only start tile can have height of zero.");
    }
    levelStart = isStart;
  }


  /**
   * Reduces height of pile by one.
   * Height will be reduced to at most 0 if it is a start, 1 otherwise.
   * @return <code>true</code> if the height was reduced.
   */
  public boolean decHeight() {
    if (height > 1 || (levelStart && height == 1)) {
      height -= 1;
      return true;
    } else {
      return false;
    }
  }

  
  /**
   * Sets height of tile.
   * @param height to which to set this pile to. Note: If a height below 0 is
   * specified the height will still be set to 0. If the tile is not a start
   * then the height will be set to at least 1.
   */
  public void setHeight(int height) {
    if (levelStart) {
      this.height = (height < 0) ? 0 : height;
    } else {
      this.height = (height < 1) ? 1 : height;
    }
  }

}