/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

/**
 * Serves as parent for all types of tile. Most methods
 * described here are overloaded.
 * @author V.Termanis
 */
public class Tile implements java.io.Serializable {

  /**
   * A normal tile.
   */  
  public static final int NORMAL = 0;
  /**
   * Slides to adjacent tile.
   */  
  public static final int ICE = 1;
  /**
   * Teleports to another tile.
   */  
  public static final int TELEPORT = 2;
  /**
   * Exerts force on surrounding tiles
   */  
  public static final int GRAVITY = 3;
  /**
   * Only allows movement in certain directions.
   */  
  public static final int ARROW = 4;


  /**
   * Height of this pile
   */
  protected int height;
  /**
   * One of <code>NORMAL</code>, <code>ICE</code>, <code>TELEPORT</code>,
   * <code>GRAVITY</code> or <code>ARROW</code>
   */
  protected int type;


  /**
   * Used to check for tile attribute validity upon creation.
   * @throws TileException if one or more attributes are invalid for a
   * particular tile type.
   */
  protected void tileCheck() throws TileException {
    if (height < 1) {
      throw new TileException("Tile.tileCheck(): Invalid tile height.");
    }
  }


  /**
   * Used to apply all attributes to this tile from another one.
   * @param tile the <code>Tile</code> instance from which to get attributes.
   */
  protected void transferAttributesFrom(Tile tile) {
    height = tile.height;
    if (height == 0) {
      height++;
    }
  }


  /**
   * Returns the type of tile.
   * @return one of <code>NORMAL</code>, <code>ICE</code>,
   * <code>TELEPORT</code>, <code>GRAVITY</code> or <code>ARROW</code>.
   */
  public int getType() {
    return type;
  }

  
  /**
   * Returns height of this pile.
   * @return height of pile in as an <code>int</code>.
   */
  public int getHeight() {
    return height;
  }


  /**
   * Reduces height of pile by one (to at most 1).
   * @return <code>true</code> if the height was reduced.
   */
  public boolean decHeight() {
    if (height > 1) {
      height -= 1;
      return true;
    } else {
    return false;
    }
  }


  /**
   * Increases height of pile.
   */
  public void incHeight() {
    height += 1;
  }


  /**
   * Sets height of tile.
   * @param height to which to set this pile to. Note: If a height below 1 is
   * specified the height will still be set to one.
   */
  public void setHeight(int height) {
    this.height = (height < 1) ? 1 : height;
  }


  /**
   * Specifies whether this can be a start tile.
   * Not all types of tile can be a start (e.g. <code>ICE</code>).
   * @return <code>true</code> if this can be a start tile.
   */
  public boolean isStartable() {
    return false;
  }

  
  /**
   * Used to compare two tiles, ignoring their respective heights.
   * @param t the <code>Tile</code> instance to compare the current tile with.
   * @return <code>true</code> if the two <code>Tile</code> instances
   * are equal disregarding their heights.
   */
  public boolean equalsApartFromHeight(Tile t) {
    return false;
  }


  /**
   * Creates an exact (non-shallow) copy of the current tile
   * @return a copy of the <code>Tile</code> instance.
   */
  public Tile copy() {
    return null;
  }

}