/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

import java.awt.*;
import java.awt.image.*;
import java.awt.geom.Rectangle2D;
import java.awt.font.TextLayout;

/**
 * Handles tile rendering
 * @author V.Termanis
 */
public class TilePainter {

  /* used to calculate character offsets - all other character will be
   * positioned relative to these
   */
  private static final char
    calcChar = 'Z',
    calcDigit = '3';

  private Color gridColor, gridHighlight;
  // separate offset for '1' because the custom font is not very well aligned
  private Point numberOffset, number1Offset, charOffset;
  private Font numberFont, charFont;
  private Images imgData;
  private ScreenFrames frames;
  private Screen screen;
  private Color numberColor, charColor;


  /**
   * Constructs a new <code>TilePainter</code> instance with the given
   * attributes.
   * @param imgData graphics resources
   * @param frames geometry information
   * @param numberFont font to use for tile numbering
   * @param charFont font to use for all other characters
   * @param numColor color of numbers on tiles
   * @param charColor color of all other characters on tiles
   * @param gridColor color to draw grid with (editor)
   * @param gridHighlight color to use for highlighting grid position
   * @param g graphics context for calculating font positioning
   * @throws NullPointerException if any of the supplied parameters are
   * <code>null</code>.
   */
  public TilePainter(Images imgData, ScreenFrames frames, Screen screen,
		     Font numberFont, Font charFont, Color numColor,
		     Color charColor, Color gridColor, Color gridHighlight,
		     Graphics2D g) throws NullPointerException {
    if (imgData == null || frames == null || screen == null ||
	numberFont == null || charFont == null || gridColor == null ||
	gridHighlight == null) {
      throw new NullPointerException("at least one parameter is null");
    }
    this.imgData = imgData;
    this.frames = frames;
    this.screen = screen;
    this.numberFont = numberFont;
    this.charFont = charFont;
    this.gridColor = gridColor;
    this.gridHighlight = gridHighlight;
    this.numberColor = numColor;
    this.charColor = charColor;
    numberOffset = Geom.calcCharOnTileOffset(calcDigit, numberFont, frames, g);
    number1Offset = Geom.calcCharOnTileOffset('1', numberFont, frames, g);
    charOffset = Geom.calcCharOnTileOffset(calcChar, charFont, frames, g);
  }


  /**
   * Renders the given tiles (editor).
   * Warning: For optimum performance <b>no checks are performed</b> on any of
   * the parameters (i.e. this method might not fail gracefully) though some of
   * the parameters can be set to <code>null</code> if they do not apply.
   * @param tiles the tiles to render
   * @param activePos currently active grid position
   * @param activePos2 second active position. Used in conjuction with
   * <code>activePos</code> to highlight a whole area
   * @param oldPos specifies previously active position. Only having to
   * render over the top of one grid position improves performance.
   * @param player2Start location of second player
   * @param dualStart indicates both players start from the same position
   * @param g graphics context for drawing
   */
  public void drawTiles(Tile[][] tiles, Point activePos, Point activePos2,
			Point oldPos, Point player2Start, boolean dualStart,
			Graphics2D g) {
    drawTiles(tiles, frames.getGridArea(), frames.getTileSize(), activePos,
	      activePos2, oldPos, player2Start, true, dualStart, true, g);
  }

  
  /**
   * Renders the given tiles (in-game).
   * Warning: For optimum performance <b>no checks are performed</b> on any of
   * the parameters (i.e. this method might not fail gracefully) though some of
   * the parameters can be set to <code>null</code> if they do not apply.
   * @param tiles the tiles to render
   * @param activePos position of first player
   * @param activePos2 position of second player
   * @param numbered toggle tile height numbers
   * @param g graphics context for drawing
   */
  public void drawGTiles(GameTile[][] tiles, Point activePos, Point activePos2,
			 boolean numbered, Graphics2D g) {
    drawGTiles(tiles, frames.getGridArea(), frames.getTileSize(), activePos,
	       activePos2, numbered, g);
  }

  
  /**
   * Renders the given tiles (editor/level-chooser).
   * Warning: For optimum performance <b>no checks are performed</b> on any of
   * the parameters (i.e. this method might not fail gracefully) though some of
   * the parameters can be set to <code>null</code> if they do not apply.
   * @param tiles the tiles to render
   * @param gridArea absolute location and dimesions of the grid
   * @param tileSize size of each individual grid slot
   * @param activePos currently active grid position
   * @param activePos2 second active position. Used in conjuction with
   * <code>activePos</code> to highlight a whole area
   * @param oldPos specifies previously active position. Only having to
   * render over the top of one grid position improves performance.
   * @param player2Start location of second player
   * @param drawP2Start toggle drawing of second player's start position.
   * Note: Above parameter required to distinguish between players' start
   * positions.
   * @param dualStart indicates both players start from the same position
   * @param numbered toggle tile height numbers
   * @param g graphics context for drawing
   */
  protected void drawTiles(Tile[][] tiles, Rectangle gridArea, int tileSize,
			   Point activePos, Point activePos2, Point oldPos,
			   Point player2Start, boolean drawP2Start,
			   boolean dualStart, boolean numbered, Graphics2D g) {
    int x, y;

    if (oldPos != null) {
      if (activePos2 == null) {
	g.setColor(gridColor);
	g.drawRect(gridArea.x + (tileSize * oldPos.x),
		   gridArea.y + (tileSize * oldPos.y),
		   tileSize - 1, tileSize - 1);
      } else {
	Geom.drawRect(Geom.tilePosToArea(frames, oldPos).
		      union(Geom.tilePosToArea(frames, activePos2)),
		      gridColor, g);
      }
    }

    g.setFont(numberFont);
    for (y = 0; y < tiles.length; y++) {
      for (x = tiles.length - 1; x >= 0; x--) {
	if (tiles[y][x] != null) {
	  // draw 1P or not start
	  if ((!drawP2Start && dualStart) ||
	      !(player2Start != null &&
		player2Start.x == x && player2Start.y == y)) {
	    drawTile(gridArea.x + tileSize * x, gridArea.y + tileSize * y,
		     tiles[y][x], false, true, false, false, numbered, g);
	    // draw 2P or dual start
	  } else {
	    drawTile(gridArea.x + tileSize * x, gridArea.y + tileSize * y,
		     tiles[y][x], false, drawP2Start, dualStart, true,
		     numbered, g);
	  }
	}
      }
    }
    if (activePos != null) {
      x = activePos.x;
      y = activePos.y;
      // existing tile
      if (tiles[y][x] != null && activePos2 == null) {
	if (!(player2Start != null &&
	      player2Start.x == x && player2Start.y == y)) {
	  drawTile(gridArea.x + tileSize * x, gridArea.y + tileSize * y,
		   tiles[y][x], true, true, dualStart, false,
		   numbered, g);
	} else {
	  drawTile(gridArea.x + tileSize * x, gridArea.y + tileSize * y,
		   tiles[y][x], true, true, dualStart, true,
		   numbered, g);
	}
	// empty space
      } else if (tiles[y][x] == null && activePos2 == null) {
	g.setColor(gridHighlight);
	g.drawRect(gridArea.x + (tileSize * activePos.x),
		   gridArea.y + (tileSize * activePos.y),
		   tileSize - 1, tileSize - 1);
	// whole area
      } else {
	Geom.drawRect(Geom.tilePosToArea(frames, activePos).
		      union(Geom.tilePosToArea(frames, activePos2)),
		      gridHighlight, g);
      }
    }
  }

  
  /**
   * Renders tiles (in-game).
   */
  private void drawGTiles(GameTile[][] tiles, Rectangle gridArea,
			  int tileSize, Point activePos, Point activePos2,
			  boolean numbered, Graphics2D g) {
    g.setFont(numberFont);
    for (int y = 0; y < tiles.length; y++) {
      for (int x = tiles.length - 1; x >= 0; x--) {
	if (tiles[y][x] != null) {
	  boolean
	    isP1Pos =
	    activePos != null && y == activePos.y && x == activePos.x,
	    isP2Pos =
	    activePos2 != null && y == activePos2.y && x == activePos2.x;
	  drawGTile(gridArea.x + tileSize * x, gridArea.y + tileSize * y,
		    tiles[y][x], (isP1Pos || isP2Pos), (isP1Pos && isP2Pos),
		    isP2Pos, numbered, g);
	}
      }
    }
  }


  /**
   * Renders an individual tile (editor/level-chooser).
   */
  private void drawTile(int x, int y, Tile t, boolean highlighted,
			boolean drawStart, boolean dualStart,
			boolean p2, boolean numbered, Graphics2D g) {
    // for editor always show start tile (2nd last argument)
    drawImgTile(x, y, t, t.getHeight(), highlighted, drawStart, dualStart, p2,
		numbered, g);
  }


  /**
   * Renders an individual tile (in-game).
   */
  private void drawGTile(int x, int y, GameTile gTile, boolean highlighted,
			 boolean equalActives, boolean is2pPos,
			 boolean numbered, Graphics2D g) {
    if (!(gTile.getHeight() < 1 && !highlighted && !gTile.isDisabled())) {
      int
	tileHeight = gTile.getHeight(),
	height = tileHeight * frames.getHeightPerLevel();
      Tile t = gTile.getTile();


      if (!highlighted) {
	if (!gTile.isDisabled()) {
	  drawImgTile(x, y, t, tileHeight, false, false, false, false,
		      numbered, g);
	} else {
	  tileBase(x, y, tileHeight, gTile.getType(), highlighted, g);
	  g.drawImage(imgData.getImage(Images.O_DISABLED_TILE), null,
		      x + (tileHeight * frames.getHeightPerLevel()),
		      y - (tileHeight * frames.getHeightPerLevel()));
	  // add arrows if other player might still uses this pile
	  if (tileHeight > 0 && gTile.getType() == Tile.ARROW) {
	    drawArrows((ArwTile) t, false, x + height, y - height, g);
	  }
	}
      } else {
	int imgType;

	if (!equalActives) {
	  if (!is2pPos) {
	    imgType = !gTile.isDisabled() ?
	      Images.T_START : Images.O_DISABLED_TILE_P1;
	  } else {
	    imgType = !gTile.isDisabled() ?
	      Images.T_START2 : Images.O_DISABLED_TILE_P2;
	  }
	} else {
	  imgType = Images.O_START_BOTH;
	}

	tileBase(x, y, tileHeight, gTile.getType(), highlighted, g);
	g.drawImage(imgData.getImage(imgType), null,
		    x + (tileHeight * frames.getHeightPerLevel()),
		    y - (tileHeight * frames.getHeightPerLevel()));
	// add arrows if needed
	if (gTile.getType() == Tile.ARROW) {
	  drawArrows((ArwTile) t, true, x + height, y - height, g);
	}
	// add pile height number if needed
	if (numbered) {
	  if (tileHeight == 1) {
	    Geom.drawNumber(tileHeight, x + height + number1Offset.x,
			    y - height + number1Offset.y, numberColor, g);
	  } else {
	    Geom.drawNumber(tileHeight, x + height + numberOffset.x,
			    y - height + numberOffset.y, numberColor, g);
	  }
	}
      }
    }
  }


  /**
   * Used by <code>drawTile()</code> and <code>drawGTile()</code>.
   */
  private void drawImgTile(int x, int y, Tile t, int tileHeight,
			   boolean highlighted, boolean drawStart,
			   boolean dualStart, boolean p2, boolean numbered,
			   Graphics2D g) {
    int height = tileHeight * frames.getHeightPerLevel();

    tileBase(x, y, tileHeight, t.getType(), highlighted, g);

    if (t.getType() == Tile.NORMAL || t.getType() == Tile.ARROW) {
      // draw start if necessary
      if (drawStart && t.isStartable() && ((StartableTile) t).isStart()) {
	if (!dualStart) {
	  if (!p2) {
	    g.drawImage(imgData.getImage(Images.T_START, highlighted), null,
			x + height, y - height);
	  } else {
	    g.drawImage(imgData.getImage(Images.T_START2, highlighted), null,
			x + height, y - height);
	  }
	} else {
	  g.drawImage(imgData.getImage(Images.O_START_BOTH, highlighted), null,
		      x + height, y - height);
	}
      } else if (tileHeight > 0) {
	g.drawImage(imgData.getImage(Images.T_NORMAL, highlighted), null,
		    x + height, y - height);      	
      }
      if (t.getType() == Tile.ARROW) {
	drawArrows((ArwTile) t, highlighted, x + height, y - height, g);
      }
    } else {
      // all others
      g.drawImage(imgData.getImage(t.getType(), highlighted), null, x + height,
		  y - height);
    }
    if (numbered) {
      if (tileHeight == 1) {
	Geom.drawNumber(tileHeight, x + height + number1Offset.x,
			y - height + number1Offset.y, numberColor, g);
      } else {
	Geom.drawNumber(tileHeight, x + height + numberOffset.x,
			y - height + numberOffset.y, numberColor, g);
      }
    }
  }


  /**
   * Draws arrows (used by <code>drawGTile()</code> and
   * <code>drawImgTile()</code>.
   */
  private void drawArrows(ArwTile t, boolean highlighted, int x, int y,
			  Graphics2D g) {
    if (t.allowsDirection(ArwTile.UP)) {
      g.drawImage(imgData.getImage(Images.T_ARROW_UP, highlighted),
		  null, x, y);      	
    }
    if (t.allowsDirection(ArwTile.DOWN)) {
      g.drawImage(imgData.getImage(Images.T_ARROW_DOWN, highlighted),
		  null, x, y);      	
    }
    if (t.allowsDirection(ArwTile.LEFT)) {
      g.drawImage(imgData.getImage(Images.T_ARROW_LEFT, highlighted),
		  null, x, y);      	
    }
    if (t.allowsDirection(ArwTile.RIGHT)) {
      g.drawImage(imgData.getImage(Images.T_ARROW_RIGHT, highlighted),
		  null, x, y);      	
    }
  }


  /**
   * Renders "3D" (isometric) tile base.
   */
  private void tileBase(int x, int y, int height, int tileType,
			boolean highlighted, Graphics2D g) {

    for (int i = 0; i < height; i++) {
      g.drawImage(imgData.get3DImage(tileType, highlighted), null,
		  x + (frames.getHeightPerLevel() * i),
		  // needs to be 4 higher for 5 pixel edge
		  y - 4 - (frames.getHeightPerLevel() * i));
    }
  }


  /**
   * Draws a string with each letter on a separate tile.
   * Warning: For optimum performance <b>no checks are performed</b> on any of
   * the parameters (i.e. this method might not fail gracefully).
   * @param s the string to render
   * @param raised indicates whether each letter has a 'base' or not.
   * @param location and dimensions for this string.
   * @param g graphics context for drawing.
   */
  public void drawMenuString(String s, boolean raised, Rectangle location,
			     Graphics2D g) {
    if (!screen.scaled()) {
      Geom.fillRectMax(location, frames.getBgColor(), g);
    } else {
      Geom.fillRectOversize(location, frames.getBgColor(), g);
    }
    drawTileString(s, raised, true, false, false, location.x, location.y, g);
  }


  /**
   * Variation of <code>drawMenuString</code>. This is used to draw entries in
   * the highscore table.
   * Warning: For optimum performance <b>no checks are performed</b> on any of
   * the parameters (i.e. this method might not fail gracefully).
   * @param s the string to render
   * @param appendCursor indicates whether to add a cursor to the end of the
   * string.
   * @param first highest score is highlighted.
   * @param x absolute horizontal location to render to.
   * @param y absolute vertical location to render to.
   * @param g graphics context for drawing.
   */
  public void drawScoreString(String s, boolean appendCursor, boolean first,
			      int x, int y, Graphics2D g) {
    drawTileString(s, true, false, first, appendCursor, x, y, g);
  }


  /**
   * Draws a string with each character on an individual tile.
   */
  private void drawTileString(String s, boolean raised, boolean raisedByTwo,
			      boolean ice, boolean appendCursor, int x, int y,
			      Graphics2D g) {
    g.setFont(charFont);

    if (appendCursor) {
      drawLetterTile(' ', false, false, true, false,
		     x + (s.length() * frames.getTileSize()), y, g);
    }
    for (int i = s.length() - 1; i >= 0; i--) {
      drawLetterTile(s.charAt(i), raised, raisedByTwo, false, ice,
		     x + (i * frames.getTileSize()), y, g);
    }
  }


  /**
   * Draws a tile with a character on it.
   */
  private void drawLetterTile(char c, boolean raised, boolean raisedByTwo,
			      boolean appendCursor, boolean ice, int x, int y,
			      Graphics2D g) {
    if (!appendCursor) {
      if (c != ' ') {
	if (raised) {
	  g.drawImage(imgData.getImage((ice ? Images.O_3D_ICE : Images.O_3D),
				       !raised),
		      null, x + 5, y + 1);
	}
	if (raisedByTwo) {
	  g.drawImage(imgData.getImage((ice ? Images.O_3D_ICE :Images.O_3D),
				       !raised),
		      null, x, y + 6);
	}
	x += raised ? 10 : 5;
	y += raised ? 0 : 5;
	g.drawImage(imgData.getImage((ice ? Images.T_ICE : Images.T_NORMAL),
				     !raised),
		    null, x, y);
      }
    } else {
      g.drawImage(imgData.getImage(Images.T_START), null, x + 5, y + 5);
    }
    
    Geom.drawChar(c, x + charOffset.x, y + charOffset.y, charColor, g);
  }

}
