/* ========================================================================= *
 * JTilePile (C) 2008 by V.Termanis                                          *
 * TilePile game concept (C) 2008 by J.Mitchell                              *
 * ========================================================================= */
package JTP;

/**
 * Class storing timing information (used by <code>GameTimer</code>).
 * @author V.Termanis
 */
public class TimeStore {

  private long startTime, stopTime, accumulatedTime, timeLimit;
 
  
  /**
   * Sets time limit to use.
   * @param timeLimit time in seconds.
   * @throws IllegalArgumentException if <code>timeLimit < 0</code>.
   */
  public void setTimeLimit(int timeLimit) throws IllegalArgumentException {
    if (timeLimit < 0) {
      throw new IllegalArgumentException("time is less than zero");
    }
    this.timeLimit = (long) timeLimit * 1000;
  }

  
  /**
   * Returns time time limit set.
   * @return time limit in seconds.
   */
  public long getTimeLimit() {
    return timeLimit;
  }
 
  
  /**
   * Sets start time to current timestamp.
   */
  public void start() {
    startTime = System.currentTimeMillis();
    stopTime = 0;
  }


  /**
   * Returns whether start point has been set.
   * @return <code>true</code> if <code>start()</code> has been called
   * since construction or last <code>reset()</code> call.
   */
  public boolean hasStarted() {
    return (startTime > 0);
  }


  /**
   * Adjust start time to take into account pause period caused by
   * call to <code>stop()</code>.
   */
  public void resume() {
    if (stopTime > 0) {
      startTime += System.currentTimeMillis() - stopTime;
      stopTime = 0;
    } else {
      start();
    }
  }


  /**
   * Sets stop time to current timestamp assuming <code>start()</code> has been
   * called before. A second call to this method without calls to either
   * <code>start()</code> or <code>resume()</code> will have no effect.
   */
  public void stop() {
    if (startTime > 0 && stopTime == 0) {
      stopTime = System.currentTimeMillis();
    }
  }


  /**
   * Returns time added up by using <code>incrementAccumulatedTime()</code>
   * method.
   * @return accumulated time in milliseconds.
   */
  public long getAccumulatedTime() {
    return accumulatedTime;
  }


  /**
   * Returns time added up by using <code>incrementAccumulatedTime()</code>
   * method, in seconds. Note: The returned value will have been rounded up
   * to the nearest second.
   * @return accumulated time in seconds.
   */
  public long getAccumulatedTimeSeconds() {
    return secCeiling(accumulatedTime);
  }


  /**
   * Adds the currently leftover time to the total.
   * Note: It only makes sense to call this after the timer has been stopped
   * and before it has been reset.
   */
  public void incrementAccumulatedTime() {
    accumulatedTime += timeLimit - (stopTime - startTime);
  }


  /**
   * Returns time remaining for given time limit. 
   * @return time in milliseconds, and never less than zero.
   */
  public long timeLeft() {
    if (startTime == 0) {
      return timeLimit;
    } else {
      long timeLeft;
      if (stopTime == 0) {
	timeLeft = timeLimit - (System.currentTimeMillis() - startTime);
      } else {
	timeLeft = timeLimit - (stopTime - startTime);
      }
      return (timeLeft > 0) ? timeLeft : 0;
    }
  }


  /**
   * Resets start and stop times.
   */
  public void reset() {
    startTime = stopTime = 0;
  }


  /**
   * Resets the time added up by using <code>incrementAccumulatedTime()</code>
   * method back to zero for active timer.
   */
  public void resetAccumulatedTime() {
    accumulatedTime = 0;
  }


  /**
   * Returns time remaining in <code>mm:ss</code> format.
   * @return time in <code>String</code> representation.
   */
  public String toString() {
    long inSeconds = secCeiling(timeLeft()), seconds = (inSeconds % 60);
    String string = Long.toString(inSeconds / 60) + " : ";
    if (seconds < 10) {
      string += "0";
    }
    string += Long.toString(seconds);
    return string;
  }


  /**
   * Rounds time up to the nearest second.
   */
  private static long secCeiling(long milliSeconds) {
    if (milliSeconds != 0) {
      long sec = milliSeconds / 1000;
      if (milliSeconds - (sec * 1000) > 0) {
	sec += 1;
      }
      return sec;
    } else {
      return 0;
    }
  }

}